/** PalmOS data base i/o for GPilotS.
  @file dbio.c
  @author C. Claveleira
  @date 1999-2004
  
  Handles transfers between gpslib and data bases,
  active track setup and feeding and beaming (experimental)
*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
 */

/* 	$Id: dbio.c,v 1.11 2004/11/17 21:21:00 clavelei Exp $	 */

#ifndef lint
// static char vcid[] = "$Id: dbio.c,v 1.11 2004/11/17 21:21:00 clavelei Exp $";
#endif /* lint */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "gtalkRsc.h"
#include "gpslib.h"
#include "gtalk.h"
#include "gpsutil.h"
#include "log.h"
#include "dbio.h"
#include "gpsio.h"

static Raw_D210_Rte_Link_Type rteLink = {
    {0, 0},
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    {0}
};


/**
   Writes a record contents and attributes.
 Record is released on return.

 @param  RecH handle on the record (may be null)
 @param  p pointer on datas to put in record (if null the record is cleared)
 @param  l size of datas
 @param  DB pointer on an open data base
 @param  pos position of the record in the data base
 @param  Category category to put the record 
 @return error code, 0 if no error
*/

Err
SetRecord (VoidHand RecH, void *p, ULong l, DmOpenRef DB, UInt pos, UInt Category)
{
    Ptr RecP;
    Err error;

    if (!RecH)
	return (FrmAlert (NoRoomID) - 1);

    if (MemHandleSize (RecH) != l)
	RecH = DmResizeRecord (DB, pos, l);		   /* need to resize record ? */
    if (!RecH)
	return (FrmAlert (NoRoomID) - 1);

    RecP = MemHandleLock (RecH);
    if (p)
	error = DmWrite (RecP, 0, p, l);
    else
	error = DmSet (RecP, 0L, l, 0);
    ErrFatalDisplayIf (error, "Could not write to new record.");
    MemHandleUnlock (RecH);
    DmReleaseRecord (DB, pos, true);
    error = SetCategory (DB, pos, Category);
    SetBackupBit(DB);
    return (error);
}							   // SetRecord


/**
   Sets a record's category and dmRecAttrDirty flag

 @param  DB data base open reference
 @param  index index of the record
 @param  Category category to set the record into
 @return 0 if OK
*/

Err
SetCategory (DmOpenRef DB, UInt index, UInt Category)
{
    UInt attributes;
    Err error;

    error = DmRecordInfo (DB, index, &attributes, NULL, NULL);
    if (error)
	return (error);
    //    ErrFatalDisplayIf (error, "Can't read record information");
    attributes &= ~dmRecAttrCategoryMask;
    attributes |= ((Category < 0xF ? Category : 0) | dmRecAttrDirty);
    error = DmSetRecordInfo (DB, index, &attributes, NULL);
    ErrFatalDisplayIf (error, "Can't set record information");
    return (0);
}   /* SetCategory */



#ifdef EVENTS


/**
   Callback to get a raw packet from data base.
 Extracts a raw packet from data base pointed by LogDB

 @param  current number of the packet
 @return pointer on the retrieved packet
*/

Garmin_Packet *
get_next_packet (UInt current)
{
    VoidHand RecH = NULL;
    static UInt c;
    static Garmin_Packet packet;
    Garmin_Packet *r,
     *RecP;
    ULong l;

    if (current == 0)
	c = 0;
    RecH = DmQueryNextInCategory (LogDB, &c, dmAllCategories);
    c++;
    if (RecH)
    {
	RecP = MemHandleLock (RecH);
	l = MemHandleSize (RecH);
	if (l > sizeof (packet))
	    l = sizeof (packet);
	(void) MemMove (&packet, RecP, l);
	MemHandleUnlock (RecH);
	Sent++;
	r = &packet;
    }
    else
	r = NULL;
    return (r);
}                  /* get_next_packet */


/**
   Callback to stores a raw Garmin packet into the packet data base.

 @param  p pointer on raw packet
 @param  current number of packet
 @return error code (0 if ok)
*/

Err
put_packet (Garmin_Packet * p, UInt current)
{
    UInt pos = dmMaxRecordIndex;
    static VoidHand RecH = NULL;

    RecH = DmNewRecord (LogDB, &pos, p->length);
    if (SetRecord (RecH, p, p->length, LogDB, pos, dmUnfiledCategory))
	return (-1);
    New++;
    return (0);
}              /* put_packet */
#endif


/**
   stores a raw waypoint in waypoint data base. Used as callback by gpslib

 @param  p pointer on a raw waypoint (as received from gps)
 @param  total total count of item sent by gps
 @param  current current item number
 @param  format format of waypoint
 @param  size size of waypoint
 @return 0 if OK
 */

Err
put_wpt (GarminObject * p, UInt total, UInt current, LocalFormatID format, unsigned char size)
{
    UInt pos,
      differ;
    VoidHand RecH;
    Ptr RecP;
    Err error;
    Object o;

    switch (format)
    {
    case G100Type:
    case G102Type:
    case G103Type:
    case G104Type:
    case G107Type:
    case G108Type:
    case G109Type:
      InitObject(&o);
	o.header.size = size;
	MemMove (&(o.d), p, o.header.size);
	o.header.type = format;
	SortWpts (Alpha);				   /* il faut que la base soit trie alphabtiquement */
	error = 0;
	// insertion dans la base de donnees :
	pos = DmFindSortPosition (WptDB, (VoidPtr) & o, NULL, (DmComparF *) & CompareWpt, 0);
	differ = 1;
	if (pos > 0)
	{						   /* duplicata ? */
	    RecH = DmGetRecord (WptDB, pos - 1);	   /* regardons precedent record */
	    RecP = MemHandleLock (RecH);
	    differ = CompareWpt (&o, RecP, 0, NULL, NULL, NULL);
	    MemHandleUnlock (RecH);
	    DmReleaseRecord (WptDB, pos - 1, false);
	}
	if (differ)
	    goto new;
	else if (!OnlyNews)
	{
	    if (!Prefs.allowduplicate)
	    {						   /* record de meme nom, on remplace l'ancien */
		RecH = DmGetRecord (WptDB, pos - 1);
		error = SetRecord (RecH, &o, SizeOfObject (&o), WptDB, pos - 1, Prefs.wptcategory);
		if (!error)
		    Replaced++;
	    }
	    else
	    {						   /* creation nouveau record */
	      new:RecH = DmNewRecord (WptDB, &pos, SizeOfObject (&o));
		error = SetRecord (RecH, &o, SizeOfObject (&o), WptDB, pos, Prefs.wptcategory);
		if (!error)
		    New++;
	    }
	}

	break;

    default:
	FrmCustomAlert (UnsupportedID, "waypoint", "", "");
	error = -2;
    }							   /* switch */

    return (error);
}							   // put_wpt


/**
   Callback to store a route in the current routes data base.

 @param  p pointer on the raw object from the gps
 @param  total total count of items sent by gps
 @param  current number of current item (first=1)
 @param  format format of the object (route header or waypoint)
 @param  size size of object
 @return error code (0 if no error)
*/

Err
put_rte (GarminObject * p, UInt total, UInt current, LocalFormatID format, unsigned char size)
{
    Err error;
    static UInt pos;
    ULong l;
    static VoidHand RecH;
    Ptr RecP;
    //    struct RteDB *t = p;
    //    byte *cp = p;
    //    void *v;
    Object o;

    debug_in (GTALKDBG_CALLS,
	      ("->put_rte(total=%d,current=%d,format=%d,size=%d", total, current, format, size));

    error = 0;
    InitObject(&o);
    o.header.size = size;
    MemMove (&(o.d), p, o.header.size);
    o.header.type = format;

    switch (format)
    {
    case G200Type:
    case G201Type:
    case G202Type:
    case CustomRteHdr:					   /* utilise en interne pour format custom */
	pos = dmMaxRecordIndex;				   /* reception header,initialisation du record */
	RecH = DmNewRecord (RteDB, &pos, SizeOfObject (&o));
	error = SetRecord (RecH, &o, SizeOfObject (&o), RteDB, pos, Prefs.rtecategory);
	if (!error)
	    New++;
	break;

    case G210Type:					   /* route link */
#ifndef SUPPORT_LINKS
	break;						   /* we drop links */
#endif

    case G100Type:					   /* waypoints */
    case G102Type:
    case G103Type:
    case G104Type:
    case G107Type:
    case G108Type:
    case G109Type:
	if (RecH)
	{						   // reception waypoint
	    l = MemHandleSize (RecH);
	    RecH = DmResizeRecord (RteDB, pos, l + SizeOfObject (&o));
	    RecH = DmGetRecord (RteDB, pos);
	    if (RecH)
	    {
		RecP = MemHandleLock (RecH);
		error = DmWrite (RecP, l, &o, SizeOfObject (&o));
		MemHandleUnlock (RecH);
		error = DmReleaseRecord (RteDB, pos, true);
	    }
	}
	break;

    default:
	FrmCustomAlert (UnsupportedID, "route element", "", "");
	error = -2;
    }
    debug_out (GTALKDBG_CALLS, ("<-put_rte returns %d", error));

    return (error);
}							   // put_rte


/**
   Initialise a track header object from a Garmin raw header and a (first) track point

 @param  trkhdrp pointer on a track header object
 @param  format format of the Garmin object
 @param  p pointer on the Garmin object
 @param  trkp pointer on track points
 @return nothing
*/

static void
MakeTrackHeader (Object * trkhdrp, LocalFormatID format, GarminObject * p, TrkPts * trkp)
{
    D310_Trk_Hdr_Type *p310;
    DateTimeType dtp;
    char date[dateStringLength];
    char time[timeStringLength];

    InitObject(trkhdrp);
    //    trkhdrp->header.version = OBJECTVERSION;
    trkhdrp->header.type = CustomTrkHdr;
    trkhdrp->header.size = sizeof (Custom_Trk_Hdr_Type);
    trkhdrp->d.CustTrkHdr.dspl = trkhdrp->d.CustTrkHdr.color = 0;
    trkhdrp->d.CustTrkHdr.type = CompactTracks ? CompactTrkPt : CustomTrkPt;	/* format des trk pts a suivre */
    trkhdrp->d.CustTrkHdr.number = 0;
    trkhdrp->d.CustTrkHdr.latmin = LATMAX;
    trkhdrp->d.CustTrkHdr.latmax = LATMIN;
    trkhdrp->d.CustTrkHdr.lonmin = LONMAX;
    trkhdrp->d.CustTrkHdr.lonmax = LONMIN;
    if (format == G310Type)
    {
	//        trkhdrp->header.type = G301Type;
	p310 = (D310_Trk_Hdr_Type *) p;
	StrNCopySafe (trkhdrp->d.CustTrkHdr.name, p310->trk_ident,
		      sizeof (trkhdrp->d.CustTrkHdr.name));
	trkhdrp->d.CustTrkHdr.dspl = p310->dspl;
	trkhdrp->d.CustTrkHdr.color = p310->color;
    }
    else
    {
	if (CompactTracks)
	    TimSecondsToDateTime (TimGetSeconds (), &dtp);
	else
	    TimSecondsToDateTime (GarminTime2PalmSeconds (trkp->custom.time), &dtp);
	DateToAscii (dtp.month, dtp.day, dtp.year, dfDMYWithDashes, date);
	TimeToAscii (dtp.hour, dtp.minute, tfColon24h, time);
	StrNCopySafe (trkhdrp->d.CustTrkHdr.name, date, sizeof (trkhdrp->d.CustTrkHdr.name));
	StrNCat (trkhdrp->d.CustTrkHdr.name, " ", sizeof (trkhdrp->d.CustTrkHdr.name));
	StrNCat (trkhdrp->d.CustTrkHdr.name, time, sizeof (trkhdrp->d.CustTrkHdr.name));
    }
}							   /* MakeTrackHeader */


/**
   Callback to store tracks in the current tracks data base

  @param  p pointer on a raw track point or track header received from gps
  @param  total total count of items sent by gps
  @param  current current item number (first=1)
  @param  format format of the object received
  @param  size size of object
  @return error code (0 if no error)
*/

Err
put_trk (GarminObject * p, UInt total, UInt current, LocalFormatID format, unsigned char size)
{
    Err error;
    static UInt pos;
    static VoidHand RecH = NULL;
    Ptr RecP;
    UInt sizeoftrkpt = 0;
    static ULong l = 0;
    static Object trkhdr;
    TrkPts trk;

    debug (GTALKDBG_CALLS,
	   ("<->put_trk(total=%d,current=%d,format=%d,size=%d)", total, current, (int) format,
	    (int) size));
    if (format != G310Type)
    {
	sizeoftrkpt =
	    AnyRawToAnyTrkPt (format, CompactTracks ? CompactTrkPt : CustomTrkPt, p, &trk);
	if (sizeoftrkpt == 0)
	{
	    FrmCustomAlert (UnsupportedID, "track", "", "");
	    return (GErrUnsupportedFmt);
	}
    }							   /* switch */

    if (current > total)
	return (GErrProtocolError);

    if ((current == 1) || (format == G310Type))
    {							   /* initialisation du record */
	if (current != 1)
	    CleanUpTracks ();				   /* cleaning of previous track */
	MakeTrackHeader (&trkhdr, format, p, &trk);	   /* init du header */
	RecH = NULL;
	if (format == G310Type)
	    return (0);
    }
    if ((RecH == NULL) && (format != G310Type))
    {							   /* initialisation du record */

	l = SizeOfObject (&trkhdr) + (long) sizeoftrkpt *(long) total - current + 1;	/* espace qu'il faudrait */
	//debug (GTALKDBG_INFOS, ("put_trk : space needed : %ld bytes",l)); // $$
	if (l > MaxRecordSize)
	    l = (((MaxRecordSize - SizeOfObject (&trkhdr)) / sizeoftrkpt)) * sizeoftrkpt + SizeOfObject (&trkhdr);	/* ce que l'on peut mettre */
	//debug (GTALKDBG_INFOS, ("put_trk : allocated %ld bytes record (%d b/trkpt)", l,sizeoftrkpt));  // $$

	pos = dmMaxRecordIndex;
	RecH = DmNewRecord (TrkDB, &pos, l);
	error = SetRecord (RecH, NULL, l, TrkDB, pos, Prefs.trkcategory);
	if (error)
	    return (error);				   /* initialisation */
	l = SizeOfObject (&trkhdr);
	New++;
    }

    if (trk.custom.posn.lat > trkhdr.d.CustTrkHdr.latmax)
	trkhdr.d.CustTrkHdr.latmax = trk.custom.posn.lat;
    else if (trk.custom.posn.lat < trkhdr.d.CustTrkHdr.latmin)
	trkhdr.d.CustTrkHdr.latmin = trk.custom.posn.lat;
    if (trk.custom.posn.lon > trkhdr.d.CustTrkHdr.lonmax)
	trkhdr.d.CustTrkHdr.lonmax = trk.custom.posn.lon;
    else if (trk.custom.posn.lon < trkhdr.d.CustTrkHdr.lonmin)
	trkhdr.d.CustTrkHdr.lonmin = trk.custom.posn.lon;
    //debug (GTALKDBG_INFOS, ("<->put_trk(format=%d,size=%d) : lat=%ld, lon=%ld", format, (int) size,   trk.custom.posn.lat, trk.custom.posn.lon));
    if (RecH)
    {
	if (l >= MemHandleSize (RecH))
	{
	    error = put_trk (p, total - current + 1, 1, format, size);	/* essayons un nouveau record */
	    return (error);
	}
	RecH = DmGetRecord (TrkDB, pos);
	RecP = MemHandleLock (RecH);
	trkhdr.d.CustTrkHdr.number++;
	error = DmWrite (RecP, 0, &trkhdr, SizeOfObject (&trkhdr));	/* header */
	error = DmWrite (RecP, l, &trk, sizeoftrkpt);
	ErrFatalDisplayIf (error, "Can't write trackpoint");
	l += sizeoftrkpt;
	MemHandleUnlock (RecH);
	DmReleaseRecord (TrkDB, pos, true);
	return (GErrNoError);
    }

    else
	return (GErrNoHandle);
}							   // put_trk


static int active_trk_index = -1;
#define ACTIVETRKNAME "ACTIVE TRACK"


/**
   Creates and initialises an active track in the current track data base.
 The record has a particuliar name defined in ACTIVETRKNAME.

 @return index of record containing active track
*/

UInt
init_active_trk ()
{
    VoidHand RecH = NULL;
    Object *op,
      trkhdr;
    UInt t;
    Boolean gotit = 0;

    for (t = DmNumRecords (TrkDB); t > 0; t--)
    {
	RecH = DmQueryRecord (TrkDB, t - 1);
	if (RecH)
	{
	    op = MemHandleLock (RecH);
	    if (op->header.type == CustomTrkHdr)
	    {
		if (!StrCompare (op->d.CustTrkHdr.name, ACTIVETRKNAME))
		    gotit = 1;
	    }
	    MemHandleUnlock (RecH);
	}
	if (gotit)
	    break;
    }							   // for
    if (!t)
    {
	RecH = DmNewRecord (TrkDB, &t, SizeOfObject (&trkhdr));	/* factorisable avec put_trk ? */
	InitObject(&trkhdr);
		   //	trkhdr.header.version = OBJECTVERSION;
	trkhdr.header.type = CustomTrkHdr;
	trkhdr.header.size = sizeof (Custom_Trk_Hdr_Type);
	trkhdr.d.CustTrkHdr.dspl = trkhdr.d.CustTrkHdr.color = 0;
	trkhdr.d.CustTrkHdr.type = CustomTrkPt;		   /* format des trk pts a suivre */
	trkhdr.d.CustTrkHdr.number = 0;
	trkhdr.d.CustTrkHdr.latmin = LATMAX;
	trkhdr.d.CustTrkHdr.latmax = LATMIN;
	trkhdr.d.CustTrkHdr.lonmin = LONMAX;
	trkhdr.d.CustTrkHdr.lonmax = LONMIN;
	StrNCopySafe (trkhdr.d.CustTrkHdr.name, ACTIVETRKNAME, sizeof (trkhdr.d.CustTrkHdr.name));
	SetRecord (RecH, &trkhdr, SizeOfObject (&trkhdr), TrkDB, t, dmUnfiledCategory);
	active_trk_index = t;
    }
    else
	active_trk_index = t - 1;
    return (active_trk_index);
}							   // init_active_trk


/**
   Pushes a track point in the active track.
 The active track must be initialised.

@param  trkpt pointer on the track point
@return nothing
*/

void
push_active_trkpt (Custom_Trk_Point_Type trkpt)
{
    static VoidHand RecH = NULL;
    Ptr RecP;
    static Object trkhdr;
    ULong l;
    Err error;

    if (active_trk_index >= 0)
    {
	RecH = DmQueryRecord /*  DmGetRecord */ (TrkDB, active_trk_index);
	if (RecH)
	{
	    RecP = MemHandleLock (RecH);
	    MemMove (&trkhdr, RecP, SizeOfObject ((ObjectPtr) RecP));
	    trkhdr.d.CustTrkHdr.number++;
	    if (trkpt.posn.lat > trkhdr.d.CustTrkHdr.latmax)
		trkhdr.d.CustTrkHdr.latmax = trkpt.posn.lat;
	    else if (trkpt.posn.lat < trkhdr.d.CustTrkHdr.latmin)
		trkhdr.d.CustTrkHdr.latmin = trkpt.posn.lat;
	    if (trkpt.posn.lon > trkhdr.d.CustTrkHdr.lonmax)
		trkhdr.d.CustTrkHdr.lonmax = trkpt.posn.lon;
	    else if (trkpt.posn.lon < trkhdr.d.CustTrkHdr.lonmin)
		trkhdr.d.CustTrkHdr.lonmin = trkpt.posn.lon;
	    MemHandleUnlock (RecH);
	}
	l = SizeOfObject (&trkhdr) + trkhdr.d.CustTrkHdr.number * sizeof (Custom_Trk_Point_Type);
	RecH = DmResizeRecord (TrkDB, active_trk_index, l);
	if (RecH)
	{
	    RecH = DmGetRecord (TrkDB, active_trk_index);
	    RecP = MemHandleLock (RecH);
	    error = DmWrite (RecP, 0, &trkhdr, SizeOfObject (&trkhdr));	/* header */
	    error =
		DmWrite (RecP, l - sizeof (Custom_Trk_Point_Type), &trkpt,
			 sizeof (Custom_Trk_Point_Type));
	    ErrFatalDisplayIf (error, "Can't write trackpoint");
	    MemHandleUnlock (RecH);
	    DmReleaseRecord (TrkDB, active_trk_index, true);
	}
	else
	    FrmCustomAlert (ErrorID, "Can't store new track point", " ", " ");
    }
}							   /* push_active_trkpt */


/**
   Extracts next waypoint from current wpt category
   and converts it to specified format

 @param current if 0, the first waypoint is returned
 @param format wanted format of waypoint
 @param size pointer on a byte to write returned object size
 @return pointer on converted  waypoint
*/

GarminObject *
get_next_wpt (UInt current, LocalFormatID format, unsigned char *size)
{
    static Object obj;
    Object *RecP, *p;
    GarminObject *r;
    VoidHand RecH = NULL;
    static UInt c;

    debug_in (GTALKDBG_CALLS, ("->get_next_wpt(current=%d,format=%d)", current, format));
    if (current == 0)
	c = 0;
#ifdef SELECT
    if (Prefs.selectmode)
    {
	while ((RecH = DmQueryNextInCategory (WptDB, &c, Prefs.wptcategory)))
	{
	    if (RecIsSecret (WptDB, c))
		break;
	    else
		c++;
	}
    }
    else
#endif
	RecH = DmQueryNextInCategory (WptDB, &c, Prefs.wptcategory);
    c++;
    if (RecH)
      {
	RecP = MemHandleLock (RecH);
	//      (void) memcpy (&buf, RecP, sizeof (Raw_D103_Wpt_Type));
	p=AnyToAnyWpt (RecP, &obj, format);
	MemHandleUnlock (RecH);
	if (!p){
	  FrmCustomAlert (UnsupportedID, "requested waypoint", "", "");
	  r=NULL;
	}
	else{
	  Sent++;
	  r = &obj.d.g;
	  *size = obj.header.size;
	}
      }
    else
	r = NULL;
    debug_out (GTALKDBG_CALLS, ("<-get_next_wpt()=%ld, *size=%d", r, (int) *size));
    return (r);
}							   // get_next_wpt


/**
   Extracts next element of track from current data base record.
   Process each record of current category.
   Used as callback by gpslib

 @param  current index of current element (first=0)
 @param  format format of element requested
 @param size pointer to store size of element returned
 @param eot true if last element of current track
 @return pointer on element returned (Garmin format)
 */

GarminObject *
get_next_trk (UInt current, LocalFormatID format, unsigned char *size, Boolean * eot)
{
    static Raw_D300_Trk_Point_Type nt300;
    static Raw_D301_Trk_Point_Type nt301;
    static D310_Trk_Hdr_Type trkhdr;
    Custom_Trk_Point_Type *customp;
    Compact_Trk_Point_Type *compactp;
    GarminObject *r;
    //    struct TrkDB *RecP;
    Object *RecP;
    VoidHand RecH = NULL;
    static UInt count;
    static UInt c,
      t;

    debug_in (GTALKDBG_CALLS, ("->get_next_trk(current=%d,format=%d)", current, format));
    r = NULL;
    if ((current == 0) || (count == 0))
    {
	if (current == 0)
	    c = 0;
	else
	    c++;
#ifdef SELECT
	if (Prefs.selectmode)
	{
	    while ((RecH = DmQueryNextInCategory (TrkDB, &c, Prefs.trkcategory)))
	    {
		if (RecIsSecret (TrkDB, c))
		    break;
		else
		    c++;
	    }
	}
	else
#endif
	    RecH = DmQueryNextInCategory (TrkDB, &c, Prefs.trkcategory);
	if (!RecH)
	    goto f;
	count = t = TrkptsInTrkDB (RecH);
	Sent++;
    }

    RecH = DmQueryRecord (TrkDB, c);
    if (RecH)
    {
	RecP = MemHandleLock (RecH);
	customp = (Custom_Trk_Point_Type *) ((void *) RecP + SizeOfObject (RecP));
	compactp = (Compact_Trk_Point_Type *) customp;
	customp += t - count;
	compactp += t - count;
	if (format == G300Type)
	{						   /* conversion en format network */
	    if (RecP->d.CustTrkHdr.type == CustomTrkPt)
		CustomToRawD300TrkPt (customp, &nt300);
	    else
		CompactToRawD300TrkPt (compactp, &nt300);
	    r = (GarminObject *) & nt300;
	    *size = sizeof (nt300);
	}
	else if (format == G301Type)
	{
	    if (RecP->d.CustTrkHdr.type == CustomTrkPt)
		CustomToRawD301TrkPt (customp, &nt301);
	    else
		CompactToRawD301TrkPt (compactp, &nt301);
	    r = (GarminObject *) & nt301;
	    *size = sizeof (nt301);
	}
	else if (format == G310Type)
	{
	    CustomTrkHdrToD310 (RecP, &trkhdr);
	    r = (GarminObject *) & trkhdr;
	    *size = sizeof (trkhdr) - sizeof (trkhdr.trk_ident) + StrLen (trkhdr.trk_ident) + 1;
	    count++;
	}
	MemHandleUnlock (RecH);
    }
    if (--count == 0)
	*eot = 1;
    else
	*eot = 0;
  f:
    debug_out (GTALKDBG_CALLS, ("<-get_next_trk()=%ld, *size=%d", r, (int) *size));
    return (r);
}							   // get_next_trk


/**
   Extracts next sequential route item from current route data base.
 The item is converted to specified format

 @param  current number of current item (header or waypoint, first=0)
 @param  format wanted format of route
 @param size pointer on a byte to write returned object size
 @param  header pointer on a boolean set if item is a header
 @param  eor pointer on a boolean set if it's the last element
 @return pointer on converted item
*/

GarminObject *
get_next_rte (UInt current, LocalFormatID format, unsigned char *size, Boolean * header,
	      Boolean * eor)
{
    Object *objp,
     *RecP;
    static Object obj;
    GarminObject *gop = NULL;
    VoidHand RecH = NULL;
    static int count;
    static UInt c,
      t;

    debug_in (GTALKDBG_CALLS, ("->get_next_rte(current=%d,format=%d", current, current, format));
    *header = 0;
    if ((current == 0) || (count == 0))
    {
	if (current == 0)
	    c = 0;
	else
	    c++;
	*header = 1;
#ifdef SELECT
	if (Prefs.selectmode)
	{
	    while ((RecH = DmQueryNextInCategory (RteDB, &c, Prefs.rtecategory)))
	    {
		if (RecIsSecret (RteDB, c))
		    break;
		else
		    c++;
	    }
	}
	else
#endif
	    RecH = DmQueryNextInCategory (RteDB, &c, Prefs.rtecategory);
	if (RecH)
	{
	    count = t = ObjectsInRec (RecH);
	    Sent++;
	}
    }

    //    count--;
    RecH = DmQueryRecord (RteDB, c);
    if (RecH && count)
    {
	RecP = MemHandleLock (RecH);
	objp = (void *) RecP + OffsetOfObjectInRec (RecH, t - count);
	if (count-- == t)
	    AnyToAnyRteHdr (objp, &obj, format);	   /* header ? */
	else if ((format == G210Type) && (objp->header.type != G210Type))
	{						   /* route link */
	    InitObject(&obj);
	    obj.header.type = G210Type;			   /* s'il n'existe pas on en fabrique un */
	    obj.header.size = sizeof (rteLink);
	    //	    obj.header.version = OBJECTVERSION;
	    MemMove (&(obj.d), &rteLink, obj.header.size);
	    count++;					   /* compensation */
	}
	else
	    AnyToAnyWpt (objp, &obj, format);
	if (count == 0)
	    *eor = 1;					   /* dernier elt de la route ? */
	*size = obj.header.size;
	MemHandleUnlock (RecH);
	gop = &obj.d.g;
    }
    debug_out (GTALKDBG_CALLS,
	       ("<-get_next_rte returns size=%d, header=%d, eor=%d%d", *size, *header, *eor));
    return (gop);
}							   // get_next_rte


#ifdef FURUNO
/**
   Return next route in Furuno format


 @param  current number of item wanted
 @return pointer on string to send to gps
*/

char *
get_next_full_rte (UInt current)
{
    byte *RecP;
    VoidHand RecH;
    UInt i,
      j,
      n = 0,
      NumWpts;
    static UInt c;
    Raw_D103_Wpt_Type *p;
    char *e = NULL;
    Object *objp;
    Custom_Wpt_Type custwp;

    if (current == 0)
	c = 0;
    else
	c++;
    RecH = DmQueryNextInCategory (RteDB, &c, Prefs.rtecategory);
    if (RecH)
    {
	RecP = MemHandleLock (RecH);
	NumWpts = WptsInRteDB (RecH);
	for (i = 0; i <= ((NumWpts - 1) / 8); i++)
	{
	    n += StrPrintF (gbuf1 + n, "$GPRTE,%d,%d,C,0%d", 1 + (NumWpts - 1) / 8, i + 1,
			    ((D201_Rte_Hdr_Type *) RecP)->nmbr);
	    for (j = i * 8; (j < (i + 1) * 8) && (j < NumWpts); j++)
	    {
	  objp=(void *) RecP + OffsetOfObjectInRec (RecH, j+1);
	  PtrToCustomWpt (objp,&custwp);
/* 		p = (Raw_D103_Wpt_Type *) (RecP + sizeof (D201_Rte_Hdr_Type) + */
/* 					   j * sizeof (Raw_D103_Wpt_Type)); */
		StrNCopySafe (gbuf2, custwp.ident, sizeof (custwp.ident) + 1);
		n += StrPrintF (gbuf1 + n, ", %s", gbuf2);
	    }
	    n += StrPrintF (gbuf1 + n, "\r\n");
	}
	MemHandleUnlock (RecH);
	e = gbuf1;
    }
    return (e);
}							   /* get_next_full_rte */


/**
   Extracts next waypoint from current wpt category
   and converts it to Furuno format

@param  current number of item (first=0)
@return opinter on string in Furuno format
*/

char *
Furuno_get_next_wpt (UInt current)
{
    GarminObject *nwp;
    //    D103_Wpt_Type wp;
    char name[7],
      comment[17],
     *r = NULL;
    //    static char buf[82];
    double latitude,
      longitude;
    unsigned char size;
    int n;
    Custom_Wpt_Type custwp;

    debug_in (GTALKDBG_CALLS, ("->Furuno_get_next_wpt(current=%d)", current));
    nwp = get_next_wpt (current, G103Type, &size);
    if (nwp != NULL)
    {
	//StrPrintF (buf, "$PFEC,GPxfr,CTL,E\r\n");
	//        return (NULL);

	//        D103_ntoh (nwp, &wp);
      G103ToCustom (&(nwp->d103),&custwp );

	StrNCopySafe (name, custwp.ident, sizeof (name));
	StrNCopySafe (comment, custwp.cmnt, sizeof (comment));
	SemiToDegre (&custwp.posn, &latitude, &longitude);
	//	translate (FromWGS84, &latitude, &longitude, DatumIndex);
	n = StrPrintF (gbuf1, "$PFEC,GPwpl,%s,", DegreToDM (latitude, 0, 1));
	n += StrPrintF (gbuf1 + n, "%s,%s,,@%c%s,A,,,,\r\n", DegreToDM (longitude, 1, 1), name,
			(char) ('a' + 0x10 + (custwp.smbl % 9)), comment);
	r = gbuf1;
    }
    debug_out (GTALKDBG_CALLS, ("<-Furuno_get_next_wpt(r=%ld)", r));
    return (r);
}							   /* Furuno_get_next_wpt */
#endif


/**
   Cleaning of tracks.
 Check tracks to recover unused space in case of interrupted transfers.

@return nothing
 @bug : should use d.CustTrkHdr.number
*/

void
CleanUpTracks ()
{
    UInt n,
      c,
      hdrsize,
      t;
    char *p,
     *pp;
    ULong newsize,
      size;
    Long l;
    VoidHand RecH = NULL;
    Object *op;

    c = 0;
    for (t = DmNumRecordsInCategory (TrkDB, Prefs.trkcategory); t > 0; t--)
    {
	RecH = DmQueryNextInCategory (TrkDB, &c, Prefs.trkcategory);
	if (RecH)
	{
	    op = MemHandleLock (RecH);
	    size =
		op->d.CustTrkHdr.type ==
		CompactTrkPt ? sizeof (Compact_Trk_Point_Type) : sizeof (Custom_Trk_Point_Type);
	    p = (char *) op;
	    hdrsize = SizeOfObject (op);
	    for (pp = p + MemHandleSize (RecH) - 1; (pp > p) && !(*pp); pp--);
	    l = pp - p;
	    MemHandleUnlock (RecH);
	    if (l > hdrsize)
		n = (l - hdrsize) / size + 1;
	    else
		n = 0;
	    newsize = hdrsize + n * size;
	    if (newsize != MemHandleSize (RecH))
		DmResizeRecord (TrkDB, c, newsize);
	}
	c++;
    }
}							   // CleanUpTracks

#ifdef BEAM

/**
   Receives beamed datas
   Warning : globals not allowed

   @param  sockp pointer on ExgSocket structure (from exchange manager)
   @param  DB pointer on data base to put datas
   @return 0 if OK
 */

Err
BeamReceive (ExgSocketPtr sockp, DmOpenRef DB)
{
    Err err;
    MemHandle dataH,
      RecH;
    UInt16 size,
      ChunkSize = 1024;
    UInt8 *dataP;
    UInt16 dataLen = 0,
      idx,
      len;

    size = ChunkSize;
    dataH = MemHandleNew (size);			   /* first chunk */
    if (dataH)
    {
	err = ExgAccept (sockp);
	if (!err)
	{
	    dataP = MemHandleLock (dataH);
	    do
	    {
		len = ExgReceive (sockp, dataP + dataLen, size - dataLen, &err);
		if (len && !err)
		{
		    dataLen += len;
		    if (dataLen >= size)
		    {
			MemHandleUnlock (dataH);
			err = MemHandleResize (dataH, size + ChunkSize);
			dataP = MemHandleLock (dataH);
			if (!err)
			    size += ChunkSize;
		    }
		}

	    } while (len && !err);


	    if (!err && (dataLen > sizeof (RecordSeparator)))
	    {
		idx = dmMaxRecordIndex;
		RecH = DmNewRecord (DB, &idx, dataLen - sizeof (RecordSeparator));
		err = SetRecord (RecH, dataP + sizeof (RecordSeparator),
				 dataLen - sizeof (RecordSeparator), DB, idx, dmUnfiledCategory);

	    }
	    if (err)
		SndPlaySystemSound (sndError);		   // $$
	    MemHandleUnlock (dataH);
	    MemHandleFree (dataH);
	}						   /* if (!err) */
    }
    ExgDisconnect (sockp, err);				   // closes transfer dialog
    return (err);
}							   /* BeamReceive */

#endif


/**
   Initializes an object.
 Clears the header and set version number.

@param  op pointer on an object
@return nothing
*/

void InitObject (Object * op){
  MemSet(op,sizeof(Object),0);
  op->header.version = OBJECTVERSION;
}
