/** Waypoint edition
  @file editwpt.c
  @author C. Claveleira
  @date 1999-2004
  
*/

// $Id: editwpt.c,v 1.28 2004/11/17 21:18:23 clavelei Exp $ 


#ifndef SKIPEDIT

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Graffiti.h>
#include <string.h>
#include "gtalkRsc.h"
#include "latlong.h"
#include "MathLib.h"
#include <stdio.h>
#include "gpslib.h"
#include "gtalk.h"
#include "gpsutil.h"
#include "dbio.h"


static char LatButtonsValues[12][2];   /* contenu des bouton statiques de latitude */
static char LonButtonsValues[13][2];   /* contenu des bouton statiques de longitude */
static const char *ListChoices[] =
    { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "N", "S", "W", "E" };
static char LatMask[3][13] = {	       /* masque d'utilisation des boutons (attention  la cohrence) */
    "0060.000'N",		       /* DMM */
    "0060'60.0\"N",		       /* DMS */
    "00.00000N"		       /* DDD */
};
static char LonMask[3][14] = {
    "20060.000'W",		       /* DMM */
    "20060'60.0\"W",		       /* DMS */
    "200.00000W"		       /* DDD */
};
static const char * LocalId2String []={"Unkn", "D100","D101", "D102", "D103", "D104", "D105", "D106", "D107", "D108", "", "", "", "", "", "", "", "", "", "", "D109"};

static void EditListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText) EDIT_SECTION;
static void updateLatButtons (double latitude) EDIT_SECTION;
static void updateLonButtons (double longitude) EDIT_SECTION;
static void DegreToButtons (double latitude, double longitude) EDIT_SECTION;
static double ReadLon (void) EDIT_SECTION;
static double ReadLat (void) EDIT_SECTION;
static void SetButtons (int firstbutton, char *maskp, char *datas, char label[][2]) EDIT_SECTION;
static Err CustomToDBWpt (Custom_Wpt_Type * pcust, LocalFormatID format, long index, Boolean merge) EDIT_SECTION;

static Boolean changed = 0,
  ReducedSymbolsSet;


/**
   Get longitude from longitude buttons

@return longitude
 */

static double
ReadLon ()
{
    int butind,
      i;
    char lon[14];
    Boolean b;
    ControlPtr butptr;

    for (i = 0, butind = EditCoordButLon0; butind <= EditCoordButLonEW; butind++, i++)
    {
	butptr = (ControlPtr) GetObjPtr (butind);
	lon[i] = CtlGetLabel (butptr)[0];
    }
    lon[i] = 0;
    return (ReadCoord (lon, &b));
}


/**
   Get latitude from latitude buttons

@return latitude
 */

static double
ReadLat ()
{
    int butind,
      i;
    char lat[13];
    Boolean b;
    ControlPtr butptr;

    for (i = 0, butind = EditCoordButLat0; butind <= EditCoordButLatNS; butind++, i++)
    {
	butptr = (ControlPtr) GetObjPtr (butind);
	lat[i] = CtlGetLabel (butptr)[0];
    }
    lat[i] = 0;
    return (ReadCoord (lat, &b));
}


/**
   Sets buttons contents from given mask and datas 

   @param  firstbutton id of first button (left)
   @param maskp mask (format) to apply
   @param datas datas to display on buttons
   @param label string array where to put static labels
   @return nothing
 */

static void
SetButtons (int firstbutton, char *maskp, char *datas, char label[][2])
{
    int butind,
      j;
    ControlPtr butptr;
    ListPtr LstPtr;

    for (butind = firstbutton; *maskp; butind++, maskp++, datas++)
    {
	butptr = (ControlPtr) GetObjPtr (butind);
	LstPtr = (ListPtr) GetObjPtr (butind + EditListOffset);
	if (*maskp == 'N')
	{
	    LstSetListChoices (LstPtr, (char **)&ListChoices[10], 2);
	    LstSetSelection (LstPtr, *datas == 'N' ? 0 : 1);
	    LstSetHeight (LstPtr, 2);
	    CtlSetLabel (butptr, LstGetSelectionText (LstPtr, LstGetSelection (LstPtr)));
	    CtlSetEnabled (butptr, 1);
	}
	else if (*maskp == 'W')
	{
	    LstSetListChoices (LstPtr, (char **)&ListChoices[12], 2);
	    LstSetSelection (LstPtr, *datas == 'W' ? 0 : 1);
	    LstSetHeight (LstPtr, 2);
	    CtlSetLabel (butptr, LstGetSelectionText (LstPtr, LstGetSelection (LstPtr)));
	    CtlSetEnabled (butptr, 1);
	}
	else if ((*maskp >= '0') && (*maskp <= '9'))
	{
	    LstSetListChoices (LstPtr, (char **)&ListChoices[0], *maskp == '0' ? 10 : *maskp - '0');
	    LstSetHeight (LstPtr, *maskp == '0' ? 10 : *maskp - '0');
	    LstSetSelection (LstPtr, *datas - '0');
	    CtlSetLabel (butptr, LstGetSelectionText (LstPtr, LstGetSelection (LstPtr)));
	    CtlSetEnabled (butptr, 1);
	}
	else
	{
	    j = butind - firstbutton;
	    label[j][0] = *maskp;
	    label[j][1] = 0;
	    CtlSetLabel (butptr, label[j]);
	    CtlSetEnabled (butptr, 0);
	}
    }							   /* for */
}							   /* SetButtons */


/**
   Set content of latitude buttons from given latitude

@param  latitude latitude to display
@return nothing
 */

static void
updateLatButtons (double latitude)
{
    char *p,
     *maskp = LatMask[Prefs.displaymode];

    p = DegreToChars (latitude, 0, Prefs.displaymode);
    SetButtons (EditCoordButLat0, maskp, p, LatButtonsValues);
}


/**
  Set content of longitude buttons from given longitude

@param  longitude longitude to display
@return nothing
 */

static void
updateLonButtons (double longitude)
{
    char *p,
     *maskp = LonMask[Prefs.displaymode];

    p = DegreToChars (longitude, 1, Prefs.displaymode);
    SetButtons (EditCoordButLon0, maskp, p, LonButtonsValues);
}


/**
   Display latitude and longitude on latitude and longitude buttons

@param  latitude latitude to display
@param  longitude longitude to display
@return nothing
 */

static void
DegreToButtons (double latitude, double longitude)
{
    updateLatButtons (latitude);
    updateLonButtons (longitude);
    return;
}

/**
   Converts the given Custom_Wpt_Type waypoint to an object of given format 
   and place it in waypoint data base 

   @param  pcust pointeur on Custom_Wpt_Type waypoint type
   @param  format format to convert to
   @param  index If >= 0 : index of record to replace
                 If < 0  : new record to create
   @param  merge If != 0 the Custom_Wpt_Type waypoint is to be merged with the record of index index
   @return 0 if OK
 */

static Err
CustomToDBWpt (Custom_Wpt_Type * pcust, LocalFormatID format, long index, Boolean merge)
{
    Object o;
    ObjectPtr newp = &o,
      destob;
    VoidHand RecH = NULL;
    Err error;

    InitObject(newp);
    if (index >= 0)
    {							   /* replace existing record */
	RecH = DmGetRecord (WptDB, index);
	if (!RecH)
	    return (DmGetLastErr ());
	if (merge)
	{
	    destob = MemHandleLock (RecH);
	    MemMove (&(newp->d), &(destob->d),
		     min (sizeof (o.d), MemHandleSize (RecH) - sizeof (destob->header)));
	    MemHandleUnlock (RecH);
	}
    }

    CustomToAnyWpt (pcust, format, newp, merge);

    if (index < 0)
	error = put_wpt (&(newp->d.g), 0, 0, newp->header.type, newp->header.size);
    else
	error = SetRecord (RecH, newp, SizeOfObject (newp), WptDB, index, Prefs.wptcategory);	// 
    return (error);
}							   /* CustomToDBWpt */

#if 0
/**
   Callback proc. to display datum names in datum list

@param  itemNum item number to display
@param  bounds bounds to display in
@param  itemsText
@return nothing
 */

static void
EditListDatumDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)
{

	WinDrawChars (gDatum[itemNum].name, StrLen (gDatum[itemNum].name), bounds->topLeft.x,
		      bounds->topLeft.y);
}
#endif


/**
   Callback proc. to draw waypoints symbols in symbols list

@param  itemNum item number to draw
@param  bounds bounds to display in
@param  itemsText
@return nothing
 */

static void
EditListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)
{
    VoidHand RecH = NULL;
    BitmapPtr *RecP;

	if (ReducedSymbolsSet)
    {
	RecH = DmGetResource ('Tbmp', itemNum + DotBMID);
	if (RecH)
	{
	    RecP = MemHandleLock (RecH);
	    WinDrawBitmap ((BitmapPtr) RecP, bounds->topLeft.x, bounds->topLeft.y);
	    MemHandleUnlock (RecH);
	    DmReleaseResource (RecH);
	}
    }
    else
	DrawSymbol (itemNum, bounds->topLeft.x + 8, bounds->topLeft.y + 7);
}							   /* EditListDraw */



/**
   Edit form event handler

@param  e event
@return conventional handled
 */

Boolean
EditFormHandleEvent (EventPtr e)
{
    Boolean Edited,
      handled = false;
    FormPtr frm;
    ListPtr LstPtr;
    static ListPtr ColorLstPtr,
      SymbolLstPtr;
    FieldPtr fdPtr;
    ControlPtr PopupPtr,
      butPtr;
    static Custom_Wpt_Type wp;
    char date[dateStringLength];
    char time[timeStringLength];
    static LocalFormatID wptformat,
      curformat;
    static double latitude,
      longitude,
      latfrombut,		       /* buttons latitude */

      lonfrombut;		       /* buttons longitude */
    double nlatitude,
      nlongitude;
    static UInt MyCategory;
    static int butID;
    DateTimeType dtp;
    char name[8],
      CategoryName[dmCategoryLength],
     *cp;
    UInt i,
      j,
      category,
      attributes;
    Err error = 0;
    short choice;
    RectangleType r;
    int focus;

	switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	wptformat = Prefs.emulatedgps == G12CX ? G107Type :
	    Prefs.emulatedgps == GII ? G100Type :
	    Prefs.emulatedgps == MAP175 ? G102Type :
	    Prefs.emulatedgps == GIII ? G104Type :
	    Prefs.emulatedgps == eTrex ? G108Type : Prefs.emulatedgps ==
	    GPS76 ? G109Type : G103Type;
	if (EditMode >= NewWpt)
	{
	    /* creation nouveau waypoint  */
	    MemSet (&wp, sizeof (wp), 0);
	    wp.posn.lat = CurrentWGS84Pos.lat;
	    wp.posn.lon = CurrentWGS84Pos.lon;
	    wp.dst = 1e25;
	    wp.alt = 1e25;
	    //      if((wptformat != G103Type) && (wptformat != Garmin107Type))wp.smbl = sym_wpt_dot;
	    TimSecondsToDateTime (TimGetSeconds () - Prefs.utcoffset * 3600, &dtp);	/* cmnt par defaut a base de date */
	    DateToAscii (dtp.month, dtp.day, dtp.year, dfDMYWithDashes, date);
	    TimeToAscii (dtp.hour, dtp.minute, tfColon24h, time);
	    StrPrintF (wp.cmnt, "%s %s", date, time);
	    if (EditMode == MarkWpt)
	    {						   /* Mark point ? */
		for (i = 1; i < 100; i++)
		{
		    StrCopy (name, "MARK");
		    StrIToA (&name[4], i);
		    if (SearchWptByName (name, 1, 1) < 0)
			break;
		}
		StrNCopySafe (wp.ident, name, sizeof (wp.ident));
		changed = 1;
	    }
	    if (EditMode == FromRefWpt)
	    {						   /* nouveau wpt driv de rfrence ? */
		//              FrmCopyLabel (frm, EditRefNameID, ReferencePtName); /* nom de la reference */
		StrNToField (ReferencePtName, StrLen (ReferencePtName), EditRefNameID);	/* nom de la reference */
		FrmCopyLabel (frm, EditunitLabelID, DistanceUnitNames[Prefs.distunit]);
	    }
	    HideObject (EditBtRefID);			   /* effacement bouton Ref */
	}
	else
	  {						   /* edition waypoint existant */
	    curformat = IndexToCustomWpt (CurrentRecordIndex, WptDB, &wp);
	    if (wptformat != curformat)
	      if (!FrmAlert (ConvWptID))
		wptformat = curformat;
	    if (wptformat<=G109Type) StrNToField (LocalId2String[wptformat], StrLen (LocalId2String[wptformat]), EditTypeFieldID);
	    CurrentWGS84Pos.lat = wp.posn.lat;
	    CurrentWGS84Pos.lon = wp.posn.lon;
	    HideObject (EditBtNextID);
	  }

	/* symboles */
	SymbolLstPtr = (ListPtr) GetObjPtr (EditSymbolListID);
	LstSetDrawFunction (SymbolLstPtr, &EditListDraw);
	if ((wptformat == G103Type) || (wptformat == Garmin107Type))
	{
	    ReducedSymbolsSet = 1;
	    LstSetListChoices (SymbolLstPtr, NULL, 16);
	    wp.smbl = SymbolTypeTosmbl (wp.smbl);
	    //      LstSetSelection (SymbolLstPtr, wp.smbl);
	}
	else
	{
	    //        HideObject (EditSymbolPopupID);
	    ReducedSymbolsSet = 0;
	    LstSetListChoices (SymbolLstPtr, NULL, 2400 / 16);
	    wp.smbl = symb_gid2lid (wp.smbl);
	}
	LstSetSelection (SymbolLstPtr, wp.smbl);
	LstMakeItemVisible (SymbolLstPtr, LstGetSelection (SymbolLstPtr));

	/* nom */
	StrNToField (wp.ident, StrLen (wp.ident), EditFdNameID);
	if (EditMode >= NewWpt)
	{
	    fdPtr = (FieldPtr) GetObjPtr (EditFdNameID);
	    FldSetSelection (fdPtr, 0, FldGetTextLength (fdPtr));
	    FrmSetFocus (frm, FrmGetObjectIndex (frm, EditFdNameID));
	    /* FldGrabFocus(fdPtr); // ne marche pas */
	}

	/* commentaire */
	StrNToField (wp.cmnt, StrLen (wp.cmnt), EditFdCommentID);

	/* coordonnees */
	if (EditMode != FromRefWpt)
	{
	    SemiToDegre (&wp.posn, &latitude, &longitude); /* traduction */
	    nlatitude = latitude;
	    nlongitude = longitude;
	    translate (FromWGS84, &nlatitude, &nlongitude, Prefs.datum);	/* conversion */
	    DegreToButtons (nlatitude, nlongitude);	   /* et affichage */
	    latfrombut = ReadLat ();
	    lonfrombut = ReadLon ();
	}

	/* display mode */
	if ((wptformat == G103Type) || (wptformat == G104Type) || (wptformat == G107Type)
	    || (wptformat == G108Type) || (wptformat == G109Type))
	{						   // optimisable, a voir pour 109 $$
	    LstPtr = (ListPtr) GetObjPtr (EditDispListID);
	    LstSetSelection (LstPtr, wp.dspl);
	    PopupPtr = (ControlPtr) GetObjPtr (EditDispPopupID);
	    CtlSetLabel (PopupPtr, LstGetSelectionText (LstPtr, wp.dspl));
	}
	else
	{
	    HideObject (EditPDispLabelID);
	    HideObject (EditDispPopupID);
	}
	//#if 0
	/* proximity distance */
	if ((wptformat == G102Type) || (wptformat == G104Type) || (wptformat == G107Type) || 
	    (wptformat == G108Type) || (wptformat == G109Type))
	  if (wp.dst < 1.0e9) {
	    cp = DToA (wp.dst, 0, 0);  // suspect principal $$
	    StrNToField (cp, StrLen (cp), EditPDistFieldID);
	  }
	  else;
	else
	{
	    //FldSetUsable (fdPtr, 0);
	    HideObject (EditPDistFieldID);
	    HideObject (EditPDistLabelID);
	}
	/* altitude */
	if ((wptformat == G108Type) || (wptformat == G109Type))
	  {
	    if (wp.alt < 1.0e9) {
	      cp = DToA (wp.alt, 0, 0);
	      StrNToField (cp, StrLen (cp), EditAltFieldID);
	    }
	    ShowObject (EditAltLabelID);
	    ShowObject (EditAltFieldID);
	  }

	/* color */
	ColorLstPtr = NULL;
	PopupPtr = NULL;
	if (wptformat == G107Type)
	{
	    ColorLstPtr = (ListPtr) GetObjPtr (EditColorListID);
	    LstSetSelection (ColorLstPtr, wp.color < 4 ? wp.color : 0);
	    PopupPtr = (ControlPtr) GetObjPtr (EditColorPopupID);
	    ShowObject (EditColorPopupID);
	}
	else if ((wptformat == G108Type) || (wptformat == G109Type))
	{
	    ColorLstPtr = (ListPtr) GetObjPtr (EditColor2ListID);
	    LstSetSelection (ColorLstPtr, wp.color < 16 ? wp.color : 0);
	    PopupPtr = (ControlPtr) GetObjPtr (EditColor2PopupID);
	    ShowObject (EditColor2PopupID);
	}
	if (ColorLstPtr)
	    CtlSetLabel (PopupPtr,
			 LstGetSelectionText (ColorLstPtr, LstGetSelection (ColorLstPtr)));

	/* catgorie */
	if (EditMode >= NewWpt)
	{
	    MyCategory = Prefs.wptcategory;
	    if (MyCategory == dmAllCategories)
		MyCategory = dmUnfiledCategory;
	}
	else
	{
	    error = DmRecordInfo (WptDB, CurrentRecordIndex, &attributes, NULL, NULL);
	    MyCategory = attributes & dmRecAttrCategoryMask;
	}
	CategoryGetName (WptDB, MyCategory, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (EditCatPopupID), CategoryName);

	/* datum */
#if 0
	LstPtr = (ListPtr) GetObjPtr (EditDatumListID);
	LstSetListChoices (LstPtr, NULL, nDatums);
	LstSetDrawFunction (LstPtr, &EditListDatumDraw);
	LstSetSelection (LstPtr, Prefs.datum);
	LstMakeItemVisible (LstPtr, Prefs.datum);
	PopupPtr = (ControlPtr) GetObjPtr (EditDatumPopupID);
	CtlSetLabel (PopupPtr, gDatum[Prefs.datum].name);
#else
	FldSetTextPtr (GetObjPtr (EditDatumFieldID), gDatum[Prefs.datum].name);
#endif

	FrmDrawForm (frm);

	//      PopupPtr = (ControlPtr) GetObjPtr (EditSymbolPopupID);
	GetObjectBounds (EditSymbolPopupID, &r);
	EditListDraw (LstGetSelection (SymbolLstPtr), &r, NULL);
	GrfSetState (1, 0, 0);
	//      butID = EditCoordButLat0; ??
	handled = true;
	break;

    case frmCloseEvent:
	if (EditMode >= NewWpt)
	    //        SelectionAvailable = 0;
	    break;

/*      case menuEvent: */
/*  	MenuEraseStatus (NULL); */
/*  	switch (e->data.menu.itemID) */
/*  	{ */
/*  	} */
/*  	handled = true; */
/*  	break; */

    case lstSelectEvent:
	switch (e->data.lstSelect.listID)
	{
	}
	//      handled = true;
	break;

    case popSelectEvent:
#if 0
	if (e->data.popSelect.controlID == EditDatumPopupID)
	{
	    CtlSetLabel (e->data.popSelect.controlP, gDatum[e->data.popSelect.selection].name);
	    nlongitude = ReadLon ();
	    nlatitude = ReadLat ();
	    if ((nlongitude == lonfrombut) && (nlatitude == latfrombut))	/* contenu boutons modifie ? */
	    {
		nlatitude = latitude;
		nlongitude = longitude;
	    }
	    else
		translate (ToWGS84, &nlatitude, &nlongitude, Prefs.datum);	/* conversion depuis datum courant */
	    Prefs.datum = e->data.popSelect.selection;
	    translate (FromWGS84, &nlatitude, &nlongitude, Prefs.datum);	/* conversion vers nouveau datum */
	    DegreToButtons (nlatitude, nlongitude);
	    latfrombut = ReadLat ();
	    lonfrombut = ReadLon ();			   /* et affichage */
	    handled = true;
	}
#endif
	break;

    case keyDownEvent:
	frm = FrmGetActiveForm ();
	focus = FrmGetFocus (frm);
	if (focus >= 0)
	    i = FrmGetObjectId (frm, focus);
	else
	    i = 0;
	switch (e->data.keyDown.chr)
	{
	case nextFieldChr:
	    switch (i)
	    {
	    case EditFdNameID:
		j = EditFdCommentID;
		break;
	    default:
		j = EditFdNameID;
	    }						   /* switch (i) */
	    FrmSetFocus (frm, FrmGetObjectIndex (frm, j));
	    handled = true;
	    break;

	case prevFieldChr:
	    break;
	}						   /* switch (e->data.keyDown.chr) */
	break;						   /* pb avec focus des fields */
#if 0
	butPtr = (ControlPtr) GetObjPtr (butID);
	if (butPtr)
	{
	    p = CtlGetLabel (butPtr);
	    name[0] = *p;
	    *p = e->data.keyDown.chr;
	    CtlSetLabel (butPtr, p);
	    if ((LonFieldToDegree () < -180) || (LonFieldToDegree () > 180)
		|| (LatFieldToDegree () < -90) || (LatFieldToDegree () > 90))
	    {
		SndPlaySystemSound (sndError);
		*p = name[0];
		CtlSetLabel (butPtr, p);
	    }
	}
	handled = true;
	break;
#endif
    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case EditBtNextID:
	case EditBtOKID:
	    //                frm = FrmGetActiveForm ();
	    LstPtr = (ListPtr) GetObjPtr (EditDispListID);
	    if (LstGetSelection (LstPtr) != wp.dspl)	   /* display mode chang ? */
	    {
		wp.dspl = LstGetSelection (LstPtr);
		changed = 1;
	    }

	    fdPtr = (FieldPtr) GetObjPtr (EditFdNameID);
	    if (FldDirty (fdPtr))
	    {						   /* nom chang ? */
		StrNCopySafe (wp.ident, FldGetTextPtr (fdPtr), sizeof (wp.ident));
		Canonicalize (wp.ident, /*  sizeof */ StrLen (wp.ident));
		changed = 1;
	    }

	    fdPtr = (FieldPtr) GetObjPtr (EditFdCommentID);
	    if (FldDirty (fdPtr))
	    {						   /* commentaire chang ? */
		StrNCopySafe (wp.cmnt, FldGetTextPtr (fdPtr), sizeof (wp.cmnt));
		Canonicalize (wp.cmnt, /*  sizeof */ StrLen (wp.cmnt));
		changed = 1;
	    }

	    fdPtr = (FieldPtr) GetObjPtr (EditPDistFieldID);
	    if (FldDirty (fdPtr))
	    {						   /* distance change ? */
		wp.dst = StrToD (FldGetTextPtr (fdPtr), NULL);
		if(wp.dst < 0.0)wp.dst=-wp.dst;
		changed = 1;
	    }

	    fdPtr = (FieldPtr) GetObjPtr (EditAltFieldID);
	    if (FldDirty (fdPtr))
	    {						   /* altitude change ? */
		wp.alt = StrToD (FldGetTextPtr (fdPtr), NULL);
		changed = 1;
	    }

	    //      LstPtr = (ListPtr) GetObjPtr (EditColorListID);
	    if (ColorLstPtr && (LstGetSelection (ColorLstPtr) != wp.color))	/* couleur change ? */
	    {
		wp.color = LstGetSelection (ColorLstPtr);
		changed = 1;
	    }
	    if (ReducedSymbolsSet)
		wp.smbl = smblToSymbolType (wp.smbl);	   /* symbole */
	    else
		wp.smbl = symb_lid2gid (wp.smbl);
	    if (EditMode == FromRefWpt)
	    {						   /* cas du wpt driv d'une rfrence */
		double distance = 0;
		unsigned int bearing = 0;
		fdPtr = (FieldPtr) GetObjPtr (EditDistFieldID);

		if (FldDirty (fdPtr))
		    distance = StrToD (FldGetTextPtr (fdPtr), NULL);

		if (Prefs.distunit == metric)
		    distance /= 1.852;
		else if (Prefs.distunit == statute)
		    distance /= 1.15078;
		fdPtr = (FieldPtr) GetObjPtr (EditBearFieldID);
		if (FldDirty (fdPtr))
		    bearing = StrAToI (FldGetTextPtr (fdPtr));
		wp.posn.lat = ReferencePt.lat;
		wp.posn.lon = ReferencePt.lon;
		DistBearToSemi (distance, bearing, &wp.posn);
		changed = 1;
	    }
	    else
	    {
		nlongitude = ReadLon ();
		nlatitude = ReadLat ();
		if ((nlatitude > 90.) || (nlatitude < -90.) || (nlongitude > 180.)
		    || (nlongitude < -180.))
		{
		    FrmCustomAlert (CoordErrorID, "", "", "");
		    changed = 0;
		    handled = true;
		    break;
		}
		if ((nlongitude != lonfrombut) || (nlatitude != latfrombut))
		{					   /* coord. modifies ? */
		    translate (ToWGS84, &nlatitude, &nlongitude, Prefs.datum);
		    DegreToSemi (nlatitude, nlongitude, &wp.posn);
		    changed = 1;
		}
	    }
	    if (changed)
	    {
		WptSortCritType oldsortcrit = Prefs.wptsortcrit;
		Prefs.wptcategory = MyCategory;
		if (EditMode >= NewWpt)
		{					   /* creation nouveau waypoint ? */
		    error = CustomToDBWpt (&wp, wptformat, -1, 0);
		    CurrentRecordIndex = 0;
		}
		else
		{					   /* modification de waypoint */
		    error = CustomToDBWpt (&wp, wptformat, CurrentRecordIndex, 1);
		}
		SortWpts (oldsortcrit);			   /* remise en ordre */
		MainListDirty = 1;
	    }						   /* EditBtOKID */
	    if (e->data.ctlSelect.controlID == EditBtNextID)
	    {						   /* on enchaine sur un autre */
		wp.ident[0] = 0;
		StrNToField (wp.ident, StrLen (wp.ident), EditFdNameID);
		break;
	    }

	case EditBtCancelID:
	    FrmReturnToForm (0);
	    handled = true;
	    break;

	case EditBtRefID:
	    if (SetReferencePt (CurrentRecordIndex, 1))
	    {
		if ((Prefs.wptsortcrit == Dist)||(Prefs.wptsortcrit == Bear))
		    SortWpts (Current);
		FrmReturnToForm (0);
	    }
	    handled = true;
	    break;

	case EditBtDeleteID:				   /* suppression du waypoint */
	    if (EditMode >= NewWpt)
	    {						   /* si nouveau wpt abandon */
		//                SelectionAvailable = 0;
		FrmReturnToForm (0);
	    }
	    else if (FrmCustomAlert (RmvWptID, wp.ident, NULL, NULL))
	    {
		error = DmRemoveRecord (WptDB, CurrentRecordIndex);
		CurrentRecordIndex = CurrentRecordIndex == 0 ? 0 : CurrentRecordIndex - 1;
		MainListDirty = 1;
		FrmReturnToForm (0);
	    }

	    handled = true;
	    break;

	case EditCatPopupID:				   /* categorie */
	    frm = FrmGetActiveForm ();
	    category = MyCategory;
	    Edited =
		CategorySelect (WptDB, frm, EditCatPopupID, EditCatListID, false, &MyCategory,
				CtlGetLabel (e->data.ctlSelect.pControl), 1, 0);
	    if (Edited)
		Prefs.wptcategory = dmAllCategories;
	    if ((MyCategory != category) && (EditMode == EditWpt))
		changed = 1;
	    handled = true;
	    break;

	case EditSymbolPopupID:			   /* symbole */
	    //        LstPtr = (ListPtr) GetObjPtr (EditSymbolListID);
	    choice = LstPopupList (SymbolLstPtr);
	    if ((choice != -1) && (choice != wp.smbl))
	    {
		changed = 1;
		wp.smbl = choice;
	    }
	    //      PopupPtr = (ControlPtr) GetObjPtr (EditSymbolPopupID);
	    GetObjectBounds (EditSymbolPopupID, &r);
	    EditListDraw (choice, &r, NULL);
	    handled = true;
	    break;

	case EditCoordButLat0:
	case EditCoordButLat0 + 1:
	case EditCoordButLat0 + 2:
	case EditCoordButLat0 + 3:
	case EditCoordButLat0 + 4:
	case EditCoordButLat0 + 5:
	case EditCoordButLat0 + 6:
	case EditCoordButLat0 + 7:
	case EditCoordButLat0 + 8:
	case EditCoordButLat0 + 9:
	case EditCoordButLat0 + 10:
	case EditCoordButLat0 + 11:
	case EditCoordButLon0:
	case EditCoordButLon0 + 1:
	case EditCoordButLon0 + 2:
	case EditCoordButLon0 + 3:
	case EditCoordButLon0 + 4:
	case EditCoordButLon0 + 5:
	case EditCoordButLon0 + 6:
	case EditCoordButLon0 + 7:
	case EditCoordButLon0 + 8:
	case EditCoordButLon0 + 9:
	case EditCoordButLon0 + 10:
	case EditCoordButLon0 + 11:
	case EditCoordButLon0 + 12:
	    butID = e->data.ctlSelect.controlID;
	    butPtr = e->data.ctlSelect.pControl;
	    LstPtr = (ListPtr) GetObjPtr (butID + EditListOffset);
	    i = LstPopupList (LstPtr);
	    CtlSetLabel (butPtr, LstGetSelectionText (LstPtr, LstGetSelection (LstPtr)));
	    break;
	}
	break;

    default:
	break;
    }

	return handled;
}							   /* EditFormHandleEvent */

#endif
