/** GPS interactions forms for GPilotS 
  @file
  @author C. Claveleira
  @date 1999-2004
  
*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)
*/

/* 	$Id: gpsio.c,v 1.8 2004/11/17 21:17:25 clavelei Exp $	 */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "gtalkRsc.h"
#include "MathLib.h"
#include "gpslib.h"
#include "gtalk.h"
#include "gpsutil.h"
#include "dbio.h"
#include "gpsio.h"
#include "memo.h"
#include "log.h"

GpsOperation GpsOp;
static char StatHost[32];
static char CategoryName[dmCategoryLength];
static UInt ItemsToSendInCategory (DmOpenRef DB) IO_SECTION;
static Boolean OkForGremlins (unsigned int prob) IO_SECTION;


/**
   Event handler of Host form

  @param  e pointer on event received
  @return true if handled
 */

Boolean
HostFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    static char WptCategoryName[dmCategoryLength],
      TrkCategoryName[dmCategoryLength],
      RteCategoryName[dmCategoryLength];

	switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	CategoryGetName (WptDB, Prefs.wptcategory, WptCategoryName);
	CategorySetTriggerLabel (GetObjPtr (HostWptPopupID), WptCategoryName);
	CategoryGetName (TrkDB, Prefs.trkcategory, TrkCategoryName);
	CategorySetTriggerLabel (GetObjPtr (HostTrkPopupID), TrkCategoryName);
	CategoryGetName (RteDB, Prefs.rtecategory, RteCategoryName);
	CategorySetTriggerLabel (GetObjPtr (HostRtePopupID), RteCategoryName);
	FrmDrawForm (frm);
	handled = true;
	break;

    case frmCloseEvent:
	break;


    case ctlSelectEvent:
	frm = FrmGetActiveForm ();
	switch (e->data.ctlSelect.controlID)
	{
	case HostWptPopupID:
	    CategorySelect (WptDB, frm, HostWptPopupID, HostWptListID, true,
			    (WordPtr) & Prefs.wptcategory, WptCategoryName, 1, 0);
	    break;
	case HostRtePopupID:
	    CategorySelect (RteDB, frm, HostRtePopupID, HostRteListID, true,
			    (WordPtr) & Prefs.rtecategory, RteCategoryName, 1, 0);
	    break;
	case HostTrkPopupID:
	    CategorySelect (TrkDB, frm, HostTrkPopupID, HostTrkListID, true,
			    (WordPtr) & Prefs.trkcategory, TrkCategoryName, 1, 0);
	    break;
	case BtCancelID:
	    FrmReturnToForm (0);
	    break;
	case BtGoID:
	    CompactTracks = GetControlValue (HostCmptID);
	    //CtlGetValue (GetObjPtr (HostCmptID));
	    GpsOp = gpsHostMode;
	    FrmGotoForm (ProgressForm);
	    break;
	}
	handled = true;
	break;

    default:
	break;
    }

	return handled;
}							   // HostFormHandleEvent


/**
   Event handler of progress form

  @param  e pointer on event received
  @return true if handled
*/

Boolean
ProgressFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    //    FormPtr frm;
    static FieldPtr fdPtr;
    ULong GPSTime;
    static ULong start,
      t;
    Long TimeDiff;
    char buf[16];
    static char *p,
     *statp;
    EventType evt;
    Radian_Type *PositionP;
    static FieldPtr StatHostFdPtr;
    static Boolean IgnoreEooEvt;       /* flag pour ignorer EndOfOperationEvent */
    Err error=0;

	if (e->eType!=nilEvent) debug_in (GTALKDBG_CALLS, ("->ProgressFormHandleEvent(%s)", 
e->eType > LASTVALIDEVTNAME ?StrIToA(buf,e->eType):EvtNames[e->eType]));
	switch (e->eType)
    {
    case nilEvent:
	GpslibTicks () /* traiter future erreur renvoyee */ ;
	if (statp && (TimGetTicks () > t + sysTicksPerSecond / 3))
	{						   /* gadget */
	    //buf[0]=0;
	    StrNCopy (buf, p >= (statp + StrLen (statp)) ? p = statp : p, sizeof (buf));
	    p++;
	    StrNCat (buf, statp, sizeof (buf));
	    //      buf[sizeof(buf)-1]=0;
	    FldSetTextPtr (StatHostFdPtr, buf);
	    FldRecalculateField (StatHostFdPtr, true);
	    FldDrawField (StatHostFdPtr);
	    t = TimGetTicks ();
	}
	handled = true;
	break;

    case frmOpenEvent:
	//        frm = FrmGetActiveForm ();
	StatHostFdPtr = (FieldPtr) (GetObjPtr (FdStatus2ID));
	p = statp = NULL;
	t = TimGetTicks ();
	New = Replaced = Sent = 0;
	fdPtr = (FieldPtr) (GetObjPtr (FdStatusID));
	FrmDrawForm (FrmGetActiveForm ());
#ifdef LOOSE
	if (Prefs.looseprotocol) GpslibControl(GpslibCtrl_LooseProto,&Prefs.looseprotocol);
#endif
	switch (GpsOp)
	{
	case gpsHostMode:
	case gpsEnableAsyncEvts:
	    SysCopyStringResource (StatHost, StrHostRunning);
	    statp = p = StatHost;
	    error=GarminSetHostMode (fdPtr, Prefs.emulatedgps);
	    if (!error) if (GpsOp == gpsHostMode)
	    {
		GarminGetWpts (put_wpt, fdPtr);
		GarminGetTrks (put_trk, fdPtr);
		GarminGetRtes (put_rte, fdPtr);
	    }
#ifdef EVENTS
	    else
		error=GarminEnableAsyncEvents (&put_packet, fdPtr, /*0xffff */ 0x7201);
#endif
	    break;

	case gpsUpload:
#ifdef FURUNO
	  if (Furuno) {
		if (Prefs.todisplay == DispWpt)
		    error=FurunoPutWpts (&Furuno_get_next_wpt, fdPtr,
				   DmNumRecordsInCategory (WptDB, Prefs.wptcategory));  //$$
		else if (Prefs.todisplay == DispRte)
		    error=FurunoPutRtes (&get_next_full_rte, fdPtr,
				   DmNumRecordsInCategory (RteDB, Prefs.rtecategory));//$$
	  } else
#endif
	    if (Prefs.todisplay == DispWpt)
	      error=GarminPutWpts (&get_next_wpt, fdPtr,ItemsToSendInCategory(WptDB));
	    else if (Prefs.todisplay == DispTrk)
		error=GarminPutTrks (&get_next_trk, fdPtr, ItemsToSendInCategory (TrkDB));
	    else if (Prefs.todisplay == DispRte)
		error=GarminPutRtes (&get_next_rte, fdPtr, ItemsToSendInCategory (RteDB));
	    break;

	case gpsDownload:
	    if (Prefs.todisplay == DispWpt)
		error=GarminGetWpts (put_wpt, fdPtr);
	    else if (Prefs.todisplay == DispTrk)
		error=GarminGetTrks (put_trk, fdPtr);
	    else if (Prefs.todisplay == DispRte)
		error=GarminGetRtes (put_rte, fdPtr);
	    break;

	case gpsGetTime:
	    error=GarminGetTime (fdPtr);			   /* requete time */
	    Prefs.todisplay -= 4;
	    break;

	case gpsGetPosition:
	    error=GarminGetPosition (fdPtr);			   /* requete de position */
	    Prefs.todisplay -= 8;
	    break;

	case gpsGetInfos:
	  debug (GTALKDBG_INFOS, ("ProgressFormHandleEvent: before GarminGetRemoteInfos"));  // $$
	    error=GarminGetRemoteInfos (fdPtr);		   /* requete d'identification */
	    debug (GTALKDBG_INFOS, ("ProgressFormHandleEvent: after GarminGetRemoteInfos"));  // $$
	    Prefs.todisplay -= 32;
	default:
	}
	if (error) {
	  HideObject(BtStopID);
	  ShowObject(BtOKID);
	}
	else {
	start = TimGetSeconds ();
	MustCallGpslibTicks = 1;
	IgnoreEooEvt = 0;
	}
	handled = true;
	break;

    case frmCloseEvent:
	GPSEndOfOperation ();
	MustCallGpslibTicks = 0;
	CleanUpTracks ();
/*  	FldSetSelection (fdPtr, 0, FldGetTextLength (fdPtr)); */
/*  	FldCopy (fdPtr);  */
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case BtStopID:
	    if (OkForGremlins(20))GPSEndOfOperation ();
	    handled = true;
	    break;
	case BtOKID:
	    FrmReturnToForm (0);
	    handled = true;
	    break;
	}
	break;

    case TransferWptReqEvent:				   /* requete de transfert de wpt */
	GarminPutWpts (&get_next_wpt, fdPtr, DmNumRecordsInCategory (WptDB, Prefs.wptcategory));
	handled = true;
	break;

    case TransferRteReqEvent:				   /* requete de transfert de routes */
	GarminPutRtes (&get_next_rte, fdPtr, ItemsToSendInCategory (RteDB));
	handled = true;
	break;

    case TransferTrkReqEvent:				   /* requete de transfert de traces */
	GarminPutTrks (&get_next_trk, fdPtr, ItemsToSendInCategory (TrkDB));
	break;

#ifdef EVENTS
    case AsyncEventReqEvent:				   /* requete d'envoi d'events */
      {
	AsyncEventMask * evtmaskp=(AsyncEventMask *)& (e->data);
	if(*evtmaskp)GarminPutPacket (&get_next_packet, fdPtr, DmNumRecords (LogDB));
      }
	break;
#endif

    case TimeReceivedEvent:				   /* time recu */
	GPSTime = e->data.sclRepeat.time;
	TimeDiff = TimGetSeconds () - (GPSTime + (long) Prefs.utcoffset * 3600);
	SysCopyStringResource (gbuf1, StrTimeFormat);
	StrPrintF (gbuf2, gbuf1, TimeDiff);
	if (FrmCustomAlert (ConfID, " ", gbuf2, " "))
	    TimSetSeconds (TimGetSeconds () - TimeDiff);
	IgnoreEooEvt = 1;
	FrmReturnToForm (0);
	handled = true;
	break;

    case PositionReceivedEvent:			   /* position recue */
	PositionP = (Radian_Type *) & (e->data);
	debug (GTALKDBG_INFOS, ("ProgressFormHandleEvent(), got position : Lat=%ld", PositionP->lat));
	CurrentWGS84Pos.lat = PositionP->lat * OnePower31Pi;
	CurrentWGS84Pos.lon = PositionP->lon * OnePower31Pi;
	//  CurrentRecordIndex = dmMaxRecordIndex;
	debug (GTALKDBG_INFOS, ("ProgressFormHandleEvent(), CurrentWGS84Pos.lat=%ld",CurrentWGS84Pos.lat ));
	EditMode = MarkWpt;
	IgnoreEooEvt = 1;
#ifndef SKIPEDIT
	FrmGotoForm (EditForm);
#endif
	handled = true;
	break;

    case PowerOffEvent:				   /* power off */
	evt.eType = keyDownEvent;
	evt.data.keyDown.chr = hardPowerChr;
	evt.data.keyDown.keyCode = 0;
	evt.data.keyDown.modifiers = commandKeyMask;
	EvtAddEventToQueue (&evt);
	IgnoreEooEvt = 1;
	FrmReturnToForm (0);
	handled = true;
	break;

    case EndOfOperationEvent:
	if (!IgnoreEooEvt)
	  {
	    MustCallGpslibTicks = 0;
	    CleanUpTracks ();
	    if (GpsOp<=gpsDownload){
	      SysCopyStringResource (gbuf1, StrProgFormat);
	      StrPrintF (gbuf2, gbuf1, TimGetSeconds () - start, New, Replaced, Sent);
	      if (Verbose)
		FrmCustomAlert (AlertID, gbuf2, " ", " ");
	      FrmReturnToForm (0);
	    }
	    else {
	      HideObject(BtStopID);
	      ShowObject(BtOKID);
	      FldSetSelection (fdPtr, 0, FldGetTextLength (fdPtr));
	      FldCopy (fdPtr);
	    }
	  }
	MainListDirty = 1;
	handled = true;
	break;

    default:
	break;
    }							   // switch (e->eType)

    if (e->eType!=nilEvent) debug_out (GTALKDBG_CALLS, ("<-ProgressFormHandleEvent()=%d", handled));
      return handled;
}							   // ProgressFormHandleEvent


/**
   Event handler of download form

  @param  e pointer on event received
  @return true if handled
*/

Boolean
DownloadFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;

    frm = FrmGetActiveForm ();
    switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	if (*CurrentCategory == dmAllCategories)
	    *CurrentCategory = 0;
	CategoryGetName (CurrentDB, *CurrentCategory, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (MainCatPopupID), CategoryName);
	//      CompactTracks=0;
	if (Prefs.todisplay == DispTrk)
	    ShowObject (DownCkID);
	//      Upload = 0;
	FrmDrawForm (frm);
	handled = true;
	break;

    case frmCloseEvent:
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case BtGoID:
	    CompactTracks = GetControlValue (DownCkID);
	    //CtlGetValue (GetObjPtr (DownCkID));
	    GpsOp = gpsDownload;
	    FrmGotoForm (ProgressForm);
	    break;
	case BtCancelID:
	    FrmReturnToForm (0);
	    break;
	case MainCatPopupID:
	    CategorySelect (CurrentDB, frm, MainCatPopupID, MainCatListID, false, CurrentCategory,
			    CategoryName, 1, 0);
	    break;
	}
	handled = true;
	break;

    default:
	break;
    }

    return handled;
}							   //DownloadFormHandleEvent


/**
   Upload form event handler

  @param  e pointer on event received
  @return true if event handled
*/

Boolean
UploadFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;

    frm = FrmGetActiveForm ();
    switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	CategoryGetName (CurrentDB, *CurrentCategory, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (MainCatPopupID), CategoryName);
	//      Upload = 1;
	Furuno = 0;
	FrmDrawForm (frm);
	handled = true;
	break;

    case frmCloseEvent:
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case BtGoID:
	    GpsOp = gpsUpload;
	    FrmGotoForm (ProgressForm);
	    break;
	case BtCancelID:
	    FrmReturnToForm (0);
	    break;
	case MainCatPopupID:
	    CategorySelect (CurrentDB, frm, MainCatPopupID, MainCatListID, true, CurrentCategory,
			    CategoryName, 1, 0);
	    break;
	}
	handled = true;
	break;

#ifdef FURUNO
    case keyDownEvent:					   /* activation mode furuno */
	if (e->data.keyDown.chr == pageDownChr)
	  {
	    WinDrawChars ("FURUNO", 6, 20, 80);
	    Furuno = 1;
	  }
	break;
#endif

    default:
	break;
    }

    return handled;
}							   // UploadFormHandleEvent


/** Number of items to send.
   Return total number of items, in current category of given data base, 
   to send given the transmission protocol and selection mode

 @param  DB pointer on data base
 @return number of items
*/

static UInt
ItemsToSendInCategory (DmOpenRef DB)
{
  UInt n,
    c,
    t,
    cat;
  VoidHand RecH = NULL;
  GarminProtoID Proto;

  c = n = 0;
  if (DB == WptDB) {	/* wpt : simple, un item = un record */
#ifdef SELECT
    if (Prefs.selectmode){
      while ((RecH = DmQueryNextInCategory (DB, &c, Prefs.wptcategory)))
	if(RecIsSecret(DB,c++))n++;
    }
    else    
#endif
      n= DmNumRecordsInCategory (WptDB, Prefs.wptcategory);
    return(n);
  }

  else if (DB == TrkDB){
    cat = Prefs.trkcategory;
    Proto=GarminGetProtocol(TrkProto);
    if (Proto==A000)Proto=A300; // valeur par defaut provisoire $$
  }
  else {
    cat = Prefs.rtecategory;
    Proto=GarminGetProtocol(RteProto);
    if (Proto==A000)Proto=A200; // valeur par defaut provisoire $$
  }
  if (Proto==A000)return(0);  /* erreur */

  while ((RecH = DmQueryNextInCategory (DB, &c, cat)))
    {
      if (!Prefs.selectmode
#ifdef SELECT
	  || RecIsSecret(DB, c)
#endif
	  ){
	if (DB == TrkDB){
	  n += TrkptsInTrkDB (RecH)+ (Proto==A301?1:0) ;
	}
	else {
	  t=WptsInRteDB (RecH);
	  n += t + 1;
	  if (Proto==A201) n+= --t;  /* links */
	}
      }
      c++;
    }
  return (n);
}							   /* ItemsToSendInCategory */


static Boolean OkForGremlins (unsigned int prob){
  if(SysGremlins (GremlinIsOn, NULL))
    return (SysRandom(0)<=(sysRandomMax/prob)); /* if gremlins, return probability */
  else return(1);
}
