/**  Main code for GPilotS.
  @file gtalk.c
  @author C. Claveleira
  @date 1999-2004
 
  Contains entry point of GPilotS and management code of the main form, eg
  to display list of waypoints, routes or tracks. It's named gtalk.c for
  historical reasons...

*/

/*
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef DEBUG
static char rcsid[] = "$Id: gtalk.c,v 1.41 2004/11/17 21:11:32 clavelei Exp $";
static char * GetCharEncoding(void);
#endif

#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS // very dirty, en attendant que pilrc permette hascrollbar pour les listes $$

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <string.h>
#include <stdio.h>
#include "gtalkRsc.h"
#include "latlong.h"
#include "MathLib.h"
#include "gpslib.h"
#include "gtalk.h"
#include "route.h"
#include "gpsutil.h"
#include "log.h"
#include "map.h"
#include "gpsio.h"
#include "memo.h"
#include "dbio.h"
#include "track.h"
#include <LstGlue.h>
#include <TxtGlue.h>

// The minimum OS version supported :
#define ourMinVersion		sysMakeROMVersion(3,0,0,sysROMStageRelease,0)
#define MaxDBNames  10 /* maximum data bases names managed by SetPreferences */

PrefType Prefs =
    { DispWpt, dmAllCategories, 0, Alpha, metric, DMM, 0, WGS84Index, 0, DBVersion, GIIPlus, 10,
      dmAllCategories, dmAllCategories, dmAllCategories, false, 0, WptDBName, RteDBName, TrkDBName, 0
};
int DebugLevel;
const char *DistanceUnitNames[] = { "km", "nm", "mi" };
char *BeamTypes[] = { "Un", "Wp", "Rt", "Tk" };	/* see ItemToDisplay enum for order */
EditWptMode EditMode;
Semicircle_Type ReferencePt = { 0, 0 };	/* coordonnees point de reference pour tri wpts */
char ReferencePtName[32] = "";


UInt CurrentRecordIndex,	       /* current record index */
  CurrentRouteIndex,
 *CurrentCategory;

UInt New,
  Replaced,
  Sent;
char gbuf1[512],
  gbuf2[128],			       /* buffers "general purpose" */

  search_buf[6];		       /* buffer recherche par debut de nom */


Semicircle_Type CurrentWGS84Pos;
Boolean MustCallGpslibTicks = 0;
Boolean OnlyNews,
  Furuno,
  CompactTracks = 0;
Boolean MainListDirty = 1;
DmOpenRef CurrentDB,
  WptDB,
  TrkDB,
  RteDB,
  LogDB;

static UInt lastrecordindex,
  lastitemnum = -2;
static ULong CurrentRecordID;	       /* current record id */
static char CategoryName[dmCategoryLength];
static WinHandle SymbolsWH;	       /* offscreen window containing symbols bitmaps */

// prototypes :
static Boolean MainFormHandleEvent (EventPtr e);
static Boolean ApplicationHandleEvent (EventPtr e);
static Word StartApplication (void);
static void StopApplication (void);
static void EventLoop (void);
DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags);
static Err OpenOrCreateDB (DmOpenRef * DBP, char *DBName, ULong DBType, ULong creator);
static void UpdateMainForm (void);
static void AlarmProc (void);
static Err RomVersionCompatible (UInt32 requiredVersion, UInt16 launchFlags);
#ifdef SELECT
static Boolean CatFormHandleEvent (EventPtr e);
static Err RecSetSecret (DmOpenRef * DBP, UInt16 index, Boolean value);
#endif
static void NewDB (void);
static UInt16 FindDBNames(UInt32 type, UInt32 creator,  char * namep,Int16 * matched, char * bufp, char ** bufpp,UInt16 bufpsize);
static Err SwitchDB(DmOpenRef * DBP, char *DBName, ULong DBType);
static void ChangeDB (void);
#ifdef DEBUG
static void SetDebug (void);
#endif

#ifdef EXPORT
static void ExportWpt (void);


/**
   Experimental procedure to export waypoints as text.

@return nothing
*/

static void
ExportWpt ()			       /* experimental, not tested */
{
    UInt i,
      l,
      pos = 0;
    Custom_Wpt_Type wpt;
    FormPtr fp = NULL;
    double latitude,
      longitude;
    char *p;

    fp = FrmInitForm (ExportForm);
    FrmDrawForm (fp);

    memo_open ("GPilotS waypoints\n");
    for (i = DmNumRecordsInCategory (WptDB, Prefs.wptcategory); i > 0; i--)
    {
	DmQueryNextInCategory (WptDB, &pos, Prefs.wptcategory);
	IndexToCustomWpt (pos, WptDB, &wpt);
	pos++;
	SemiToDegre (&wpt.posn, &latitude, &longitude);
	p = DegreToChars (latitude, 0, Prefs.displaymode);
	l = StrPrintF (gbuf1, "W\t%s\t%s\t", wpt.ident, p);
	p = DegreToChars (longitude, 1, Prefs.displaymode);
	l += StrPrintF (gbuf1 + l, "%s\t%s\t%s", p, DToA (wpt.alt, 0, 0), wpt.cmnt);
	memo_write (gbuf1);
    }							   /* for */
    memo_close (NULL);
    FrmEraseForm (fp);
    FrmDeleteForm (fp);
}							   /* ExportWpt */
#endif


/**
   Draws a waypoint symbol centered at given coordinates

   @param  lid local bitmap index
   @param  x x coordinate of center
   @param  y y  coordinate of center
   @return nothing
 */

void
DrawSymbol (unsigned int lid, int x, int y)
{
#define SymbH 14					   // height
#define SymbW 16					   // width
#define SymbO 2						   // y offset
    RectangleType rect = { {lid * SymbW, SymbO}, {SymbW, SymbH} };

    if (SymbolsWH)
	WinCopyRectangle (SymbolsWH, WinGetDrawWindow (), &rect, x - SymbW / 2, y - SymbH / 2,
			  scrCopy);
}							   /* DrawSymbol */


#if 0
GetWptName (Custom_Wpt_Type * wp, char *Header)
{
    UInt i;

    for (i = 1; i < 100; i++)
    {
	StrNCopySafe (wp->ident, Header, sizeof (wp->ident));
	StrIToA (&name[4], i);
	if (SearchWptByName (name, 1, 1) < 0)
	    break;
    }
    StrNCopySafe (wp.ident, name, sizeof (wp.ident));
}
#endif


/**
   Compares two waypoints features according to Prefs.wptsortcrit.
  Used as callback.

  @param p1 pointer on first waypoint
  @param p2 pointer on second waypoint
  @param other ignored
  @param sp1 ignored
  @param sp2 ignored
  @param aph ignored
  @return >0 if *p1>*p2, 0 if *p1==*p2, <0 if *p1<*p2
 */

Int
CompareWpt (void *p1, void *p2, Int other,
	    SortRecordInfoPtr sp1, SortRecordInfoPtr sp2, VoidHand aph)
{
    Int c,b1,b2;

    Custom_Wpt_Type wpt1,
      wpt2;

    PtrToCustomWpt (p1, &wpt1);
    PtrToCustomWpt (p2, &wpt2);
    switch (Prefs.wptsortcrit)
    {
    case Lat:
	c = wpt1.posn.lat > wpt2.posn.lat ? 1 : wpt1.posn.lat == wpt2.posn.lat ? 0 : -1;
	break;

    case Lon:
	c = wpt1.posn.lon > wpt2.posn.lon ? 1 : wpt1.posn.lon == wpt2.posn.lon ? 0 : -1;
	break;

    case Dist:
	c = DistComp (wpt1.posn.lat, wpt1.posn.lon, wpt2.posn.lat, wpt2.posn.lon, ReferencePt.lat,
		      ReferencePt.lon);
	break;

    case Symbol:
	if (wpt1.smbl == wpt2.smbl)
	    c = StrCompare (wpt1.ident, wpt2.ident);
	else
	    c = wpt1.smbl > wpt2.smbl ? 1 : -1;
	break;

    case Alt:
	c = wpt1.alt > wpt2.alt ? 1 : wpt1.alt == wpt2.alt ? 0 : -1;
	break;

    case Bear:
      b1 = Bearing (ReferencePt.lat, ReferencePt.lon, wpt1.posn.lat, wpt1.posn.lon);
      b2 = Bearing (ReferencePt.lat, ReferencePt.lon, wpt2.posn.lat, wpt2.posn.lon);
      c = b1 > b2 ? 1 :b1 ==b2 ? 0 : -1;
      break;

    case Alpha:
    default:
	c = StrCompare (wpt1.ident, wpt2.ident);
    }
    return c;
}							   /* CompareWpt */


/**
   Sort waypoints according to given criteria.
   If allready sorted following given criteria nothing is done
   except if s == Current.

   @param  s sort criteria. If s == Current, the sort is forced on current criteria
   @return nothing
 */

void
SortWpts (WptSortCritType s)
{
    FormPtr fp = NULL;

    if (Prefs.wptsortcrit == s)
	return;
    Prefs.wptsortcrit = s == Current ? Prefs.wptsortcrit : s;
    fp = FrmInitForm (SortForm);
    FrmDrawForm (fp);
    DmQuickSort (WptDB, (DmComparF *) & CompareWpt, 0);
    FrmEraseForm (fp);
    FrmDeleteForm (fp);
    MainListDirty = 1;
}							   /* SortWpts */


/** Search waypoint by name
 @param name  part of name to search (null terminated)
 @param exact  if true, name contains the exact string to search
       si false,name contains the begining of the name to search
 @param all  si true, search in all categories
         else search in current category
 @return index of wayoint if found, -1 else
 */

long
SearchWptByName (char *name, Boolean exact, Boolean all)
{
    ObjectHandle RecH;
    //    Raw_D103_Wpt_Type *wp;
    Custom_Wpt_Type wp;
    UInt i,
      c = 0;
    Boolean stop = 0;

    while ((RecH =
	    (ObjectHandle) DmQueryNextInCategory (WptDB, &c,
						  all ? dmAllCategories : Prefs.wptcategory))
	   && !stop)
    {
	c++;
	//      wp = MemHandleLock (RecH);
	HandleToCustomWpt (RecH, &wp);
	i = sizeof (wp.ident);
	if (exact)
	{
	    for (; wp.ident[i - 1] == ' '; i--)
		if (i == 0)
		    break;				   /* pour ignorer blancs en fin de nom */
	    i = max (i, StrLen (name));
	}
	if (StrNCaselessCompare (wp.ident, name, min (i, StrLen (name))) == 0)
	    stop = 1;
	//      MemHandleUnlock (RecH);
	if (stop)
	    break;
    }
    if (stop)
	return (c - 1);
    else
	return (-1);
}							   /* SearchWptByName */

#define XName 1						   /* coord. aff. nom */
#define XCmnt 48					   /* coord. aff. commentaire */
#define XBear 105					   /* coord. aff. bearing */
#define XSymb 136					   /* coord. aff. symbole */
#define XEnd 150

/**
   Draws an item of the main page's list.
 Take appropriate actions given the type of the object concerned
 and the display and sort options.

  @param itemNum item number
  @param bounds rectangle to draw into
  @param itemsText ignored
  @return nothing
 */

void
MainListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)
{
    VoidHand RecH = NULL;
    Custom_Wpt_Type wpt;
    Custom_Rte_Hdr_Type rth;
    Object *TRecP;
    LocalFormatID wptformat;
    Object *objp;
    Err error;
    char *p;
    int i,
      l;
    double latitude,
      longitude;
    double d;

    debug (GTALKDBG_INFOS, ("MainListDraw(): itemNum=%d", itemNum));
    if (*CurrentCategory == dmAllCategories)
    {
	RecH = DmQueryRecord (CurrentDB, (UInt) itemNum);
	lastrecordindex = itemNum + 1;
    }
    else
    {							   /* categorie particuliere */
	if (itemNum == (lastitemnum + 1))		   /* optimisation */
	{
	    RecH = DmQueryNextInCategory (CurrentDB, &lastrecordindex, *CurrentCategory);
	}
	else
	{
	    lastrecordindex = 0;
	    error =
		DmSeekRecordInCategory (CurrentDB, &lastrecordindex, itemNum, dmSeekForward,
					*CurrentCategory);
	    RecH = DmQueryRecord (CurrentDB, lastrecordindex);
	}
	lastitemnum = itemNum;
	lastrecordindex++;
    }
    error = DmGetLastErr ();				   // debug

    if (RecH)
    {
	switch (Prefs.todisplay)
	{
	case DispWpt:					   /* affichage de waypoints */
/*  	    WRecP = MemHandleLock (RecH); */
/*  	    if (MemHandleSize (RecH)==sizeof(Raw_D107_Wpt_Type)) G107ToCustom((Raw_D107_Wpt_Type *)WRecP,&wpt); */
/*  else G103ToCustom(WRecP,&wpt); */
	    wptformat = HandleToCustomWpt ((ObjectHandle) RecH, &wpt);
	    DrawCharsInWidth (wpt.ident, XName, bounds->topLeft.y, XCmnt - XName, 0);
	    switch (Prefs.wptsortcrit)
	    {
	    case Dist:
	    case Bear:
		d = Distance (ReferencePt.lat, ReferencePt.lon, wpt.posn.lat, wpt.posn.lon,
			      Prefs.distunit);
		l = StrPrintF (gbuf1, "%s %s", DToA (d, 0, 2), DistanceUnitNames[Prefs.distunit]);
		DrawCharsInWidth (gbuf1, XCmnt, bounds->topLeft.y, XBear - XCmnt, 1);
		i = Bearing (ReferencePt.lat, ReferencePt.lon, wpt.posn.lat, wpt.posn.lon);
		l = StrPrintF (gbuf1, "%d ", i);
		DrawCharsInWidth (gbuf1, XBear, bounds->topLeft.y, XSymb - XBear, 1);
		break;
	    case Alpha:
	    case Symbol:
		DrawCharsInWidth (wpt.cmnt, XCmnt, bounds->topLeft.y, XSymb - XCmnt, 0);
		break;
	    case Lat:
		SemiToDegre (&wpt.posn, &latitude, &longitude);
		translate (FromWGS84, &latitude, &longitude, Prefs.datum);
		p = DegreToChars (latitude, 0, Prefs.displaymode);
		DrawCharsInWidth (p, XCmnt, bounds->topLeft.y, XSymb - XCmnt, 0);
		break;
	    case Lon:
		SemiToDegre (&wpt.posn, &latitude, &longitude);
		translate (FromWGS84, &latitude, &longitude, Prefs.datum);
		p = DegreToChars (longitude, 1, Prefs.displaymode);
		DrawCharsInWidth (p, XCmnt, bounds->topLeft.y, XSymb - XCmnt, 0);
		break;
	    case Alt:
		l = StrPrintF (gbuf1, "%s m", DToA (wpt.alt, 0, 0));
		DrawCharsInWidth (gbuf1, XCmnt, bounds->topLeft.y, XSymb - XCmnt, 1);
		break;
	    default:
	    }
#ifdef SELECT
	    if (Prefs.selectmode)
	    {
		FontID fnt = FntSetFont (symbol11Font);

		if (RecIsSecret (CurrentDB, lastrecordindex - 1))
		    WinDrawChar (1, XSymb + 3, bounds->topLeft.y);
		else
		    WinDrawChar (0, XSymb + 3, bounds->topLeft.y);
		FntSetFont (fnt);
	    }
	    else
#endif
	      {   /* optimisable avec DrawSymbol $$ */
		if ((wptformat == G103Type) || (wptformat == Garmin107Type))
		  {
		    BitmapPtr BitP;
		    VoidHand BitH =  DmGetResource ('Tbmp',  SymbolTypeTosmbl (wpt.smbl)  + DotBMID);
		    if (BitH)
		      {
			BitP = MemHandleLock (BitH);
			WinDrawBitmap ( BitP, bounds->extent.x-13, bounds->topLeft.y-1);
			MemHandleUnlock (BitH);
			DmReleaseResource (BitH);
		      }
		  }
		else		DrawSymbol (symb_gid2lid (wpt.smbl), bounds->extent.x - /* 12 */5, bounds->topLeft.y + 7);
	      }
	    break;

	case DispTrk:					   /* affichage de track */
	    TRecP = MemHandleLock (RecH);
	    if (lastrecordindex == Prefs.PermTrackIndex)
		WinDrawChars ("*", 1, XName, bounds->topLeft.y);
	    l = XName + FntCharWidth ('*');
	    DrawCharsInWidth (TRecP->d.CustTrkHdr.name, l, bounds->topLeft.y, XBear - l, 0);
	    l = StrPrintF (gbuf1, "%d %spts", (int) TrkptsInTrkDB (RecH),
			   TRecP->d.CustTrkHdr.type == CompactTrkPt ? "c-" : "  ");
	    DrawCharsInWidth (gbuf1, XBear, bounds->topLeft.y, XEnd - XBear, 0);
	    MemHandleUnlock (RecH);
#ifdef SELECT
	    if (Prefs.selectmode)
	    {
		FontID fnt = FntSetFont (symbol11Font);

		if (RecIsSecret (CurrentDB, lastrecordindex - 1))
		    WinDrawChar (1, XSymb + 3, bounds->topLeft.y);
		else
		    WinDrawChar (0, XSymb + 3, bounds->topLeft.y);
		FntSetFont (fnt);
	    }
#endif
	    break;

	case DispRte:					   /* affichage de route */
	    objp = MemHandleLock (RecH);
	    PtrToCustomRteHdr (objp, &rth);
	    StrIToA (gbuf1, (long) (rth.nmbr));
	    StrNCat (gbuf1, " : ", sizeof (gbuf1));
	    if (rth.cmnt[0] && StrNCompare (rth.cmnt, "________", 8))
		StrNCatSafe (gbuf1, rth.cmnt, sizeof (gbuf1), sizeof (rth.cmnt));
	    else if (ObjectsInRec (RecH) > 1)
	    {						   /* pas de nom : on affiche les extremites */
		PtrToCustomWpt ((void *) objp + OffsetOfObjectInRec (RecH, 1), &wpt);
		StrNCatSafe (gbuf1, wpt.ident, sizeof (gbuf1), sizeof (wpt.ident));	/* $$ */
		StrNCatSafe (gbuf1, " to ", sizeof (gbuf1), 4);
		PtrToCustomWpt ((void *) objp + OffsetOfObjectInRec (RecH, ObjectsInRec (RecH) - 1),
				&wpt);
		StrNCatSafe (gbuf1, wpt.ident, sizeof (gbuf1), sizeof (wpt.ident));	/* $$ */
	    }
	    //              buf[sizeof (buf) - 1] = 0;
	    //      WinDrawChars (gbuf1, StrLen (gbuf1), bounds->topLeft.x + XName, bounds->topLeft.y);
	    DrawCharsInWidth (gbuf1, bounds->topLeft.x + XName, bounds->topLeft.y, XBear - XName,
			      0);
	    l = StrPrintF (gbuf1, "(%d pts)", (int) WptsInRteDB (RecH));
	    //      WinDrawChars (gbuf1, l, 100, bounds->topLeft.y);
	    DrawCharsInWidth (gbuf1, bounds->topLeft.x + XBear, bounds->topLeft.y, XEnd - XBear, 0);
	    MemHandleUnlock (RecH);
#ifdef SELECT
	    if (Prefs.selectmode)
	    {
		FontID fnt = FntSetFont (symbol11Font);
		if (RecIsSecret (CurrentDB, lastrecordindex - 1))
		    WinDrawChar (1, XSymb + 3, bounds->topLeft.y);
		else
		    WinDrawChar (0, XSymb + 3, bounds->topLeft.y);
		FntSetFont (fnt);
	    }
#endif

	case DispUnknown:
	}
/*  	MemHandleUnlock (RecH); */
    }
    else
	WinDrawChars ("******", 6, bounds->topLeft.x, bounds->topLeft.y);

}							   // MainListDraw



/**
   Set reference point to waypoint whose index is passed if user confirms

   @param  index index of waypoint record
   @param ask 1 -> a confirmation dialog is displayed, 0 -> it's set without asking
   @return 1 if user accepts, 0 otherwise
 */

int
SetReferencePt (UInt index, Boolean ask)
{
    int r = 0;
    Custom_Wpt_Type wpt;

    IndexToCustomWpt (index, WptDB, &wpt);
    SysCopyStringResource (gbuf1, StrWptRefFormat);
    StrPrintF (gbuf2, gbuf1, wpt.ident);
    if (!ask || FrmCustomAlert (ConfID, gbuf2, " ", " "))
    {
	ReferencePt.lat = wpt.posn.lat;
	ReferencePt.lon = wpt.posn.lon;
	StrNCopySafe (ReferencePtName, wpt.ident, sizeof (ReferencePtName));
	DmRecordInfo (WptDB, index, NULL, &Prefs.refptuid, NULL);
	r = 1;
    }
    return (r);
}							   /* SetReferencePt */


/**
   Updates scrollbar of main list

@param  LstPtr pointer on list associated with scrollbar
@param  ScrollId id of the scrollbar in the form
@return nothing
*/

static void
UpdateScrollBar (ListPtr LstPtr, Word ScrollId)
{
  ScrollBarPtr scrollp = GetObjPtr (ScrollId);
    Word currentPosition,
      textHeight,
      fieldHeight,
      maxValue;

    //    FldGetScrollValues (FdPtr, &currentPosition, &textHeight, &fieldHeight);
    textHeight=LstGetNumberOfItems (LstPtr);
    fieldHeight=LstGetVisibleItems (LstPtr);
    currentPosition=LstGlueGetTopItem (LstPtr);
    if (textHeight > fieldHeight)
	maxValue = textHeight - fieldHeight;
    else if (currentPosition)
	maxValue = currentPosition;
    else
	maxValue = 0;
    SclSetScrollBar (scrollp, currentPosition, 0, maxValue, fieldHeight - 1);
}     /* UpdateScrollBar */


/**
   Scroll lines of a list and updates associated scrollbar

@param  LstPtr pointer on list to scroll
@param  ScrollId id of associated scrollbar to update
@param  numLinesToScroll number of lines to scroll
@param  redraw unused
@return nothing
*/

static void
ScrollLines (ListPtr LstPtr,Word ScrollId, int numLinesToScroll, Boolean redraw)
{

    if (numLinesToScroll < 0)
	 LstScrollList(LstPtr, winUp, -numLinesToScroll );
    else
	LstScrollList (LstPtr, winDown, numLinesToScroll );
/*     if ((FldGetNumberOfBlankLines (FdPtr) && numLinesToScroll < 0) || redraw) */
	UpdateScrollBar (LstPtr,ScrollId);
}  /* ScrollLines */


/**
   Updates variables and list display of main form.
   Takes into account changes of categories, of items to display,...
  Updates menus, buttons, title,...

  @return nothing
 */

static void
UpdateMainForm ()
{
  //    char buf[20];
    UInt RecordsNumber;
    Boolean AtTop = 0,
      fit;
    Word pos = -2;
    ListPtr LstPtr = (ListPtr) GetObjPtr (MainListID);
    FormPtr frm = FrmGetActiveForm ();
    static ItemToDisplay PrevDisplay = -1;
    static UInt PrevCategory = dmAllCategories;
    static Boolean selmode = 0;
    Int16 width,length;

    debug_in (GTALKDBG_CALLS, ("->UpdateMainForm()"));
    if ((Prefs.todisplay != PrevDisplay) || ((CurrentDB != WptDB)&&(CurrentDB != TrkDB)&&(CurrentDB != RteDB))
#ifdef SELECT
	|| (Prefs.selectmode != selmode)
#endif
	)
    {
	switch (Prefs.todisplay)
	  {
	  default:
	    Prefs.todisplay = DispWpt;

	  case DispWpt:
	    CurrentDB = WptDB;
	    CurrentCategory = &Prefs.wptcategory;
#ifdef SELECT
	    if (Prefs.selectmode)
	      FrmSetMenu (frm, MainSelMenuID);
	    else
#endif
	      FrmSetMenu (frm, MainMenuID);
	    MainListDirty = 1;
	    break;

	  case DispTrk:
	    CurrentDB = TrkDB;
	    CurrentCategory = &Prefs.trkcategory;
#ifdef SELECT
	    if (Prefs.selectmode)
	      FrmSetMenu (frm, MainSelMenuID2);
	    else
#endif
	      FrmSetMenu (frm, MainMenuID2);
	    MainListDirty = 1;
	    break;

	  case DispRte:
	    //        ToDisplay = DispRte;                                   /* garde-fou */
	    CurrentDB = RteDB;
	    CurrentCategory = &Prefs.rtecategory;
#ifdef SELECT
	    if (Prefs.selectmode)
	      FrmSetMenu (frm, MainSelMenuID3);
	    else
#endif
	      FrmSetMenu (frm, MainMenuID3);
	    MainListDirty = 1;
	    break;
	    /*          default: */
	    /*              MainListDirty = 0; */
	  }
	PrevDisplay = Prefs.todisplay;
	selmode = Prefs.selectmode;
	AtTop = 1;
	//      MainListDirty = 1;
	debug (GTALKDBG_INFOS, ("UpdateMainForm(): ToDisplay changed to %d", Prefs.todisplay));
    }
    else
    {
	if (*CurrentCategory != PrevCategory)
	{
	    debug (GTALKDBG_INFOS,
		   ("UpdateMainForm(): cat changed from %d to %d", PrevCategory, *CurrentCategory));
	    MainListDirty = 1;
	    if ((*CurrentCategory != dmAllCategories) && (PrevCategory != dmAllCategories))
		AtTop = 1;
	}
    }
    PrevCategory = *CurrentCategory;
    if (MainListDirty)
      {
	debug (GTALKDBG_INFOS, ("UpdateMainForm(): MainListDirty=1"));
	if((Prefs.todisplay==DispWpt)&&(Prefs.wptsortcrit==Alpha)){
	  ShowObject (MainSearchID);
	  ShowObject (MainFdSearchID);
	}
	else {
	  HideObject (MainSearchID);
	  HideObject (MainFdSearchID);
	}
	CategoryGetName (CurrentDB, *CurrentCategory, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (MainCatPopupID), CategoryName);

	RecordsNumber = DmNumRecordsInCategory (CurrentDB, *CurrentCategory);
	length=StrPrintF (gbuf1, "%s",
			  Prefs.todisplay == DispWpt ?  Prefs.CurWptDBName:
			  Prefs.todisplay == DispTrk ?  Prefs.CurTrkDBName: 
			  Prefs.CurRteDBName);
	width=80;
	//length=StrLen(gbuf1);
	FntCharsInWidth (gbuf1, &width, &length, &fit);
	gbuf1[length]=0;
	FrmCopyTitle (frm, gbuf1);
	length=StrPrintF (gbuf1, "%d %s", RecordsNumber, 
			  Prefs.todisplay == DispWpt ? "Waypoint(s)" :
			  Prefs.todisplay == DispTrk ? "Track(s)" : "Route(s)");
	HideObject (MainLabelID);
	FrmCopyLabel (frm, MainLabelID,gbuf1);
	ShowObject (MainLabelID);
	LstSetListChoices (LstPtr, NULL, RecordsNumber);
	pos = RecIdToPosition (CurrentDB, *CurrentCategory, CurrentRecordID, NULL);
	LstSetSelection (LstPtr, pos);
	debug (GTALKDBG_INFOS, ("UpdateMainForm(): before if (pos"));
	if (pos != -1)
	  {
	    //    if (AtTop)LstSetTopItem (LstPtr,0);
	    /*  else */ LstMakeItemVisible (LstPtr, LstGetSelection (LstPtr));
	  }
	else
	  LstSetTopItem (LstPtr, 0);
	debug (GTALKDBG_INFOS, ("UpdateMainForm(): before LstDrawList"));
	if (FrmVisible (frm))
	  LstDrawList (LstPtr);
	UpdateScrollBar (LstPtr,MainScrollID);
      }
    debug_out (GTALKDBG_CALLS,
	       ("<-UpdateMainForm(), AtTop=%d, pos=%d, %s", AtTop, pos,
		MainListDirty ? "updated" : "unchanged"));
    MainListDirty = 0;
}							   /* UpdateMainForm */


/**
   Event handler of main form.

  @param  e pointer on event received
  @return true if event handled
 */

static Boolean
MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false,
      Edited;
    static FormPtr frm;
    UInt category,
      itemNum,
      windex,
      d,
      t;
    static ListPtr LstPtr;
    VoidHand RecH;
    VoidHand ResH;
    Err error;
    BitmapPtr BitP;
#ifdef SELECT
    Int16 x,
      y;
#endif

    if (e->eType != nilEvent) debug_in (GTALKDBG_CALLS, ("->MainFormHandleEvent(%s)", e->eType > LASTVALIDEVTNAME ?StrIToA(gbuf1,e->eType):EvtNames[e->eType]));
    switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	LstPtr = (ListPtr) (GetObjPtr (MainListID));
	LstPtr->attr.hasScrollBar =1; // very dirty, en attendant correction pilrc $$
	LstSetDrawFunction (LstPtr, &MainListDraw);
	//      CurrentCategory = &Prefs.wptcategory;
	UpdateMainForm ();
	SymbolsWH = WinCreateOffscreenWindow (2400, 17, screenFormat, &error);	/* offscreen work window for bitmaps */
	if (SymbolsWH)
	{
	    ResH = DmGetResource ('Tbmp', SymbolsBMID);
	    if (ResH)
	    {
		WinHandle CurWH = WinGetDrawWindow ();
		BitP = MemHandleLock (ResH);
		WinSetDrawWindow (SymbolsWH);
		WinDrawBitmap (BitP, 0, 0);
		MemHandleUnlock (ResH);
		WinSetDrawWindow (CurWH);
		DmReleaseResource (ResH);
	    }
	}
	else
	    FrmCustomAlert (ErrorID, "Can't create window (error ", StrIToA (gbuf1, error), ")");
	if (RecIdToPosition (WptDB, dmAllCategories, Prefs.refptuid, &windex) >= 0)
	    SetReferencePt (windex, 0);			   /* recherche point de ref */
	else
	{
	    StrNCopySafe (ReferencePtName, "??????", sizeof (ReferencePtName));
	    ReferencePt.lat = 0;
	    ReferencePt.lon = 0;
	}
	FrmSetFocus (frm, FrmGetObjectIndex (frm, MainFdSearchID));
	FrmDrawForm (frm);
	//      UpdateMainForm ();
	//      FrmDrawForm (frm);
	search_buf[0] = OnlyNews = 0;
#ifdef NEWSERIALMGR
	GpslibSetSerialPort (Prefs.portid);
#endif
	handled = true;
	break;

    case frmCloseEvent:
      if(SymbolsWH)WinDeleteWindow (SymbolsWH, 0);
	break;

    case /*  frmUpdateEvent */ winEnterEvent:
	if ((e->data.winEnter.enterWindow == FrmGetWindowHandle (frm)) && FrmVisible (frm))
	    UpdateMainForm ();
	break;

    case lstSelectEvent:				   // un element est designe
	if (e->data.lstSelect.listID == MainListID)
	{
	    RectangleType r;

	    EditMode = EditWpt;
	    itemNum = e->data.lstSelect.selection;
	    CurrentRecordIndex = 0;
	    (void) DmSeekRecordInCategory (CurrentDB, &CurrentRecordIndex, itemNum, dmSeekForward,
					   *CurrentCategory);
	    CurrentRecordID = GetRecordID (CurrentDB, CurrentRecordIndex);
#ifdef SELECT
	    EvtGetPen (&x, &y, &Edited);		   // test si colonne des check boxes
	    GetObjectBounds (MainListID, &r);
	    if ((x > XSymb + 4) && (x < XSymb + 14))
	    {
		//toggle select flag
		RecSetSecret (CurrentDB, CurrentRecordIndex,
			      !RecIsSecret (CurrentDB, CurrentRecordIndex));
		MainListDirty = 1;
		UpdateMainForm ();
		handled = true;
		break;
	    }
#endif
	    switch (Prefs.todisplay)
	    {
	    case DispWpt:
#ifndef SKIPEDIT
		FrmPopupForm (EditForm);
		//      MainListDirty=1;
#endif
		break;
	    case DispTrk:
		FrmPopupForm (TrackForm);
		break;
	    case DispRte:
		CurrentRouteIndex = CurrentRecordIndex;
		FrmPopupForm (RouteForm);
	    default:
	    }
	    handled = true;
	}
	break;

    case menuEvent:
	MenuEraseStatus (NULL);
	search_buf[0] = 0;
	switch (e->data.menu.itemID)
	{
	case MenuWptID:
	    Prefs.todisplay = DispWpt;
	    break;

	case MenuRteID:
	    Prefs.todisplay = DispRte;
	    break;

	case MenuTrkID:
	    Prefs.todisplay = DispTrk;
	    break;

#ifdef SELECT
	case MenuSelID:
	    Prefs.selectmode = !Prefs.selectmode;
	    break;
#endif

	case MenuViewID:
	    if (Prefs.todisplay == DispTrk)
	    {
		MapFlags.AllowedOps = MapOpMapTrack;
	    }
	    else
	    {
		MapFlags.AllowedOps = MapOpWptDisp;
		Prefs.PermTrackIndex = 0;
		MapFlags.DrawWpts = 1;
	    }
	    FrmPopupForm (MapForm);
	    break;

	case MenuNavID:
	    MapFlags.AllowedOps = /*  MapOpWptDisp| */ MapOpActiveRte | MapOpTrack;
	    FrmPopupForm (MapForm);
	    break;

#ifdef EXPORT
	case MenuExportID:
	    ExportWpt ();
	    break;
#endif

	case MenuLatID:
	    SortWpts (Lat);
	    break;

	case MenuLonID:
	    SortWpts (Lon);
	    break;

	case MenuNameID:
	    SortWpts (Alpha);
	    break;

	case MenuSymbID:
	    SortWpts (Symbol);
	    break;

	case MenuDistID:
	    SortWpts (Dist);
	    break;

	case MenuAltID:
	    SortWpts (Alt);
	    break;

	case MenuBearID:
	    SortWpts (Bear);
	    break;

	case MenuNewDBID:
	    NewDB();
	    MainListDirty = 1;
	    break;

	case MenuPrefID:				   /* preferences */
	    d = Prefs.distunit;
	    SetPreferences ();
	    //      if (Prefs.distunit != d)
	    MainListDirty = 1;
	    break;

	case MenuAboutID:
	    FrmAlert (idAbout);
	    break;

	case MenuDeleteID:				   /* suppression contenu catgorie */
	    CategoryGetName (CurrentDB, *CurrentCategory, CategoryName);
#ifdef SELECT
	    if (Prefs.selectmode)
	    {
		SysCopyStringResource (gbuf1, StrRmSelItemsID);
		CategoryName[0] = 0;
	    }
	    else
#endif
		SysCopyStringResource (gbuf1, StrRemoveItemsID);
	    if (FrmCustomAlert (ConfID, gbuf1, CategoryName, " ?"))
	    {
		d = 0;
		for (t = DmNumRecordsInCategory (CurrentDB, *CurrentCategory); t > 0; t--)
		{
		    RecH = DmQueryNextInCategory (CurrentDB, &d, *CurrentCategory);
#ifdef SELECT
		    if (Prefs.selectmode)
		    {
			if (RecIsSecret (CurrentDB, d))
			    DmRemoveRecord (CurrentDB, d);
			else
			    d++;
		    }
		    else
#endif

			DmRemoveRecord (CurrentDB, d);
		}
		MainListDirty = 1;
	    }
	    break;

#ifdef SELECT
	case MenuSelAllID:
	    d = 0;
	    while ((RecH = DmQueryNextInCategory (CurrentDB, &d, *CurrentCategory)))
		RecSetSecret (CurrentDB, d++, 1);
	    MainListDirty = 1;
	    break;

	case MenuUnSelAllID:
	    d = 0;
	    while ((RecH = DmQueryNextInCategory (CurrentDB, &d, *CurrentCategory)))
		RecSetSecret (CurrentDB, d++, 0);
	    MainListDirty = 1;
	    break;

	case MenuCatID:
	    FrmPopupForm (CategoryForm);
	    break;
#endif

	case MenuDBID:
	     ChangeDB();
	    MainListDirty = 1;
	    break;

	case MenuNewID:				   /* nouveau waypoint/route */
	    EditMode = NewWpt;
	    OnlyNews = 0;
	    if (Prefs.todisplay == DispWpt)
#ifndef SKIPEDIT
		FrmPopupForm (EditForm)
#endif
		    ;
	    else
	    {
#ifdef GRAPHICALROUTEEDIT
		CurrentRouteIndex = 0;
		MapFlags.DrawRtes = 1;
		MapFlags.AllowedOps = MapOpRteNew | MapOpWptDisp;
		FrmPopupForm (MapForm);
#else
		FrmPopupForm (EditRouteForm);
#endif
	    }
	    break;

	case MenuNewFRID:				   /* nouveau waypoint a partir de ref. */
	    EditMode = FromRefWpt;
	    //      (void) memcpy (&CurrentWGS84Pos, &ReferencePt, sizeof (CurrentWGS84Pos));
	    FrmPopupForm (EditRefForm);
	    break;

	case MenuMarkID:				   /* marque position actuelle */
	    GpsOp = gpsGetPosition;
	    FrmPopupForm (ProgressForm);
	    break;

	case MenuTimeID:				   /* time request */
	    GpsOp = gpsGetTime;
	    FrmPopupForm (ProgressForm);
	    break;

	case MenuCheckID:				   /* check gps  */
	    GpsOp = gpsGetInfos;
	    FrmPopupForm (ProgressForm);
	    break;

	case MenuHostID:				   /* host mode */
	    FrmPopupForm (HostForm);
	    break;

	case MenuUploadID:				   /* upload */
#ifdef SELECT
	    if (Prefs.selectmode)
	    {
		GpsOp = gpsUpload;
		FrmPopupForm (ProgressForm);
	    }
	    else
#endif
		FrmPopupForm (UploadForm);
	    break;

	case MenuDownNewID:
	    OnlyNews = 1;
	case MenuDownloadID:				   /* download */
	    FrmPopupForm (DownloadForm);
	    break;

#ifdef EVENTS
	case MenuLogEventsID:				   /* log d'evenements asynchrones */
	  GpsOp = gpsEnableAsyncEvts;
	    //              for (t = DmNumRecords(LogDB);t;t--)DmRemoveRecord(LogDB,t-1);
	    FrmPopupForm (ProgressForm);
	    break;
#endif
	    //            case MenuDumpEventsID:                      /* (r)envoi des evenements asynchrones */
	    //              ToDisplay += 16;
	    //              FrmGotoForm (ProgressForm);
	    //              break;

	}
	UpdateMainForm ();
	handled = true;
	break;

    case sclRepeatEvent:
	ScrollLines (LstPtr,MainScrollID,e->data.sclRepeat.newValue - e->data.sclRepeat.value, false);
	break;

    case ctlSelectEvent:
	search_buf[0] = 0;
	switch (e->data.ctlSelect.controlID)
	{
	case MainCatPopupID:
	    category = *CurrentCategory;
	    Edited =
		CategorySelect (CurrentDB, frm, MainCatPopupID, MainCatListID, true,
				CurrentCategory, CategoryName, 1, 0);
	    if (category != *CurrentCategory)
	    {						   /* changement de categorie ? */
		UpdateMainForm ();
	    }
	    break;
	}
	handled = true;
	break;

    case ctlRepeatEvent:
	switch (e->data.ctlSelect.controlID)
	{

	}
	break;

    case keyDownEvent:
	switch (e->data.keyDown.chr)
	{
	case pageUpChr:
	  //	    LstScrollList (LstPtr, winUp, LstGetVisibleItems (LstPtr) - 1);
	ScrollLines (LstPtr,MainScrollID,1-LstGetVisibleItems (LstPtr) , false);
	handled = true;
	    break;
	case pageDownChr:
	  //	    LstScrollList (LstPtr, winDown, LstGetVisibleItems (LstPtr) - 1);
	ScrollLines (LstPtr,MainScrollID,LstGetVisibleItems (LstPtr)- 1 , false);
	handled = true;
	    break;
#ifdef DEBUG
	case '.':
	  SetDebug ();
	  //	    SndPlaySystemSound (sndConfirmation /*  sndInfo */ );
	    handled = true;
	    break;
#endif
	default:					   /* recherche dans les noms */
	    if ((Prefs.todisplay == DispWpt) && (e->data.keyDown.chr))
	    {
		long l;
		char c = e->data.keyDown.chr;
		if (c == backspaceChr)
		{
		    if (StrLen (search_buf) > 0)
			search_buf[StrLen (search_buf) - 1] = 0;
		    else
			SndPlaySystemSound (sndError);
		}
		else
		    StrNCatSafe (search_buf, &c, sizeof (search_buf), 1);
		//              LstPtr = (ListPtr) (GetObjPtr (MainListID));
		if (StrLen (search_buf))
		{
		    l = SearchWptByName (search_buf, 0, 0);
		    if (l >= 0)
		    {
			CurrentRecordIndex = l;
			CurrentRecordID = GetRecordID (CurrentDB, CurrentRecordIndex);
			//LstMakeItemVisible (LstPtr, DmPositionInCategory (CurrentDB, c, *CurrentCategory));
			LstSetSelection (LstPtr,
					 DmPositionInCategory (CurrentDB, CurrentRecordIndex,
							       *CurrentCategory));
		    }
		    else
		    {
			SndPlaySystemSound (sndError);
			if (StrLen (search_buf) > 0)
			    search_buf[StrLen (search_buf) - 1] = 0;
		    }
		}
		else
		    LstSetSelection (LstPtr, -1);
	    }
	}						   /* switch */
	break;

    default:
	break;
    }
     if (e->eType != nilEvent) debug_out (GTALKDBG_CALLS, ("<-MainFormHandleEvent(%s)=%d", e->eType > LASTVALIDEVTNAME ?StrIToA(gbuf1,e->eType):EvtNames[e->eType], handled));

    return handled;
}							   // MainFormHandleEvent


/**
   Display and setting of preferences with a modal form.

@return nothing
*/

void
SetPreferences ()
{
  Err error;
  FormPtr previousForm = FrmGetActiveForm ();
  FormPtr frm = FrmInitForm (PrefForm);
  ControlPtr DupPtr,
    PopupOffsetPtr,
    PopupUnitPtr,
    PopupDispPtr,
    PopupGPSPtr,
    PopupLogPtr,
    PopupSerPtr,
    PopupWptPtr,
    PopupRtePtr,
    PopupTrkPtr,
    LoosePtr;
  Int16 offset = Prefs.utcoffset,
    i,
    j,matched;
  unsigned int index;
  ListPtr LstOffsetPtr,
    LstUnitPtr,
    LstDispPtr,
    LstGPSPtr,
    LstDatumPtr,
    LstLogPtr,
    LstSerPtr,
    LstWptPtr,
    LstRtePtr,
    LstTrkPtr;
  static const UInt LogIntervals[] =
    { 5, 10, 15, 20, 30, 40, 50, 60, 120, 180, 300, 600, 900, 1200 };
  char *DatumNames[nDatums];
#ifdef NEWSERIALMGR
  SerialDevInfo DevInfos;
#endif
  char * WptNames[MaxDBNames],
    * RteNames[MaxDBNames],
    * TrkNames[MaxDBNames];

  FrmSetActiveForm (frm);

  LstOffsetPtr = (ListPtr) (GetObjPtr (PrefListID));
  LstUnitPtr = (ListPtr) (GetObjPtr (PrefUnitListID));
  LstDispPtr = (ListPtr) (GetObjPtr (PrefDispListID));
  LstGPSPtr = (ListPtr) (GetObjPtr (PrefGPSListID));
  LstLogPtr = (ListPtr) (GetObjPtr (PrefLogListID));
  LstWptPtr = (ListPtr) (GetObjPtr (PrefWptListID));
  LstRtePtr = (ListPtr) (GetObjPtr (PrefRteListID));
  LstTrkPtr = (ListPtr) (GetObjPtr (PrefTrkListID));
#ifdef NEWSERIALMGR
  LstSerPtr = (ListPtr) (GetObjPtr (PrefSerListID));
#endif
  DupPtr = GetObjPtr (PrefDupID);
#ifdef LOOSE
  LoosePtr=GetObjPtr (PrefLooseID);
#endif
  PopupOffsetPtr = (ControlPtr) (GetObjPtr (PrefPopupID));
  PopupUnitPtr = (ControlPtr) (GetObjPtr (PrefPopUnitID));
  PopupDispPtr = (ControlPtr) (GetObjPtr (PrefPopDispID));
  PopupGPSPtr = (ControlPtr) (GetObjPtr (PrefGPSPopID));
  PopupLogPtr = (ControlPtr) (GetObjPtr (PrefPopLogID));
  PopupWptPtr = (ControlPtr) (GetObjPtr (PrefWptPopID));
  PopupRtePtr = (ControlPtr) (GetObjPtr (PrefRtePopID));
  PopupTrkPtr = (ControlPtr) (GetObjPtr (PrefTrkPopID));
#ifdef NEWSERIALMGR
  PopupSerPtr = (ControlPtr) (GetObjPtr (PrefSerPopID));
#endif
  for (i = 0; i < nDatums; i++)
    DatumNames[i] = gDatum[i].name;
  LstDatumPtr = (ListPtr) (GetObjPtr (PrefDatumListID));
  LstSetListChoices (LstDatumPtr, DatumNames, nDatums);
  LstSetSelection (LstDatumPtr, Prefs.datum);
  LstMakeItemVisible (LstDatumPtr, Prefs.datum);
  CtlSetLabel (GetObjPtr (PrefDatumPopupID), DatumNames[Prefs.datum]);

  if ((offset < -12) || (offset > 13))		   /* UTC offset */
    offset = 0;
  index = -offset + 13;
  LstSetSelection (LstOffsetPtr, index);
  LstMakeItemVisible (LstOffsetPtr, index);
  CtlSetLabel (PopupOffsetPtr, LstGetSelectionText (LstOffsetPtr, index));

  LstSetSelection (LstUnitPtr, Prefs.distunit);	   /* distance units */
  // LstMakeItemVisible (LstUnitPtr,DistUnit);
  CtlSetLabel (PopupUnitPtr, LstGetSelectionText (LstUnitPtr, Prefs.distunit));

  if (Prefs.displaymode > DDD)
    Prefs.displaymode = DDD;			   /* display format */
  LstSetSelection (LstDispPtr, Prefs.displaymode);
  CtlSetLabel (PopupDispPtr, LstGetSelectionText (LstDispPtr, Prefs.displaymode));

  if (Prefs.emulatedgps > GPS76)
    Prefs.emulatedgps = GIIPlus;			   /* emulated GPS */
  LstSetSelection (LstGPSPtr, Prefs.emulatedgps);
  CtlSetLabel (PopupGPSPtr, LstGetSelectionText (LstGPSPtr, Prefs.emulatedgps));

  for (i = (sizeof (LogIntervals) / sizeof (UInt)) - 1; i >= 0; i--)	/* log interval */
    if (Prefs.loginterval == LogIntervals[i])
      break;
  if (i < 0)
    i = 0;
  LstSetSelection (LstLogPtr, i);
  LstMakeItemVisible (LstLogPtr, i);
  CtlSetLabel (PopupLogPtr, LstGetSelectionText (LstLogPtr, i));

#ifdef NEWSERIALMGR
  i = GetSerialDevsInfos (&DevInfos);			   /* serial port */
  for (j = i - 1; j; j--)
    if (Prefs.portid == DevInfos.u[j].serDevCreator)
      break;
  //    if (!j)Prefs.portid=DevInfos.serDevCreator[0];
  LstSetHeight (LstSerPtr, min (i, 6));
  LstSetListChoices (LstSerPtr, DevInfos.serDevPortInfoStr, i);
  LstSetSelection (LstSerPtr, j);
  //    LstMakeItemVisible (LstDatumPtr, Prefs.datum);
  CtlSetLabel (PopupSerPtr, DevInfos.serDevPortInfoStr[j]);
#endif
  CtlSetValue (DupPtr, Prefs.allowduplicate);		   /* duplicatas */
#ifdef LOOSE
  CtlSetValue (LoosePtr, Prefs.looseprotocol);		   /* loose protocol */
#endif

  i=FindDBNames(WptDBType,GPilotSID,Prefs.CurWptDBName,&matched,gbuf1,WptNames,MaxDBNames);  /* Wpt DB names */
  LstSetHeight (LstWptPtr, min (i, 6));
  LstSetListChoices (LstWptPtr, WptNames, i		       );
  LstSetSelection (LstWptPtr,matched);
  CtlSetLabel (PopupWptPtr,WptNames[matched]);

  i=FindDBNames(RteDBType,GPilotSID,Prefs.CurRteDBName,&matched,WptNames[i-1]+StrLen(WptNames[i-1])+1,RteNames,MaxDBNames);  /* Rte DB names */
  LstSetHeight (LstRtePtr, min (i, 6));
  LstSetListChoices (LstRtePtr, RteNames, i		       );
  LstSetSelection (LstRtePtr,matched);
  CtlSetLabel (PopupRtePtr,RteNames[matched]);

  i=FindDBNames(TrkDBType,GPilotSID,Prefs.CurTrkDBName,&matched,RteNames[i-1]+StrLen(RteNames[i-1])+1,TrkNames,MaxDBNames);  /* Trk DB names */
  LstSetHeight (LstTrkPtr, min (i, 6));
  LstSetListChoices (LstTrkPtr, TrkNames, i		       );
  LstSetSelection (LstTrkPtr,matched);
  CtlSetLabel (PopupTrkPtr,TrkNames[matched]);

  ErrFatalDisplayIf ((TrkNames[i-1]+StrLen(TrkNames[i-1])+1) >= (gbuf1+sizeof(gbuf1)), "Buffer overflow in SetPreferences");

  if (FrmDoDialog (frm) == PrefBtOKID)
    {
      Prefs.utcoffset = -(LstGetSelection (LstOffsetPtr) - 13);
      Prefs.allowduplicate = CtlGetValue (DupPtr);
#ifdef LOOSE
      Prefs.looseprotocol =  CtlGetValue (LoosePtr);
#endif
      Prefs.distunit = LstGetSelection (LstUnitPtr);
      Prefs.displaymode = LstGetSelection (LstDispPtr);
      Prefs.emulatedgps = LstGetSelection (LstGPSPtr);
      Prefs.loginterval = LogIntervals[LstGetSelection (LstLogPtr)];
      Prefs.datum = LstGetSelection (LstDatumPtr);
      error = SwitchDB(&WptDB,WptNames[LstGetSelection (LstWptPtr)], WptDBType);
      if(!error)StrNCopySafe (Prefs.CurWptDBName,WptNames[LstGetSelection (LstWptPtr)] , dmDBNameLength);
      error = SwitchDB(&RteDB,RteNames[LstGetSelection (LstRtePtr)], RteDBType);
      if(!error)StrNCopySafe (Prefs.CurRteDBName,RteNames[LstGetSelection (LstRtePtr)] , dmDBNameLength);
      error = SwitchDB(&TrkDB,TrkNames[LstGetSelection (LstTrkPtr)], TrkDBType);
      if(!error)StrNCopySafe (Prefs.CurTrkDBName,TrkNames[LstGetSelection (LstTrkPtr)] , dmDBNameLength);
#ifdef NEWSERIALMGR
      Prefs.portid = DevInfos.u[LstGetSelection (LstSerPtr)].serDevCreator;
      GpslibSetSerialPort (Prefs.portid);
#endif
    }
  FrmSetActiveForm (previousForm);
  FrmDeleteForm (frm);
}							   // SetPreferences


/**
   Returns the number of data bases of given type and creator
   and, optionaly, a list of their names

@param type type to search
@param creator creator to search
@param namep if !=NULL points on name to compare with scanned data bases names
@param matched pointer on variable to receive index of db whose name match those passed in namep
@param bufp buffer to store names if != NULL
@param bufpp list of pointers on names stored in bufp
@param bufpsize size of bufpp list
@return number of data bases found, 0 if none or error
*/

static UInt16 FindDBNames(UInt32 type, UInt32 creator,  char * namep,Int16 * matched, char * bufp, char ** bufpp,UInt16 bufpsize){
  DmSearchStateType state; 
  UInt16 card,i=0; 
  LocalID currentDBid = 0; 
  Err error;
  char  name[dmDBNameLength];
  UInt16 l=0;

  name[0]=0;
  error = DmGetNextDatabaseByTypeCreator(true, &state,type ,creator , false, &card, &currentDBid); 
  while (!error && currentDBid) { 
    if (i>=bufpsize)break;
     error=DmDatabaseInfo ( card,  currentDBid,name , NULL, NULL,NULL ,NULL , NULL, NULL,NULL,NULL, NULL,NULL );
     if (bufpp)bufpp[i]=bufp+l;
     if (bufp)l+=StrPrintF(bufp+l,"%s",name);
     l++;
     if (namep&&matched)if (!StrCompare(namep,name))*matched=i;

    i++;
    error = DmGetNextDatabaseByTypeCreator( false, &state,type ,creator,false , &card, &currentDBid); 
  }
  return(i);
}  /* FindDBNames */


/**
   Switch from current data base to the one whose name is given.
   If it doesn't exist the new data base is created.
   Displays  message if error.

@param DBP pointer on DmOpenRef of data base to switch from
@param DBName name  of data base to switch to
@param DBType type of new data base
@return 0 if OK
*/

static Err SwitchDB(DmOpenRef * DBP, char *DBName, ULong DBType){
  Err error;
  DmOpenRef olddb;

  olddb=*DBP;
      error = OpenOrCreateDB (DBP,DBName, DBType, GPilotSID);
      if (!error) {
	DmCloseDatabase (olddb);
	*CurrentCategory=dmAllCategories;
	MainListDirty = 1;
      }
      else  {
	if (error!=dmErrAlreadyExists) FrmCustomAlert (ErrorID, "Error ", StrIToA (gbuf2, error), " when opening data base");
	*DBP=olddb;
      }
      return(error);
} /* SwitchDB */


/**
   Displays a form to ask a new data base name to create
   and to switch to

@return nothing
*/

static void NewDB (){
  FieldPtr FdPtr;
  Err error;
  DmOpenRef *curdbp,olddb;
  FormPtr previousForm = FrmGetActiveForm ();
  FormPtr frm = FrmInitForm (NewDBForm);
  UInt32 type;
  char  *name;

  switch (Prefs.todisplay)
    {
    case DispWpt:
      curdbp=&WptDB;
      type=WptDBType;
      name=Prefs.CurWptDBName;
      break;
    case DispTrk:
      curdbp = &TrkDB;
      type=TrkDBType;
      name=Prefs.CurTrkDBName;
      break;
    case DispRte:
      curdbp = &RteDB;
      type=RteDBType;
      name=Prefs.CurRteDBName;
      break;
    default:
      return;
    }
  olddb=*curdbp;
  FrmSetActiveForm (frm);
  if (FrmDoDialog (frm) == NewDBBtOKID)
    {
      FdPtr = GetObjPtr (NewDBFDNameID);
      error = SwitchDB(curdbp,FldGetTextPtr (FdPtr), type);
      if (!error) {
	StrNCopySafe (name,FldGetTextPtr (FdPtr) , dmDBNameLength);
      }
    }
  FrmSetActiveForm (previousForm);
  FrmDeleteForm (frm);
}                 /* NewDB */


/**
   Moves selected records from current data base to another.
 Lets the user choose the new data base

@return nothing
*/

static void ChangeDB (){
  //  FieldPtr FdPtr;
  Err error=0;
  DmOpenRef curdb,newdb;
  FormPtr previousForm = FrmGetActiveForm ();
  FormPtr frm = FrmInitForm (ChangeDBForm);
  UInt32 type;
  UInt16 i, matched,category,index,c,moved=0;
  char  *name;
  ListPtr LstWptPtr;
  ControlPtr PopupWptPtr;
  char * Names[MaxDBNames];
  MemHandle oldH,RecH;

  FrmSetActiveForm (frm);
  LstWptPtr = (ListPtr) (GetObjPtr (ChangeDBListID));
  PopupWptPtr = (ControlPtr) (GetObjPtr (ChangeDBPopupID));
  category = *CurrentCategory;

  switch (Prefs.todisplay)
    {
    case DispWpt:
      curdb=WptDB;
      type=WptDBType;
      name=Prefs.CurWptDBName;
      break;
    case DispTrk:
      curdb = TrkDB;
      type=TrkDBType;
      name=Prefs.CurTrkDBName;
      break;
    case DispRte:
      curdb = RteDB;
      type=RteDBType;
      name=Prefs.CurRteDBName;
      break;
    default:
      return;
    }

  i=FindDBNames(type,GPilotSID,name,&matched,gbuf1,Names,MaxDBNames);  /* Wpt DB names */
  LstSetHeight (LstWptPtr, min (i, 6));
  LstSetListChoices (LstWptPtr, Names, i		       );
  LstSetSelection (LstWptPtr,matched);
  CtlSetLabel (PopupWptPtr,Names[matched]);

  FrmSetActiveForm (frm);
  if (FrmDoDialog (frm) == ChangeDBBtOKID)
    if(LstGetSelection (LstWptPtr)!=matched)  {
      error = OpenOrCreateDB (&newdb,Names[LstGetSelection (LstWptPtr)], type, GPilotSID);
      if (!error){
	c = 0;
	while ((RecH = DmQueryNextInCategory (curdb, &c, category)))
	  {
#ifdef SELECT
	    if (RecIsSecret (curdb, c))
	      {
		error= DmDetachRecord (curdb, c, &oldH);
		index=dmMaxRecordIndex;
		error|=DmAttachRecord (newdb, &index, oldH, NULL);
		if (error){
		  DmAttachRecord (curdb, &index, oldH, NULL);
		  break;
		}
		else moved++;
	      }
	    else 
#endif
	      c++;
	  }
      }
      DmCloseDatabase (newdb);
    }
  StrPrintF (gbuf1, "Error %d, %d records moved",error,moved);
  if (error)FrmCustomAlert (ErrorID,gbuf1," ", " ");
  FrmSetActiveForm (previousForm);
  FrmDeleteForm (frm);
}  /* ChangeDB */


/**
   Alarm callback.
 Called on alarm wakeup.

@return nothing
*/

static void
AlarmProc (void)
{
    //          EventType evt;
    SndPlaySystemSound (sndClick);
}


/**
   frmLoadEvent event handler

  @param  e pointer on event received
  @return true if event was a frmLoadEvent
 */

static Boolean
ApplicationHandleEvent (EventPtr e)
{
    FormPtr frm;
    Word formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent)
    {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm (formId);
	FrmSetActiveForm (frm);

	switch (formId)
	{
	case MainForm:
	    FrmSetEventHandler (frm, MainFormHandleEvent);
	    break;
#ifndef SKIPEDIT
	case EditForm:
	case EditRefForm:
	    FrmSetEventHandler (frm, EditFormHandleEvent);
	    break;
#endif
	case DownloadForm:
	    FrmSetEventHandler (frm, DownloadFormHandleEvent);
	    break;
	case UploadForm:
	    FrmSetEventHandler (frm, UploadFormHandleEvent);
	    break;
	case ProgressForm:
	    FrmSetEventHandler (frm, ProgressFormHandleEvent);
	    break;
	case TrackForm:
	    FrmSetEventHandler (frm, TrackFormHandleEvent);
	    break;
	case RouteForm:
	    FrmSetEventHandler (frm, RouteFormHandleEvent);
	    break;
	case HostForm:
	    FrmSetEventHandler (frm, HostFormHandleEvent);
	    break;
#ifndef GRAPHICALROUTEEDIT
	case EditRouteForm:
	    FrmSetEventHandler (frm, RteEditFormHandleEvent);
	    break;
#endif
	case MapForm:
	    FrmSetEventHandler (frm, MapFormHandleEvent);
	    break;
#ifdef SELECT
	case CategoryForm:
	    FrmSetEventHandler (frm, CatFormHandleEvent);
	    break;
#endif
	}
	handled = true;
    }
    return handled;
}							   /* ApplicationHandleEvent */


/**
   Opens or creates GPilotS data base

  @param DBP pointer were to store the DmOpenRef on the data base
  @param DBName name of data base
  @param DBType data base type
  @param creator data base creator
  @return 0 if ok
 */

static Err
OpenOrCreateDB (DmOpenRef * DBP, char *DBName, ULong DBType, ULong creator)
{
  Err error = 0;
  VoidHand h;
  UInt cardNo;
  LocalID dbID;
  LocalID appinfoID;
  AppInfoType *appInfoP;
  UInt32 type,cr;

  *DBP=NULL;
  //    *DBP = DmOpenDatabaseByTypeCreator (DBType, creator, dmModeReadWrite | dmModeShowSecret);	// ouverture/creation  DB
  cardNo=type=cr=0;
  dbID= DmFindDatabase (cardNo, DBName);
  if(dbID) error = DmDatabaseInfo (cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL,NULL,NULL,&type , &cr);
  if ((type==DBType)&&(cr==creator)) *DBP=DmOpenDatabase ( cardNo,  dbID,dmModeReadWrite | dmModeShowSecret );
  if (!*DBP)
    {
      if ((error = DmCreateDatabase (cardNo, DBName, creator, DBType, false)))
	return error;
      //	*DBP = DmOpenDatabaseByTypeCreator (DBType, creator, dmModeReadWrite | dmModeShowSecret);
      dbID= DmFindDatabase (cardNo, DBName);
      *DBP =DmOpenDatabase ( cardNo,  dbID, dmModeReadWrite | dmModeShowSecret);
    }
  if (DmOpenDatabaseInfo (*DBP, &dbID, NULL, NULL, &cardNo, NULL))
    return dmErrInvalidParam;
  if (DmDatabaseInfo
      (cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appinfoID, NULL, NULL, NULL))
    return dmErrInvalidParam;
  if (appinfoID == 0)
    {
      h = DmNewHandle (*DBP, sizeof (AppInfoType));
      if (!h)
	return dmErrMemError;
      appinfoID = MemHandleToLocalID (h);
      DmSetDatabaseInfo (cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appinfoID, NULL,
			 NULL, NULL);
      appInfoP = MemLocalIDToLockedPtr (appinfoID, cardNo);
      DmSet (appInfoP, 0, sizeof (AppInfoType), 0);
      CategoryInitialize (appInfoP, localizedAppInfo);
      MemPtrUnlock (appInfoP);
    }
  return (error);
}							   /* OpenOrCreateDB */



/* Get preferences, open (or create) app database */


/**
   Open application data bases and retrieves preferences.
 Also loads MathLib library.

@return 0 if ok
*/

static Word
StartApplication (void)
{
    Err error;
    Word prefsSize = sizeof (Prefs);

    PrefGetAppPreferences (GPilotSID, GPilotSPrefID, &Prefs, &prefsSize, true);
    error = OpenOrCreateDB (&WptDB, /* WptDBName */Prefs.CurWptDBName, WptDBType, GPilotSID);
    error |= OpenOrCreateDB (&TrkDB, /* TrkDBName */Prefs.CurTrkDBName, TrkDBType, GPilotSID);
    error |= OpenOrCreateDB (&RteDB, /* RteDBName */Prefs.CurRteDBName, RteDBType, GPilotSID);
#ifdef EVENTS
    error |= OpenOrCreateDB (&LogDB, LogDBName, LogDBType, GPilotSID);
#endif
    if (error)
	return (error);
    error = SysLibFind (MathLibName, &MathLibRef);	   /* ouverture Mathlib */
    if (error)
    {
	error = SysLibLoad (LibType, MathLibCreator, &MathLibRef);
	if (!error)
	{
	    error = MathLibOpen (MathLibRef, MathLibVersion);
	    ErrFatalDisplayIf (error, "Can't open MathLib");
	}
	else
	{
	    FrmCustomAlert (ErrorID, "MathLib missing", " ", " ");
	    return (error);
	}
    }
    SetMathLibRef (MathLibRef);
    FrmGotoForm (MainForm);
    return 0;
}							   /* StartApplication */


/**
   Save preferences, close forms, close app database

@return nothing
*/

static void
StopApplication (void)
{
    UInt usecount;
    Err error;

    debug_in (GTALKDBG_CALLS, ("->StopApplication()"));
    //    debug_close ();
    PrefSetAppPreferences (GPilotSID, GPilotSPrefID, 2, &Prefs, sizeof (Prefs), true);
    FrmSaveAllForms ();
    FrmCloseAllForms ();
    GPSEndOfOperation ();				   /* au cas ou */
    if (WptDB)DmCloseDatabase (WptDB);
    if (TrkDB)DmCloseDatabase (TrkDB);
    if (RteDB)DmCloseDatabase (RteDB);
#ifdef EVENTS
    if (LogDB)DmCloseDatabase (LogDB);
#endif

    debug_close ();  // $$
    error = MathLibClose (MathLibRef, &usecount);
    ErrFatalDisplayIf (error, "Can't close MathLib");
    if (usecount == 0)
	SysLibRemove (MathLibRef);
}							   /* StopApplication */


/**
    The main event loop

@return nothing
*/

static void
EventLoop (void)
{
    Word err;
    EventType e;

    do
    {
	if (MustCallGpslibTicks)
	    EvtGetEvent (&e, sysTicksPerSecond / 20);
	else
	    EvtGetEvent (&e, evtWaitForever);
	if (!SysHandleEvent (&e))
	    if (!MenuHandleEvent (NULL, &e, &err))
		if (!ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    }
    while (e.eType != appStopEvent);
}


/**
   Main entry point

   @param  cmd launch code
   @param  cmdPBP parameters
   @param  launchFlags flags
   @return error code
 */

DWord
PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Err err = 0;
    Boolean appIsActive = launchFlags & sysAppLaunchFlagSubCall;

    err = RomVersionCompatible (ourMinVersion, launchFlags);
    if (err)
	return err;
    if (cmd == sysAppLaunchCmdNormalLaunch)
    {
	err = StartApplication ();
	if (err)
	    return err;
	EventLoop ();
	StopApplication ();
    }
    else if ((cmd == sysAppLaunchCmdAlarmTriggered) && appIsActive)
    {
	AlarmProc ();
	((SysAlarmTriggeredParamType *) cmdPBP)->purgeAlarm = 1;	//$$
    }
#ifdef BEAM
    else if (cmd == sysAppLaunchCmdExgReceiveData)
    {
	ExgSocketPtr sockp = (ExgSocketPtr) cmdPBP;
	DmOpenRef DB = NULL;
	// This call may be made from another application, so your
	// app must not assume that it's globals are initialized.
	if (!StrCompare (sockp->name, BeamTypes[DispWpt]))
	{						   /* Wpt */
	    if (!appIsActive)
		DB = DmOpenDatabaseByTypeCreator (WptDBType, GPilotSID, dmModeReadWrite);
	    else
		DB = WptDB;
	}
	else if (!StrCompare (sockp->name, BeamTypes[DispRte]))
	{						   /* Rte */
	    if (!appIsActive)
		DB = DmOpenDatabaseByTypeCreator (RteDBType, GPilotSID, dmModeReadWrite);
	    else
		DB = RteDB;
	}

	else if (!StrCompare (sockp->name, BeamTypes[DispTrk]))
	{						   /* Trk */
	    if (!appIsActive)
		DB = DmOpenDatabaseByTypeCreator (TrkDBType, GPilotSID, dmModeReadWrite);
	    else
		DB = TrkDB;
	}

	else
	    err = -1;

	if (!err)
	    err = BeamReceive (sockp, DB);
	if (!appIsActive && DB)
	    DmCloseDatabase (DB);
	if (appIsActive)
	    MainListDirty = 1;
    }
#endif
    return err;
}							   // PilotMain


/**
   Test PalmOs version compatibility

@param  requiredVersion required version
@param  launchFlags launch flag
@return
*/

static Err
RomVersionCompatible (UInt32 requiredVersion, UInt16 launchFlags)
{
    UInt32 romVersion;
    FtrGet (sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion)
    {
	if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
	    (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
	{
	    FrmAlert (OSUnsupportedID);
	    if (romVersion < 0x02000000)
	    {
	      //		Err err;
		AppLaunchWithCommand (sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
	    }
	}
	return (sysErrRomIncompatible);
    }
    return 0;
}

#ifdef SELECT

/**
   Category change form event handler

  @param  e pointer on event received
  @return true if event handled
 */

static Boolean
CatFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    UInt c;
    static UInt category,
      newcategory;
    MemHandle RecH;

    frm = FrmGetActiveForm ();
    switch (e->eType)
    {
    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	category = newcategory = *CurrentCategory;
	if (category == dmAllCategories)
	    category = dmUnfiledCategory;
	CategoryGetName (CurrentDB, category, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (MainCatPopupID), CategoryName);
	FrmDrawForm (frm);
	handled = true;
	break;

    case frmCloseEvent:
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case CatBtOKID:
	    c = 0;
	    while ((RecH = DmQueryNextInCategory (CurrentDB, &c, category)))
	    {
		if (RecIsSecret (CurrentDB, c))
		{
		    SetCategory (CurrentDB, c, newcategory);
		    RecSetSecret (CurrentDB, c, 0);
		}
		c++;
	    }
	    MainListDirty = 1;

	case CatBtCancelID:
	    FrmReturnToForm (0);
	    break;

	case MainCatPopupID:
	    CategorySelect (CurrentDB, frm, MainCatPopupID, MainCatListID, false, &newcategory,
			    CategoryName, 1, 0);
	    break;
	}
	handled = true;
	break;

    default:
	break;
    }

    return handled;
}							   // CatFormHandleEvent


/**
   Tests if a record has secret attribute


@param  DBP pointer on an open data base
@param  index index of record to test
@return true if secret attribute set
*/

Boolean
RecIsSecret (DmOpenRef * DBP, UInt16 index)
{
    UInt attributes;
    DmRecordInfo (DBP, index, &attributes, NULL, NULL);
    return (attributes & dmRecAttrSecret);
}


/**
   Sets the secret attribute of a record

@param  DBP pointer on an open data base
@param  index index of record to set the attribute
@param  value value to put in attribute
@return error code
*/

static Err
RecSetSecret (DmOpenRef * DBP, UInt16 index, Boolean value)
{
    UInt attributes;
    Err error;

    error = DmRecordInfo (DBP, index, &attributes, NULL, NULL);
    if (error)
	return (error);
    if (value)
	attributes |= dmRecAttrSecret;
    else
	attributes &= ~dmRecAttrSecret;
    error = DmSetRecordInfo (DBP, index, &attributes, NULL);
    return (error);
}

#endif


/**
   Set the backup flag of the given data base

@param  DBP DmOpenRef to the data base
@return nothing
*/

Err SetBackupBit (DmOpenRef * DBP) {
  UInt16 cardNo, attributes;
  LocalID dbID;
  Err err;

  err=DmOpenDatabaseInfo (DBP, &dbID, NULL, NULL, &cardNo, NULL);

  err|=DmDatabaseInfo ( cardNo,  dbID,NULL , &attributes, NULL,NULL ,NULL , NULL, NULL,NULL,NULL, NULL,NULL );
if (!err){
      attributes|=dmHdrAttrBackup;
      err=DmSetDatabaseInfo ( cardNo,  dbID, NULL, &attributes,NULL,NULL ,NULL , NULL, NULL,NULL,NULL, NULL,NULL );
    }
   return(err);
}   /* SetBackupBit */


#ifdef DEBUG
static char * GetCharEncoding(){
  UInt32 encoding; 

  if (FtrGet(sysFtrCreator, sysFtrNumEncoding, &encoding) != 0) encoding = charEncodingPalmLatin; //default encoding 
  return(TxtGlueEncodingName(encoding));
}   /* GetCharEncoding */


/**
   Debug form handler.
 Displays a form to select debug levels of differents parts of GPilotS

@return nothing
*/

static void
SetDebug ()
{
  FormPtr previousForm = FrmGetActiveForm ();
  FormPtr frm = FrmInitForm (DebugForm);
  ControlPtr ControlP;
  UInt16 i, gpslib_level, gpsutil_level;

  FrmSetActiveForm (frm);
  /*
    for (i = 0; i < 8; i++)
    {
    ControlP = GetObjPtr (DbgGpsl1ID + i);
    CtlSetValue (ControlP, (DebugLevel >> i) & 0x1);
    }
  */
  if (FrmDoDialog (frm) == DebugBtOKID)
    {
      gpslib_level= gpsutil_level=DebugLevel=0;
      for (i = 0; i < 8; i++)
	{
	  ControlP = GetObjPtr (DbgGpsl1ID - i);
	  gpslib_level += CtlGetValue (ControlP) << i;
	}
      for (i = 0; i < 8; i++)
	{
	  ControlP = GetObjPtr (DbgGpsu1ID - i);
	  gpsutil_level += CtlGetValue (ControlP) << i;
	}
      for (i = 0; i < 8; i++)
	{
	  ControlP = GetObjPtr (Dbg1ID - i);
	  DebugLevel += CtlGetValue (ControlP) << i;
	}
      if (gpslib_level+gpsutil_level+DebugLevel) debug_open ();
      else debug_close();
      if (gpslib_level)  GpslibSetDebugLevel (gpslib_level, log_write);
      if (gpsutil_level) GpsutilSetDebugLevel (gpsutil_level, log_write);
      if(DebugLevel){	
	debug (DebugLevel, ("%s, debug level = %d", rcsid, DebugLevel));
	debug (DebugLevel, ("Char encoding : %s",GetCharEncoding()));
      }
    }
  FrmSetActiveForm (previousForm);
  FrmDeleteForm (frm);
}							   /* SetDebug */

#endif
