/** main include file
  @file
  @author C. Claveleira
  @date 1999-2004
*/

/*
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   $Id: gtalk.h,v 1.22 2004/11/17 21:06:01 clavelei Exp $ 

*/


#include "types.h"

#define GPilotSID 'GPil'				   // creator ID
#define WptDBType 'Wpts'				   // type data base de wpts
#define TrkDBType 'Trks'				   // type data base de rtackpoints
#define RteDBType 'Rtes'				   // type data base de routes
#define LogDBType 'Logs'				   // type data base de paquets
#define WptDBName "WptDB-GPil"				   // nom data base de wpts
#define TrkDBName "TrkDB-GPil"				   // nom data base de trackpoints
#define RteDBName "RteDB-GPil"				   // nom data base de routes
#define LogDBName "LogDB-GPil"				   // nom data base de paquets
#define GPilotSPrefID 0					   // preferences ID

#define DBVersion 2					   // version number of data bases

#define MaxRecordSize 62000				   // taille max que l'on peut raisonablement esprer loger

// wrapper pour compatibilite 3.0 :
#define WinDrawChar(theChar, x, y) do{char mw_buf[1];mw_buf[0]=(theChar);WinDrawChars (mw_buf, 1 , (x),  (y));}while(0)

/* Debugging : */
#define GTALKDBG_INFOS 128				   // informational
#define GTALKDBG_CALLS 32				   // procedures calls/exits
#define GTALKDBG_UNEXP 8				   // unexpected things

#define LATMAX 0x3fffffffL				   // latitude max (90)
#define LATMIN 0xC0000000L				   // latitude min (-90)
#define LONMAX 0x7fffffffL				   // longitude max (180)
#define LONMIN 0x80000000L				   // longitude min (-180)
#define LONMAX23 0x55555554L				   // 2*180/3
#define LONMIN23 0xAAAAAAACL				   // 2*-180/3
#define DELTALONMAX 0X38E38E38				   // 80

/** Item type to display in main form
 */
enum ItemToDisplay
{
    DispUnknown, DispWpt, DispRte, DispTrk
};
typedef enum ItemToDisplay ItemToDisplay;

/** waypoint sort criteria enum
 */
enum WptSortCritType
{
    Alpha, Dist, Lat, Lon, Symbol, Alt, Bear, Current
};
typedef enum WptSortCritType WptSortCritType;	/* critere de tri */

/** Preferences structure
 */
typedef struct
{
    ItemToDisplay todisplay;	       /* dernier type donnes affiches (bytes 1-2) */
    UInt logcategory;		       /* last log category    (bytes 3-4) */
    Boolean allowduplicate;	       /* +sieurs noms identiques autoriss ? (byte 5) */
    WptSortCritType wptsortcrit;       /* critere de tri (byte 6) */
    DistanceUnit distunit;	       /* unit d'affichage de distance (byte 7) */
    DisplayFormat displaymode;	       /* mode d'affichage               */
    signed int utcoffset;	       /* offset heure locale <-> UTC (mn)     */
    unsigned int datum;		       /* datum index                     */
    unsigned long refptuid;	       /* uid du point de reference       */
    unsigned int version;	       /* version number                  */
    Personality emulatedgps;	       /* emulated GPS                   */
    unsigned int loginterval;	       /* logging interval in s              */
    UInt wptcategory;		       /* last waypoint category */
    UInt trkcategory;		       /* last track category */
    UInt rtecategory;		       /* last route category */
    Boolean selectmode;	               /* selection mode */
    UInt32 portid;	               /* serial port id */
    char CurWptDBName [dmDBNameLength]; /* current waypoint data base name */
    char CurRteDBName [dmDBNameLength]; /* current route data base name */
    char CurTrkDBName [dmDBNameLength]; /* current track data base name */
    Word PermTrackIndex;	       /* permanent track index + 1 */
    Boolean looseprotocol;             /* relaxed protocol */
}
PrefType;
extern PrefType Prefs;

extern DmOpenRef CurrentDB,
  WptDB,
  TrkDB,
  RteDB
#ifdef EVENTS
  ,LogDB
#endif
;

extern UInt CurrentRecordIndex;
extern UInt *CurrentCategory;
extern Semicircle_Type CurrentWGS84Pos;
extern char ReferencePtName[];
extern  char * BeamTypes[];
extern Semicircle_Type ReferencePt;
extern Boolean OnlyNews,
  Furuno,
  CompactTracks;

extern Boolean MainListDirty;


enum EditWptMode
{
    EditWpt, NewWpt, MarkWpt, FromRefWpt
};
typedef enum EditWptMode EditWptMode;  /* type d'operation d'edition */
extern EditWptMode EditMode;
extern const char *DistanceUnitNames[];      /* defini dans gtalk.c */
extern char gbuf1[512],
  gbuf2[128];
extern Boolean MustCallGpslibTicks;

typedef struct
{
    unsigned int reserved:16;
    unsigned int length:16;	       /* length of next object */
}
RecordSeparator;

#include "sections.h"

/* Code d'erreurs : */
#define GErrNoError 0
#define GErrUnsupportedFmt appErrorClass + 1
#define GErrProtocolError appErrorClass + 2
#define GErrNoHandle appErrorClass + 3

/* prototypes : */

/* editwpt.c */

extern Boolean EditFormHandleEvent (EventPtr e) EDIT_SECTION;

/* gtalk.c */
extern Int CompareWpt (void *p1, void *p2, Int other, SortRecordInfoPtr sp1,
		       SortRecordInfoPtr sp2, VoidHand aph);
extern void SortWpts (WptSortCritType s);
extern long SearchWptByName (char *name, Boolean exact, Boolean all);
extern int SetReferencePt (UInt index, Boolean ask);
//extern Boolean DrawRoute (UInt index, Boolean rescale, Semicircle_Type * position);
extern void DrawSymbol (unsigned int lid, int x, int y);
extern void MainListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText);
extern void SetPreferences (void);
extern Boolean RecIsSecret (DmOpenRef * DBP, UInt16 index);
extern Err SetBackupBit (DmOpenRef * DBP);
