/** Logging module for PalmOS
  @file 
  @author C. Claveleira
  @date 2000-2004
  
*/

/* 	$Id: log.c,v 1.7 2004/11/17 20:51:09 clavelei Exp $	 */


#include <PalmOS.h>
//#include <PalmCompatibility.h>
#include <stdarg.h>
#include "log.h"
#include "memo.h"

#define SIZE_INC 1000

static UInt32 ticks,
  secs;
static Boolean LogOpened;

/*  const */ char *EvtNames[] = {
    "nilEvent",
    "penDownEvent",
    "penUpEvent",
    "penMoveEvent",
    "keyDownEvent",
    "winEnterEvent",
    "winExitEvent",
    "ctlEnterEvent",
    "ctlExitEvent",
    "ctlSelectEvent",
    "ctlRepeatEvent",
    "lstEnterEvent",
    "lstSelectEvent",
    "lstExitEvent",
    "popSelectEvent",
    "fldEnterEvent",
    "fldHeightChangedEvent",
    "fldChangedEvent",
    "tblEnterEvent",
    "tblSelectEvent",
    "daySelectEvent",
    "menuEvent",
    "appStopEvent",
    "frmLoadEvent",
    "frmOpenEvent",
    "frmGotoEvent",
    "frmUpdateEvent",
    "frmSaveEvent",
    "frmCloseEvent",
    "frmTitleEnterEvent",
    "frmTitleSelectEvent",
    "tblExitEvent",
    "sclEnterEvent",
    "sclExitEvent",
    "sclRepeatEvent",
    "tsmFepModeEvent"  /* 35 */
};


int DebugIndentLevel;

/**
   Opens a new log record in memo pad data base

@return nothing
 */

Err
log_open ()
{
    DateTimeType dt;
    static UInt32 dateformat,
      timeformat;
    char dates[longDateStrLength],
      times[timeStringLength],
      buf[128];

    if (LogOpened)
	return (0);

    dateformat = PrefGetPreference (prefDateFormat);
    timeformat = PrefGetPreference (prefTimeFormat);
    TimSecondsToDateTime (TimGetSeconds (), &dt);
    DateToAscii (dt.month, dt.day, dt.year /*   + 1904 */ , dateformat, dates);
    TimeToAscii (dt.hour, dt.minute, timeformat, times);
    StrPrintF (buf, "\\Log record of %s %s\n", dates, times);
    ticks = TimGetTicks ();
    secs = TimGetSeconds ();

    return (memo_open (buf));
}


/**
   Closes the current log

@return nothing
 */

void
log_close ()
{
    log_write ("Log closed");
    memo_close (NULL);
    LogOpened = 0;
}


/**
   Closes and deletes the curent log 

@return nothing
 */

void
log_delete ()
{
    memo_delete ();
}


/**
   Writes the string given in the log, preceded by a time stamp and a newline


@param p string to log
@return nothing
 */

void
log_write (char *p)
{
    char buf[96];

	if (p[0] != '\\')
	buf[0] = '\n';
    else
	buf[0] = ' ';
    //      StrIToA (buf + 1, TimGetTicks () - ticks);
    StrPrintF (buf + 1, "(%ld)%ld: ", TimGetSeconds () - secs, TimGetTicks () - ticks);

    memo_write (buf);
    memo_write (p);
}


/**
   Internal use. Writes "printf" formatted messages to debug log


@param  formatStr Pointer to the format specification string
@param ... parameters
@return nothing
 */

void
log_printf (Char* formatStr, ...)
{
    va_list args;
    char buf[256];
    Int16 l = 0;

#ifdef DEBUG
    if (DebugIndentLevel < 0)
	DebugIndentLevel = 0;
    else if (DebugIndentLevel > 20)
	DebugIndentLevel = 20;
    l = DebugIndentLevel;
    MemSet (buf, DebugIndentLevel, ' ');
#endif
    va_start (args, formatStr);
    l = StrVPrintF (buf + l, formatStr, args);
    va_end (args);
    ErrFatalDisplayIf (l >= sizeof (buf), "log_printf overflow");
    log_write (buf);
}
