/**  Graphical display part of GPilotS
  @file objects.c
  @author C. Claveleira
  @date 1999-2004
 
  Manages all graphical representations of waypoints, routes and tracks
  with zooming, panning, scale,...
  

*/

/* 
   Copyright (C) 2001-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
 */

/* 	$Id: map.c,v 1.9 2004/11/17 21:50:12 clavelei Exp $	 */

#ifndef lint
//static char vcid[] = "$Id: map.c,v 1.9 2004/11/17 21:50:12 clavelei Exp $";
#endif /* lint */

#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS		   // dirty !  $$
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "gtalkRsc.h"
#include "latlong.h"
#include "MathLib.h"
#include "gpslib.h"
#include "gtalk.h"
#include "gpsutil.h"
#include "map.h"
#include "log.h"
#include "dbio.h"
#include "route.h"
#include "memo.h"

MapFlagsType MapFlags;
RectangleType DrawRect;		       /* draw rectangle in graphic window */

static WinHandle OffScreenWH;	       /* offscreen work window handle */
static Boolean DrawRectCleared;	       /* set if draw rectangle cleared */
static Semicircle_Type P0;	       /* top left displayed coord. in graphic window */
static Int32 LonByPixel,		       /* longitude coefficient (y) in graphic window */

  LatByPixel;			       /* latitude coefficient (x) in graphic window */
static UInt16 ActTrkIndex;
static UInt16 CurrentTrackIndex;

#define BottomStatPos 2, 146				   /* bottom status display position */

static void DrawGrid (void)    MAP_SECTION;
     static void DrawWaypoints (UInt16 category, Boolean rescale) MAP_SECTION;
     static Int32 MaxLonDisplayed (void) MAP_SECTION;
     static Int32 MaxLatDisplayed (void) MAP_SECTION;
     static void DisplayWpt (Custom_Wpt_Type * wp, Boolean invert) MAP_SECTION;
     static void RefreshDisplay (Boolean Erase) MAP_SECTION;
     static void DisplayPosition (Semicircle_Type * ptp) MAP_SECTION;
     static void DrawCalibrate (Int32 latmin, Int32 latmax, Int32 lonmin, Int32 lonmax) MAP_SECTION;
     static void ReCenterDisplay (Semicircle_Type * NewCenterPos) MAP_SECTION;
     static Boolean WptClicked (WordPtr index, PointType point) MAP_SECTION;
     static void MapFormClean(void) MAP_SECTION;
#ifdef GRAPHICALROUTEEDIT
     static Boolean RteWptClicked (Word rteindex, WordPtr wptindexp, PointType point) MAP_SECTION;
#endif

/**
   Macro to test if a waypoint is in the drawing area

   @param wp waypoint to test
 */

#define WptVisible(wp) (((wp).posn.lat >= P0.lat) && \
      ((wp).posn.lat <= Latmax) && \
      ((wp).posn.lon >= P0.lon) && \
      ((wp).posn.lon <= Lonmax))


/**
   Macro to test if a coordinate is visible

  @param pos coordinate
*/

#define PosVisible(pos) (((pos).lat >= P0.lat) && \
      ((pos).lat <= MaxLatDisplayed ()) && \
      ((pos).lon >= P0.lon) && \
      ((pos).lon <= MaxLonDisplayed ()))


/**
   Macro to convert from latitude/longitude to x/y of current window

   @param p Semicircle_Type point (from)
   @param pix PointType point (to)
 */

#define LatLonToXY(p,pix) pix.x=DrawRect.topLeft.x+((p).lon - P0.lon)/LonByPixel;\
	   pix.y=DrawRect.topLeft.y+DrawRect.extent.y-((p).lat -P0.lat)/LatByPixel


/**
   Macro to convert from x/y coordinates in current window to latitude/longitude

   @param pix PointType point to convert from
   @param pSemicircle_Type point to convert to
 */

#define XYToLatLon(pix,p) (p).lon= (pix.x-DrawRect.topLeft.x)*LonByPixel+P0.lon;\
          (p).lat=(DrawRect.topLeft.y+DrawRect.extent.y-pix.y)*LatByPixel+P0.lat


/**
   Reads and sets the backlight status on systems pre-4.0

@param  set if true the backlight will be set
@param  newState the new state of backlight desired if set is true
@return current state of the backlight
*/

static Boolean HwrBacklight(Boolean set, Boolean newState) SYS_TRAP(sysTrapHwrBacklightV33);
/* static UInt8 SysLCDBrightness(Boolean set, UInt8 newBrightnessLevel) */
/*        SYS_TRAP(sysTrapSysLCDBrightness); */
     
static Boolean BacklightIsOn(void) {
  UInt32 glRomVersion;

  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &glRomVersion);
  if (glRomVersion < sysMakeROMVersion(5,0,0,sysROMStageRelease,0)) {
    return HwrBacklight(false, false);
  }
  else {
    UInt8 brightnessLevel = SysLCDBrightness(false, 0);
    return brightnessLevel ? true : false;
  }
}

#define hwrDispBacklight 0x12
// this function works for get/set in 3.x-4.x and for getting in 5.x
void HwrDisplayAttributes(Boolean set, UInt8 kind, void *val)  SYS_TRAP(sysTrapHwrDisplayAttributes);

/**
   Reads and sets the backlight status

@param  set if true the backlight will be set
@param  newState the new state of backlight desired if set is true
@return current state of the backlight
*/

static Boolean HwrBacklight2(Boolean set, Boolean newState){
  UInt32 glRomVersion;
  Boolean oldState;

  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &glRomVersion);
  debug (GTALKDBG_INFOS, ("HwrBacklight2 : P1, glRomVersion=%lx",glRomVersion));
  if (glRomVersion < sysMakeROMVersion(4,0,0,sysROMStageRelease,0)) {
    oldState=HwrBacklight(false, false);
  }
  else    HwrDisplayAttributes(false, hwrDispBacklight, &oldState); // always works -> non !
  debug (GTALKDBG_INFOS, ("HwrBacklight2 : P2"));
  if (set){
    if ((oldState && !newState) || (!oldState && newState)){
      // enqueue request for backlight toggle
      EvtEnqueueKey(vchrBacklight, 0, commandKeyMask);
    }
  }
  return oldState;
}


/**
   Do beep-beep

   @param repeat number of times to repeat beep
   @return nothing
 */

static void BeepBeep (UInt16 repeat)
{
  SndCommandType cmd;

  for (; repeat; repeat--)
    {
      cmd.cmd = sndCmdFreqDurationAmp;
      cmd.param1 = 1200;
      cmd.param2 = 100;
      cmd.param3 = sndMaxAmp;
      SndDoCmd (NULL, &cmd, 0);
      cmd.param1 = 1000;
      cmd.param2 = 600;
      SndDoCmd (NULL, &cmd, 0);
      SysTaskDelay (sysTicksPerSecond / 2);
    }
}							   /* BeepBeep */



/**
   Draws a circle. It needs an offscreen window whose handle is in OffScreenWH

   @param x x coordinate of center
   @param y y coordinate of center
   @param r radius
   @return nothing
 */

static void
Circle (UInt16 x, UInt16 y, UInt16 r)
{
    RectangleType bounds = { {x - r, y - r}, {2 * r, 2 * r} }, bounds2 =
    {
	{
	bounds.topLeft.x + 1, bounds.topLeft.y + 1}
	,
	{
	bounds.extent.x - 2, bounds.extent.y - 2}
    };
    WinHandle CurWH = WinGetDrawWindow ();

    if (!OffScreenWH)
	return;
    WinSetDrawWindow (OffScreenWH);
    WinEraseWindow ();
    WinDrawRectangle (&bounds, r);
    /*      bounds.topLeft.x++; */
    /*      bounds.topLeft.y++; */
    /*      bounds.extent.x -= 2; */
    /*      bounds.extent.y -= 2; */
    WinInvertRectangle (&bounds2, r - 2);
    WinCopyRectangle (OffScreenWH, CurWH, &bounds, x - r, y - r, scrOR /*  scrAND */ );
    WinSetDrawWindow (CurWH);
}							   /* Circle */


/**
   (re-)draw a bitmap in DrawRect
   The previous bitmap display is erased if necessary before drawing

@param  BitmapID ressource id of the bitmap to draw
@param  x x coordinate to draw
@param  y y coordinate to draw
@return nothing
 */

static void
RedrawBitmap (int BitmapID, SWord x, SWord y)
{
    static SWord px,
      py,
      bw,
      bh;
    static int pbitmapid;
    WinHandle CurWH;
    RectangleType r;
    MemHandle ResH;
    BitmapPtr BitP;

    if (!OffScreenWH)
	return;
    WinSetClip (&DrawRect);
    CurWH = WinGetDrawWindow ();
    r.topLeft.x = 0;
    r.topLeft.y = 0;
    r.extent.x = bw;
    r.extent.y = bh;
    if (!DrawRectCleared)
    {							   /* effacement precedent dessin ? */
	WinCopyRectangle (OffScreenWH, CurWH, &r, px, py, scrXOR /*  scrAND */ );
	WinDrawLine (px + bw / 2, py + bh / 2, x, y);
    }
    DrawRectCleared = 0;

    ResH = DmGetResource ('Tbmp', BitmapID);
    if (ResH)
    {
	BitP = MemHandleLock (ResH);
	bh = BitP->height;
	bw = BitP->width;
	WinSetDrawWindow (OffScreenWH);
	WinDrawBitmap (BitP, 0, 0);
	WinSetDrawWindow (CurWH);
	MemHandleUnlock (ResH);
	DmReleaseResource (ResH);
	r.extent.x = bw;
	r.extent.y = bh;
	pbitmapid = BitmapID;
    }
    x -= bw / 2;					   /* recentrage  */
    y -= bh / 2;
    WinCopyRectangle (OffScreenWH, CurWH, &r, x, y, scrXOR /*  scrAND */ );
    px = x;
    py = y;
    WinResetClip ();
}							   /* RedrawBitmap */


/**
   Draws a latitude-longitude grid at integer values and a distance scale

   @return nothing
 */

static void
DrawGrid ()
{							   /* quadrillage */
#define CENTMINUTES 19884108
#define CENTKM      10736559
#define CENTMI      17278774				   // a verifier $$

    Int32 deltalon = LonByPixel * DrawRect.extent.x,	/* longitude couverte */
	//    deltalat = LatByPixel * DrawRect.extent.y, /* latitude couverte */
	//      lonmax = MaxLonDisplayed (),

      latmax = MaxLatDisplayed (),
      pas,			       /* pas de quadrillage */

      lat0,
      lon0,			       /* origine quadrillage */

      inc,
      unit,
      coef;
    UInt16 n;
    Semicircle_Type ll;		       /* current lat/lon */
    PointType p;

    WinSetClip (&DrawRect);
    unit = Prefs.distunit == metric ? CENTKM : Prefs.distunit == nautical ? CENTMINUTES : CENTMI;
    for (coef = 10, pas = unit; deltalon / pas <= 1; coef *= 10)
	pas = unit / coef;
    lon0 = (P0.lon / pas) * pas;			   /* lat/lon depart quadrillage */
    lat0 = (P0.lat / pas) * pas;
    ll.lat = lat0;
    for (ll.lon = lon0, inc = 0; inc <= deltalon; ll.lon += pas, inc += pas)
    {
	LatLonToXY (ll, p);
	WinDrawGrayLine (p.x, DrawRect.topLeft.y, p.x, DrawRect.topLeft.y + DrawRect.extent.y);
    }
    ll.lon = lon0;
    for (ll.lat = lat0; ll.lat < latmax; ll.lat += pas)
    {
	LatLonToXY (ll, p);
	WinDrawGrayLine (DrawRect.topLeft.x, p.y, DrawRect.topLeft.x + DrawRect.extent.x, p.y);
    }

    n = 1000 / coef;					   /* echelle de distance */
    if (n)
	StrPrintF (gbuf1, "%d %s", n, DistanceUnitNames[Prefs.distunit]);	/* >= 1 nm */
    else
    {							   /* < 1 nm */
	StrPrintF (gbuf1, "0,");
	for (n = coef / 100000; n; n /= 10)
	    StrNCatSafe (gbuf1, "0", sizeof (gbuf1), 1);
	//      StrNCatSafe (gbuf1, "1 nm", sizeof (gbuf1), 4);
	StrPrintF (gbuf1 + StrLen (gbuf1), "1 %s", DistanceUnitNames[Prefs.distunit]);
    }
    WinDrawChars (gbuf1, StrLen (gbuf1),
		  DrawRect.topLeft.x + DrawRect.extent.x - 2 - FntCharsWidth (gbuf1,
									      StrLen (gbuf1)),
		  DrawRect.topLeft.y + 2);
    //WinDrawLine (DrawRect.topLeft.x + DrawRect.extent.x, DrawRect.topLeft.y+10, DrawRect.topLeft.x + DrawRect.extent.x-(pas/LonByPixel), DrawRect.topLeft.y+10);
    WinResetClip ();
}							   /* DrawGrid */



/**
      Drawing of given category waypoints on drawing rectangle

   @param  category category of waypoints to display
   @param  rescale if true the graphic display is rescaled to fit exactly all waypoints
   @return nothing
 */

static void
DrawWaypoints (UInt16 category, Boolean rescale)
{							   /* trace des waypoints  */
    Custom_Wpt_Type wp;
    MemHandle font129H;
    Err error;
    UInt16 i,
      pos;
    void *font129p;
    FontID oldfont;
    Int32 Lonmax = MaxLonDisplayed (),
      Latmax = MaxLatDisplayed ();
    //              Semicircle_Type position;
    Int32 latmin = LATMAX,
      latmax = LATMIN,
      lonmin = LONMAX,
      lonmax = LONMIN;

    font129H = DmGetResource ('NFNT', MidgetXS0FontRSCID);
    font129p = MemHandleLock (font129H);
    FntDefineFont (MidgetXS0FontID, font129p);
    //    oldfont = UICurrentFontPtr;                              // $$  
    oldfont = FntSetFont (MidgetXS0FontID);
    //    UICurrentFontPtr = font129p;                             // $$  

    WinSetClip (&DrawRect);
    pos = 0;
    for (i = DmNumRecordsInCategory (WptDB, category); i > 0; i--)
    {
	DmQueryNextInCategory (WptDB, &pos, category);
	IndexToCustomWpt (pos, WptDB, &wp);
	pos++;
	if (rescale)
	{
	    if (wp.posn.lat < latmin)
		latmin = wp.posn.lat;
	    else if (wp.posn.lat > latmax)
		latmax = wp.posn.lat;
	    if (wp.posn.lon < lonmin)
		lonmin = wp.posn.lon;
	    else if (wp.posn.lon > lonmax)
		lonmax = wp.posn.lon;
	    continue;
	}
	if (WptVisible (wp))
	    DisplayWpt (&wp, 0);
    }							   /* for */
    WinResetClip ();
    //    UICurrentFontPtr = oldfont;                              // $$
    FntSetFont (oldfont);
    MemHandleUnlock (font129H);
    error = DmReleaseResource (font129H);
    if (rescale)
    {
	DrawCalibrate (latmin, latmax, lonmin, lonmax);
	DrawWaypoints (category, 0);
    }
}							   /* DrawWaypoints */


/**
   Draws the given route or test if given position is in proximity distance
 of one of it's waypoints

   @param  index record index of route to draw/test
   @param  rescale if !=0 the graphic display is rescaled to fit entire route
   @param  position if not null, test if position is inside proximity distance of one of wpt
   @return true if position at proximity of one wpt
 */

Boolean
DrawRoute (UInt16 index, Boolean rescale, Semicircle_Type * position)
{
    Custom_Wpt_Type wpt;
    MemHandle						   /*  ResH, */
	RecH = NULL;
    void *RecP;
    UInt16 n,
      i;
    LocalFormatID fmt;
    //    BitmapPtr BitP;
    PointType point,
      oldpoint = { 0, 0 };
    Boolean inside = 0;
    Int32 latmin = LATMAX,
      latmax = LATMIN,
      lonmin = LONMAX,
      lonmax = LONMIN;

    debug (GTALKDBG_CALLS, ("<->DrawRoute(%d)", (int) rescale));
    WinSetClip (&DrawRect);
    RecH = DmQueryRecord (RteDB, index);
    if (RecH)
    {
	n = WptsInRteDB (RecH);
	RecP = MemHandleLock (RecH);
	for (i = 1; i <= n; i++)
	{
	    fmt = PtrToCustomWpt ((Object *) (RecP + OffsetOfObjectInRec (RecH, i)), &wpt);
	    if (fmt)
	    {						   /* wpt ? */
		if (position)
		{
		    inside |= WptHit (&wpt, position);
		    continue;
		}
		if (wpt.posn.lat < latmin)
		    latmin = wpt.posn.lat;
		else if (wpt.posn.lat > latmax)
		    latmax = wpt.posn.lat;
		if (wpt.posn.lon < lonmin)
		    lonmin = wpt.posn.lon;
		else if (wpt.posn.lon > lonmax)
		    lonmax = wpt.posn.lon;

		if (rescale)
		    continue;
		DisplayWpt (&wpt, 0);
		LatLonToXY (wpt.posn, point);
		if (i > 1)
		    WinDrawLine (oldpoint.x, oldpoint.y, point.x, point.y);
		oldpoint.x = point.x;
		oldpoint.y = point.y;
	    }
	}
	MemHandleUnlock (RecH);
    }
    WinResetClip ();
    if (rescale && !position)
    {
	DrawCalibrate (latmin, latmax, lonmin, lonmax);
	DrawRoute (index, 0, NULL);
    }
    return (inside);
}							   /* DrawRoute */


/**
   Graphic scale determination given coordinates of area

   @param  latmin latitude min to draw
   @param  latmax latitude max to draw
   @param  lonmin longitude min to draw
   @param  lonmax longitude max to draw
   @return nothing
 */

static void
DrawCalibrate (Int32 latmin, Int32 latmax, Int32 lonmin, Int32 lonmax)
{
    double coef,
      LatBP,
      LonBP;

    debug (GTALKDBG_CALLS, ("<->DrawCalibrate(%ld,%ld,%ld,%ld)", latmin, latmax, lonmin, lonmax));
    if (latmin < -954437180L)
	latmin = -954437180L;				   /* 80 */
    if (latmax > 954437180L)
	latmax = 954437180L;
    P0.lat = latmin;					   /*  origine */
    P0.lon = lonmin;
    coef = (latmax + latmin) / 2 * PI / OnePower31;	   /* lat moyenne en radian */
    coef = cos (coef);
    if (coef < 0)
	coef = -coef;

    LonBP = (lonmax - lonmin) / DrawRect.extent.x;
    if (LonBP < 10)
	LonBP = 10;
    LatBP = LonBP * coef;

    if (((latmax - latmin) / LatBP) > DrawRect.extent.y)
    {							   /* on depasse en latitude ? */
	LatBP = (latmax - latmin) / DrawRect.extent.y;
	if (LatBP < 10)
	    LatBP = 10;
	LonBP = LatBP / coef;
	P0.lon = (lonmax / 2 + lonmin / 2) - (DrawRect.extent.x * LonBP) / 2;	/* recentrage */
    }
    else
	P0.lat = (latmax / 2 + latmin / 2) - (DrawRect.extent.y * LatBP) / 2;	/* recentrage */
    LonByPixel = LonBP;
    LatByPixel = LatBP;
}							   /* DrawCalibrate */


/**
   
   Drawing of a track in the drawing rectangle
  
   @param  index index of the track to draw
   @param  rescale if true the graphic display is rescaled to fit entire track
   @param  Clear if true the drawing rectangle is cleared before drawing
   @return nothing
 */

void
DrawTrack (UInt16 index, Boolean rescale, Boolean Clear)
{
    Object *RecP;
    Custom_Trk_Point_Type *ht;
    Compact_Trk_Point_Type *cht;
    PointType CurrentPoint,
      PreviousPoint = { -1, 0 };
    UInt16 n;
    MemHandle RecH;

    debug (GTALKDBG_CALLS, ("<->DrawTrack()"));
    if (Clear)
    {
	WinEraseRectangle (&DrawRect, 0);
	DrawRectCleared = 1;
    }
    //    else WinSetPattern(grayPattern);
    RecH = DmQueryRecord (TrkDB, index);
    if (!RecH)
	return;
    n = TrkptsInTrkDB (RecH);
    RecP = MemHandleLock (RecH);
    if (rescale)
	DrawCalibrate (RecP->d.CustTrkHdr.latmin, RecP->d.CustTrkHdr.latmax,
		       RecP->d.CustTrkHdr.lonmin, RecP->d.CustTrkHdr.lonmax);
    WinSetClip (&DrawRect);
    if (RecP->d.CustTrkHdr.type == CustomTrkPt)
    {
	ht = (Custom_Trk_Point_Type *) ((void *) RecP + SizeOfObject (RecP));
	for (; n > 0; n--, ht++)
	{
	    LatLonToXY (ht->posn, CurrentPoint);
	    if (ht->new_trk || (PreviousPoint.x < 0))
		WinDrawLine (CurrentPoint.x, CurrentPoint.y, CurrentPoint.x, CurrentPoint.y);
	    else
		WinDrawLine (PreviousPoint.x, PreviousPoint.y, CurrentPoint.x, CurrentPoint.y);
	    //else WinFillLine(PreviousPoint.x, PreviousPoint.y, CurrentPoint.x, CurrentPoint.y);
	    PreviousPoint.x = CurrentPoint.x;
	    PreviousPoint.y = CurrentPoint.y;
	}
    }
    else
    {
	cht = (Compact_Trk_Point_Type *) ((void *) RecP + SizeOfObject (RecP));
	for (; n > 0; n--, cht++)
	{
	    LatLonToXY (cht->posn, CurrentPoint);
	    if (cht->new_trk || (PreviousPoint.x < 0))
		WinDrawLine (CurrentPoint.x, CurrentPoint.y, CurrentPoint.x, CurrentPoint.y);
	    else
		WinDrawLine (PreviousPoint.x, PreviousPoint.y, CurrentPoint.x, CurrentPoint.y);
	    PreviousPoint.x = CurrentPoint.x;
	    PreviousPoint.y = CurrentPoint.y;
	}
    }
    MemHandleUnlock (RecH);
    WinResetClip ();
}							   /* DrawTrack */


/**
   Set alarm time

@param  alarmTime alarm time
@param  ref Caller-defined value
@return nothing 
 */

static void
SetTimeOfNextAlarm (UInt32 alarmTime, DWord ref)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    DmGetNextDatabaseByTypeCreator (true, &searchInfo,
				    sysFileTApplication, GPilotSID, true, &cardNo, &dbID);
    if (AlmSetAlarm (cardNo, dbID, ref, alarmTime, true))
	SndPlaySystemSound (sndError);
}							   /* SetTimeOfNextAlarm */


/**
   Map form event handler

  @param  e pointeur on event received
  @return true if event handled
 */

Boolean
MapFormHandleEvent (EventPtr e)
{
    static FormPtr frm;
    static Boolean atProximity;
    static UInt32 NextLogTime;
    static ListPtr LstPtr;
    static UInt16 RteWptIndex,
      OldRouteIndex;
    static int SelectedRteWptIndex;
    static Boolean BackLightOn;
    Custom_Rte_Hdr_Type RteHdr;
    Err error;
    ControlPtr PopupPtr;
    Boolean scaled,
      handled = false;
    UInt16 i;

    if (e->eType != nilEvent) debug_in (GTALKDBG_CALLS, ("->MapFormHandleEvent(%s)", e->eType > LASTVALIDEVTNAME ?StrIToA(gbuf1,e->eType):EvtNames[e->eType]));
    switch (e->eType)
    {
    case nilEvent:
	GpslibTicks ();
	break;

    case frmOpenEvent:
	frm = FrmGetActiveForm ();
	handled = true;
	GetObjectBounds (MapDrawingID, &DrawRect);	   /* recup bound box du gadget */
	FrmDrawForm (frm);				   // position a voir
	WinDrawRectangleFrame (simpleFrame, &DrawRect);
	if (!OffScreenWH)
	    OffScreenWH = WinCreateOffscreenWindow (160, 160, screenFormat, &error);	/* offscreen work window for bitmaps */
	CurrentTrackIndex = 0;
	scaled = 0;
	SelectedRteWptIndex = -1;

	if (Prefs.PermTrackIndex)
	{						   /* background track */
	    DrawTrack (Prefs.PermTrackIndex - 1, 1, 0);
	    scaled = 1;
	}

	if (MapFlags.AllowedOps & MapOpWptDisp)
	{						   /* drawing of current waypoints */
	    MapFlags.DrawWpts = 1;
	    if (!(MapFlags.AllowedOps & MapOpRteEdit))
	    {
		DrawWaypoints (Prefs.wptcategory, !scaled);
		scaled = 1;
	    }
	}
	else
	    MapFlags.DrawWpts = 0;

	if (MapFlags.AllowedOps & MapOpActiveRte)
	{						   /* drawing of active route */
	    MapFlags.DrawRtes = 1;
	    DrawRoute (CurrentRouteIndex, !scaled, NULL);
	    scaled = 1;
	    ShowObject(MapRteChk);
	}
	else
	{
	  // HideObject (MapRteChk);
	    MapFlags.DrawRtes = 0;
	}

	if (MapFlags.AllowedOps & MapOpTrack)
	{						   /* real time tracking */
	    MapFlags.DrawActTrk = 1;
	    ActTrkIndex = init_active_trk ();		   // $$
	    //      DrawWaypoints (Prefs.wptcategory, 1);
	    ShowObject(MapTrkChk);
	    ShowObject(MapLogChk);
	}
	else
	{
	  // HideObject (MapTrkChk);
	    //	    HideObject (MapLogChk);
	}

	if (MapFlags.AllowedOps & MapOpMapTrack)
	{						   /* supplementary selectionable track ? */
	    LstPtr = (ListPtr) GetObjPtr (MapListID);
	    LstSetListChoices (LstPtr, NULL, DmNumRecordsInCategory (TrkDB, Prefs.trkcategory));
	    LstSetDrawFunction (LstPtr, &MainListDraw);
	    LstSetSelection (LstPtr, 0);
	    //      LstMakeItemVisible (LstPtr, DatumIndex);
	    LstSetHeight (LstPtr, min (LstGetNumberOfItems (LstPtr), 10));
	    PopupPtr = (ControlPtr) GetObjPtr (MapPopupID);
	    CtlSetLabel (PopupPtr, "Choose track");
	}
	else
	    HideObject (MapPopupID);
#ifdef GRAPHICALROUTEEDIT

	if ((MapFlags.AllowedOps & MapOpRteEdit) || (MapFlags.AllowedOps & MapOpRteNew))
	{						   /* edition/creation of route */
	    ShowObject (MapNewChk);
	    ShowObject (MapDelChk);
	    ShowObject (MapInsChk);
	    ShowObject (MapBtCancelID);

	    OldRouteIndex = CurrentRouteIndex;
	    error = CopyOrCreateRoute (&RteHdr, CurrentRouteIndex, MapFlags.AllowedOps & MapOpRteNew);
	    if (error)
	    {
		FrmReturnToForm (0);
		break;
	    }
	    CurrentRouteIndex = DmNumRecords (RteDB) - 1;  /* la route de travail est la derniere */
	    if (MapFlags.AllowedOps & MapOpRteNew)	   /* MapFlags.AllowedOps|=MapOpRteEdit */
		RteWptIndex = 0;
	    else
	    {
		MemHandle RecH = DmQueryRecord (RteDB, CurrentRouteIndex);
		if (RecH)
		    RteWptIndex = WptsInRteDB (RecH);
		DrawRoute (CurrentRouteIndex, !scaled, NULL);
	    }
	    MapFlags.DrawRtes = 1;
	    MainListDirty = 1;
	}
#endif
	//      ActTrkIndex=init_active_trk(); pas a sa place ?

	SetControlValue (MapWptChk, MapFlags.DrawWpts);
	SetControlValue (MapGrdChk, MapFlags.DrawGrd);
	SetControlValue (MapRteChk, MapFlags.DrawRtes);
	atProximity = 0;
	RefreshDisplay (0);
	i = StrPrintF (gbuf1, "Datum : %s", gDatum[Prefs.datum].name);
	WinDrawChars (gbuf1, i, BottomStatPos);

	break;

    case frmCloseEvent:
      MapFormClean();
	break;

    case popSelectEvent:
	if (e->data.popSelect.controlID == MapPopupID)
	{
	    MemHandle RecH;
	    UInt16 oldindex = CurrentTrackIndex;
	    CurrentTrackIndex = 0;
	    DmSeekRecordInCategory (TrkDB, &CurrentTrackIndex, e->data.popSelect.selection,
				    dmSeekForward, Prefs.trkcategory);
	    RecH = DmQueryRecord (TrkDB, CurrentTrackIndex);	/* recup du nom */
	    if (RecH)
	    {
		Object *RecP = MemHandleLock (RecH);
		StrNCopySafe (gbuf1, RecP->d.CustTrkHdr.name, 12);
		CtlSetLabel (e->data.popSelect.controlP, gbuf1);
		MemHandleUnlock (RecH);
	    }
	    if (++CurrentTrackIndex != oldindex)
		if (!oldindex)
		    DrawTrack (CurrentTrackIndex - 1, !Prefs.PermTrackIndex,
			       !Prefs.PermTrackIndex);
		else
		    RefreshDisplay (1);
	    else
	    {
		CurrentTrackIndex = 0;
		RefreshDisplay (1);
	    }
	    handled = true;
	}
	break;

    case penDownEvent:
	if (RctPtInRectangle (e->screenX, e->screenY, &DrawRect))
	{
	    //      Semicircle_Type     ct;
	    PointType ClickPt = { e->screenX, e->screenY };
	    Object o;
	    ObjectPtr newp = &o;
	    Custom_Wpt_Type ClickPos;
	    Word index;
	    static Word lastindex=dmMaxRecordIndex;
	    static UInt32 lastclick;
	    //      UInt16 l;

	    frm = FrmGetActiveForm ();			   // optimisable $$
	    MemSet (&ClickPos, sizeof (ClickPos), 0);
	    XYToLatLon (ClickPt, ClickPos.posn);
#ifdef GRAPHICALROUTEEDIT
	    //             if ((MapFlags.AllowedOps & MapOpRteEdit)||(MapFlags.AllowedOps & MapOpRteNew)){
	    if ((GetControlValue (MapNewChk) || GetControlValue (MapInsChk)) && !RteWptClicked (CurrentRouteIndex, &index, ClickPt))	/* creation or insertion of route wpt */
	    {
		if (MapFlags.DrawWpts & WptClicked (&index, ClickPt))
		    IndexToCustomWpt (index, WptDB, &ClickPos);
		else
		{
		    StrPrintF (gbuf1, "R%d-%d", 10, RteWptIndex);
		    StrNCopySafe (ClickPos.ident, gbuf1, sizeof (ClickPos.ident));	// $$
		}
		//      error = CustomToDBWpt (&ClickPos, G103Type, -1, 0);  // $$
		CustomToAnyWpt (&ClickPos, G103Type, newp, 0);
		//    error = put_wpt (&(newp->d.g), 0, 0, newp->header.type, newp->header.size);  // $$
		if (GetControlValue (MapNewChk))
		    error = EditRouteRecord (CurrentRouteIndex, RteWptIndex + 1, newp);
		else
		    error = EditRouteRecord (CurrentRouteIndex, SelectedRteWptIndex, newp);
		if (!error)
		    RteWptIndex++;

		SetControlValue (MapInsChk, 0);
		SelectedRteWptIndex = -1;
		RefreshDisplay (1);			   // $$
		DisplayWpt (&ClickPos, 1);
	    }
	    else if (GetControlValue (MapDelChk))	   /* suppression of a route wpt */
	    {
		if (RteWptClicked (CurrentRouteIndex, &index, ClickPt))
		{
		    if (!EditRouteRecord (CurrentRouteIndex, index, NULL))
			RteWptIndex--;
		    RefreshDisplay (1);			   // $$
		    SetControlValue (MapDelChk, 0);
		}
		SelectedRteWptIndex = -1;
	    }
	    else if (GetControlValue (MapRteChk)
		     && RteWptClicked (CurrentRouteIndex, &index, ClickPt))
	    {						   /* selection of a route wpt */
		LocalFormatID fmt;
		void *RecP;
		MemHandle RecH = DmQueryRecord (RteDB, CurrentRouteIndex);
		if (RecH)
		{
		    RecP = MemHandleLock (RecH);
		    fmt =
			PtrToCustomWpt ((Object *) (RecP +
						    OffsetOfObjectInRec (RecH, index)), &ClickPos);
		    MemHandleUnlock (RecH);
		    if (SelectedRteWptIndex >= 0)
			RefreshDisplay (1);
		    DisplayWpt (&ClickPos, 1);
		    SelectedRteWptIndex = index;
		}
	    }
	    else
#endif
	      if (MapFlags.DrawWpts & WptClicked (&index, ClickPt)){  /* wpt selected ? */
		if ((lastindex==index)&&(TimGetTicks()<(lastclick+SysTicksPerSecond()))) { /* double click ? */
		  CurrentRecordIndex=index;
		  FrmPopupForm (EditForm);
		}
		else {
		  int l;
		  lastclick=TimGetTicks();
		  IndexToCustomWpt (lastindex, WptDB, &ClickPos);
		  DisplayWpt (&ClickPos, 0);
		  IndexToCustomWpt (index, WptDB, &ClickPos);
		  DisplayWpt (&ClickPos, 1);
		  l=StrPrintF (gbuf1, "%s",ClickPos.cmnt);
		  WinDrawChars (gbuf1, l, BottomStatPos);
		  lastindex=index;
		}
	      }
else
	    {						   /* recentrage */
		//              long deltalon;
		Semicircle_Type CenterPos;
		XYToLatLon (ClickPt, CenterPos);
		ReCenterDisplay (&CenterPos);
		RefreshDisplay (1);
		DisplayPosition (&ClickPos.posn);
		SelectedRteWptIndex = -1;
	    }
	    handled = true;
	}
	break;

    case PvtReceivedEvent:				   /* position recue */
	{
	    //        Semicircle_Type CurrentWGS84Pos;
/*  	    Radian_Type *PositionP; */
	    D800_Pvt_Data_Type pvt;
	    Custom_Trk_Point_Type trkpt;
/*  	    SWord epex, */
/*  	      epey; */
	    static UInt32 LastValidPvt;
	    PointType wpt;
	    //      double latitude,
	    //        longitude;
	    //      UInt16 l;

	    if(!LastValidPvt)LastValidPvt=TimGetSeconds ();
	    //      SndPlaySystemSound (sndClick);
	    GarminGetPVT (&pvt);
	    //      PositionP = (Radian_Type *) & (e->data);
	    CurrentWGS84Pos.lat = pvt.posn.lat * OnePower31Pi;
	    CurrentWGS84Pos.lon = pvt.posn.lon * OnePower31Pi;
	    if (!PosVisible (CurrentWGS84Pos))
	    {
		ReCenterDisplay (&CurrentWGS84Pos);
		RefreshDisplay (1);
	    }
	    LatLonToXY (CurrentWGS84Pos, wpt);
	    if (pvt.fix >invalid){
	      RedrawBitmap (Target1ID, wpt.x, wpt.y);
	    //      epex=(UInt16)pvt.epe/LonByPixel;
	    //      epey=(UInt16)pvt.epe/LatByPixel;
	      DisplayPosition (&CurrentWGS84Pos);
	      LastValidPvt=TimGetSeconds ();
	    }
	    if (MapFlags.DrawRtes)
	    {
		if (!atProximity)
		{
		    if ((atProximity = DrawRoute (CurrentRouteIndex, 0, &CurrentWGS84Pos)))
			BeepBeep (5);
		}
		else
		    atProximity = DrawRoute (CurrentRouteIndex, 0, &CurrentWGS84Pos);
	    }
	    if (NextLogTime && (TimGetSeconds () >= NextLogTime))
	      {
		EventType evt;
		trkpt.posn.lat = CurrentWGS84Pos.lat;
		trkpt.posn.lon = CurrentWGS84Pos.lon;
		trkpt.time = pvt.wn_days * 86400 + (UInt32) (pvt.tow + 0.5);
		trkpt.alt = pvt.alt + pvt.msl_hght;
		trkpt.new_trk = NextLogTime == 1 ? 1 :  
		  TimGetSeconds ()>(NextLogTime+5*Prefs.loginterval)?1:0;
		if (pvt.fix >invalid) { /* if datas ok, schedule next alarm and switch off */
		  push_active_trkpt (trkpt);
		  if (NextLogTime == 1)
		    NextLogTime = TimGetSeconds ();
		  NextLogTime += Prefs.loginterval;
		  if (TimGetSeconds () >= NextLogTime)
		    NextLogTime = TimGetSeconds () + Prefs.loginterval;
		  //  SysTaskDelay (sysTicksPerSecond / 4);  // anti-drapage ? $$
		  SetTimeOfNextAlarm (NextLogTime + 2, 0);
		  evt.eType = keyDownEvent;
		  evt.data.keyDown.chr = hardPowerChr;
		  evt.data.keyDown.keyCode = 0;
		  evt.data.keyDown.modifiers = commandKeyMask;
		  EvtAddEventToQueue (&evt);		   /* pb si le gps s'arrete */
		}
		else {    /* datas not ok, no switch off but precaution */
		  UInt16 d=TimGetSeconds ()-LastValidPvt;
		  if (!(d%10))SndPlaySystemSound(sndError);
		  i = StrPrintF (gbuf1, "Bad PVT (last good %us ago) ",d);
		  //		  if(LastValidPvt)i += StrPrintF (gbuf1+i, "(last good %us ago)",d);
		  WinDrawChars (gbuf1, i, BottomStatPos);
		  SetTimeOfNextAlarm (TimGetSeconds ()+15, 0);
		}
	      }
	}
	handled = true;
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	{
	case MapBtOKID:
	    if (MapFlags.AllowedOps && MapOpRteNew)
	    {
		MemHandle RecH = DmQueryRecord (RteDB, CurrentRouteIndex);
		if (RecH && !WptsInRteDB (RecH))
		    DmRemoveRecord (RteDB, CurrentRouteIndex);
	    }
	    if (MapFlags.AllowedOps & MapOpRteEdit)
	    {
		DmRemoveRecord (RteDB, OldRouteIndex);
	    }
	    MainListDirty = 1;
	    MapFormClean();
	    FrmReturnToForm (0);
	    handled = true;
	    break;

	case MapBtCancelID:
	    DmRemoveRecord (RteDB, CurrentRouteIndex);
	    MapFormClean();
	    FrmReturnToForm (0);
	    handled = true;
	    break;

	case MapPlsChk:				   /* zoom + */
	    if (LonByPixel > 10)
	    {
		P0.lat += (LatByPixel * DrawRect.extent.y) / 4;
		P0.lon += (LonByPixel * DrawRect.extent.x) / 4;
		LonByPixel /= 2;
		LatByPixel /= 2;
/*  		DrawCalibrate(  P0.lat + (LatByPixel * DrawRect.extent.y) / 4, */
/*  				P0.lat + 3*(LatByPixel * DrawRect.extent.y) / 4, */
/*  				P0.lon + (LonByPixel * DrawRect.extent.x) / 4, */
/*  				P0.lon + 3*(LonByPixel * DrawRect.extent.x) / 4); */
		RefreshDisplay (1);
	    }
	    SetControlValue (MapPlsChk, 0);
	    handled = true;
	    break;

	case MapMnsChk:				   /* zoom - */
	    {
		Int32 estlon,
		  ml = MaxLonDisplayed (),
		  dl = LonByPixel * DrawRect.extent.x;
		//        P0.lat -= (LatByPixel * DrawRect.extent.y) / 2;

		//                  if (MaxLonDisplayed () >= (P0.lon / 2 + LONMAX / 2))
		if (dl < DELTALONMAX)
		{					   // bornons le zoom
		    if (ml >= (Int32) (LONMAX23 + P0.lon / 3L))
			estlon = LONMAX;		   // pas plus de 180
		    else
			estlon = P0.lon + 3L * (dl / 2);
		    //if (P0.lon <= (MaxLonDisplayed () / 2 + LONMIN / 2))
		    if (P0.lon <= (Int32) (LONMIN23 + ml / 3L))
			P0.lon = LONMIN;
		    else
			P0.lon = P0.lon - dl / 2;

		    /*        P0.lon -= (LonByPixel * DrawRect.extent.x) / 2; */
		    /*        LonByPixel *= 2;     */
		    /*        LatByPixel *= 2; */
		    ml = (LatByPixel * DrawRect.extent.y) / 2L;
		    DrawCalibrate (P0.lat - ml, P0.lat + 3L * ml, P0.lon, estlon);

		    RefreshDisplay (1);
		}
		SetControlValue (MapMnsChk, 0);
		handled = true;
	    }
	    break;

	case MapWptChk:				   /* waypoint display */
	    MapFlags.DrawWpts = e->data.ctlSelect.on;
	    if (MapFlags.DrawWpts)
		DrawWaypoints (dmAllCategories, 0);
	    else
		RefreshDisplay (1);
	    handled = true;
	    break;

	case MapRteChk:				   /* route display */
	    MapFlags.DrawRtes = e->data.ctlSelect.on;
	    atProximity = 0;
	    if (MapFlags.DrawRtes)
		DrawRoute (CurrentRouteIndex, 0, NULL);
	    else
		RefreshDisplay (1);
	    handled = true;
	    break;

	case MapGrdChk:				   /* grid display */
	    MapFlags.DrawGrd = e->data.ctlSelect.on;
	    if (MapFlags.DrawGrd)
		DrawGrid ();
	    else
		RefreshDisplay (1);
	    handled = true;
	    break;

	case MapLogChk:				   /* logging */
	    NextLogTime = 1;

	case MapTrkChk:				   /* tracking */
	    //        MapFlags.DrawActTrk = e->data.ctlSelect.on;
	    //        if (MapFlags.DrawActTrk)
	    if (e->data.ctlSelect.on)
	    {
		MustCallGpslibTicks = 1;
		error = GarminStartPVT ();
		if (error)
		{
		    UInt16 l;
		    SetControlValue (MapLogChk, 0);
		    SetControlValue (MapTrkChk, 0);
		    SndPlaySystemSound (sndError);
		    l = StrPrintF (gbuf1, "error %d when starting PVT", error);	// $$
		    WinDrawChars (gbuf1, l, BottomStatPos);	// $$
		}
		else
		{
		  if (NextLogTime) BackLightOn=HwrBacklight2(1, 0); // $$
		    //                ActTrkIndex = init_active_trk ();      // $$
		    //                DrawTrack (ActTrkIndex, 0, 0);
		}
	    }
	    else
	    {
	      BackLightOn=HwrBacklight2(1, BackLightOn); // $$
		RefreshDisplay (1);
		GarminStopPVT ();
		MustCallGpslibTicks = 0;
		GPSEndOfOperation ();
		NextLogTime = 0;
		SetTimeOfNextAlarm (NextLogTime,0);
	    }
	    handled = true;
	    break;

	case MapNewChk:				   /* new route point */
	    SetControlValue (MapInsChk, 0);
	    SetControlValue (MapDelChk, 0);
	    break;

	case MapInsChk:				   /* insert route point */
	    SetControlValue (MapNewChk, 0);
	    SetControlValue (MapDelChk, 0);
	    break;

	case MapDelChk:				   /* delete route point */
	    SetControlValue (MapNewChk, 0);
	    SetControlValue (MapInsChk, 0);
	    break;

	}
	break;

    case EndOfOperationEvent:
	//        if (MapFlags.DrawActTrk)
	if (GetControlValue (MapTrkChk))
	{						   // $$
	    UInt16 l;		       // $$
	    SndPlaySystemSound (sndError);		   // $$
	    l = StrPrintF (gbuf1, "Unexpected disconnection !");	// $$
	    WinDrawChars (gbuf1, l, BottomStatPos);	   // $$
	}						   // $$
	handled = true;
	break;

    case keyDownEvent:
      if (e->data.keyDown.modifiers & poweredOnKeyMask)HwrBacklight2(1, BackLightOn); // $$
      break;

    default:
	break;
    }

    if (e->eType != nilEvent) debug_out (GTALKDBG_CALLS, 
					 ("<-MapFormHandleEvent(%s)=%d", 
					  e->eType > LASTVALIDEVTNAME ?StrIToA(gbuf1,e->eType):EvtNames[e->eType], handled));
    return handled;
}							   // MapFormHandleEvent

static Int32
MaxLonDisplayed ()
{
    return (P0.lon + (LonByPixel * DrawRect.extent.x));
}

static Int32
MaxLatDisplayed ()
{
    return (P0.lat + (LatByPixel * DrawRect.extent.y));
}


/**
   Draws a waypoint's name, symbol and proximity circle at the good place

   @param  wp pointer on a Custom_Wpt_Type type waypoint
   @param  invert if true the waypoint display is inverted
   @return nothing
 */

static void
DisplayWpt (Custom_Wpt_Type * wp, Boolean invert)
{
    MemHandle ResH;
    PointType point;
    BitmapPtr BitP;
    static const float d2s = ((Int32) 1 << 30) / 90.0;
    int r;
#define WptYOffset 3

    LatLonToXY (wp->posn, point);
#if 1
    ResH = DmGetResource ('Tbmp', PointBMID /*  wp->smbl + DotBMID */ );
    if (ResH)
    {
	BitP = MemHandleLock (ResH);
	WinDrawBitmap (BitP, point.x - (BitP->width) / 2, point.y - (BitP->height) / 2);
	MemHandleUnlock (ResH);
	DmReleaseResource (ResH);
    }
#else
    DrawSymbol (symb_gid2lid (wp->smbl), point.x, point.y);	// $$ -> pb de centrage et transparence
#endif
    if (invert)
	WinDrawInvertedChars (wp->ident, StrLen (wp->ident),
			      point.x - FntCharsWidth (wp->ident, StrLen (wp->ident)) / 2,
			      point.y + WptYOffset);
    else
	WinDrawChars (wp->ident, StrLen (wp->ident),
		      point.x - FntCharsWidth (wp->ident, StrLen (wp->ident)) / 2, point.y + WptYOffset);

    // long delta=(d2s/60+0.5); // 1 mn
    // long delta= wp->dst/111120.*d2s+0.5;
    r = 0.5 + wp->dst / 111120. * d2s / (float) LatByPixel;	/* rayon(pixel)=d(m)/1(m)*d2s/LatByPixel */
    if (r > 3)
	Circle (point.x, point.y, r);			   // $$

}							   /* DisplayWpt */


/**
   Refresh the display.
 Redraws all elements needed on the drawing area
 of the map form.

@param  Erase if true the drawing area is erased before drawing
@return nothing
*/

static void
RefreshDisplay (Boolean Erase)
{
    if (Erase)
    {
	WinEraseRectangle (&DrawRect, 0);
	DrawRectCleared = 1;
    }
    if (MapFlags.DrawGrd)
	DrawGrid ();
    if (MapFlags.DrawTrk || MapFlags.DrawActTrk)
	DrawTrack (ActTrkIndex, 0, 0);
    if (Prefs.PermTrackIndex)
	DrawTrack (Prefs.PermTrackIndex - 1, 0, 0);
    if (CurrentTrackIndex)
	DrawTrack (CurrentTrackIndex - 1, 0, 0);
    if (MapFlags.DrawActTrk)
	DrawTrack (ActTrkIndex, 0, 0);
    if (MapFlags.DrawRtes)
	DrawRoute (CurrentRouteIndex, 0, NULL);
    if (MapFlags.DrawWpts)
	DrawWaypoints (dmAllCategories, 0);
}							   /* RefreshDisplay */


/**
   Displays a textual position at the bottom of map form.

@param  ptp pointer on the coordinates
@return nothing
*/

static void
DisplayPosition (Semicircle_Type * ptp)
{
    double latitude,
      longitude;
    int l;

    SemiToDegre (ptp, &latitude, &longitude);
    translate (FromWGS84, &latitude, &longitude, Prefs.datum);
    l = StrPrintF (gbuf1, "%s ", DegreToChars (latitude, 0, Prefs.displaymode));
    l += StrPrintF (gbuf1 + l, "%s   ", DegreToChars (longitude, 1, Prefs.displaymode));
    WinDrawChars (gbuf1, l, BottomStatPos);
}							   /* DisplayPosition */


/**
   Recenters display.
 Recenters the coordinates of the drawing area around the given coordinates. 

@param NewCenterPos pointer on the coordinates
@return
*/

static void
ReCenterDisplay (Semicircle_Type * NewCenterPos)
{
    UInt32 delta = LonByPixel * DrawRect.extent.x / 2;

    if (delta > (UInt32) (LONMAX - NewCenterPos->lon))
    {							   // depassement vers l'E ?
	NewCenterPos->lon = LONMAX - delta;
    }
    else if (delta > (UInt32) (NewCenterPos->lon - LONMIN))
    {							   // depassement vers l'W ?
	NewCenterPos->lon = LONMIN + delta;
    }
    P0.lon = NewCenterPos->lon - delta;
    P0.lat = NewCenterPos->lat - LatByPixel * DrawRect.extent.y / 2;
    DrawCalibrate (P0.lat, MaxLatDisplayed (), P0.lon, MaxLonDisplayed ());	/* recalibrage */
}



#define SELECTDIST 6

/**
   Test if a pen down event is "near" a displayed waypoint

   @param  indexp address of a variable to hold index of waypoint hit
   @param  point coordinates of the click
   @return true if click near a waypoint, index contains the waypoint's index
 */

static Boolean
WptClicked (WordPtr indexp, PointType point)
{
    ObjectHandle RecH;
    Custom_Wpt_Type wp;
    PointType p;
    UInt16 c = 0;
    UInt16 d,
      dmin = 1000;
    Int32 Lonmax = MaxLonDisplayed (),
      Latmax = MaxLatDisplayed ();

    while ((RecH =
	    (ObjectHandle) DmQueryNextInCategory (WptDB, &c,
						  dmAllCategories /*  Prefs.wptcategory */ )))
    {
	HandleToCustomWpt (RecH, &wp);
	if (WptVisible (wp))
	{
	    LatLonToXY (wp.posn, p);
	    if (p.x > point.x)
		d = p.x - point.x;
	    else
		d = point.x - p.x;
	    if (p.y > point.y)
		d += (p.y - point.y);
	    else
		d += (point.y - p.y);
	    if (d < dmin)
	    {
		dmin = d;
		*indexp = c;
	    }
	}
	c++;
    }
    if (dmin < SELECTDIST)
	return (true);
    else
	return (false);
}							   /* WptClicked */

#ifdef GRAPHICALROUTEEDIT


/**
   Tests if a pen down event is "near" a displayed route waypoint.

@param  rteindex index of the route in current route data base
@param  wptindexp pointer on index of hit waypoint
@param  point coordinates of the pen
@return true if a waypoint is hit
*/

static Boolean
RteWptClicked (Word rteindex, WordPtr wptindexp, PointType point)
{
    MemHandle RecH = NULL;
    void *RecP;
    UInt16 n,
      i;
    PointType p;
    LocalFormatID fmt;
    Custom_Wpt_Type wp;
    UInt16 d,
      dmin = 1000;
    Int32 Lonmax = MaxLonDisplayed (),
      Latmax = MaxLatDisplayed ();

    RecH = DmQueryRecord (RteDB, rteindex);
    if (RecH)
    {
	n = WptsInRteDB (RecH);
	RecP = MemHandleLock (RecH);
	for (i = 1; i <= n; i++)
	{
	    fmt = PtrToCustomWpt ((Object *) (RecP + OffsetOfObjectInRec (RecH, i)), &wp);
	    if (fmt)
	    {						   /* wpt ? */
		if (WptVisible (wp))
		{
		    LatLonToXY (wp.posn, p);
		    if (p.x > point.x)
			d = p.x - point.x;
		    else
			d = point.x - p.x;
		    if (p.y > point.y)
			d += (p.y - point.y);
		    else
			d += (point.y - p.y);
		    if (d < dmin)
		    {
			dmin = d;
			*wptindexp = i;
		    }
		}
	    }
	}
	MemHandleUnlock (RecH);
    }
    if (dmin < SELECTDIST)
	return (true);
    else
	return (false);
}							   /* RteWptClicked */

#endif


/**
   Cleaning on exit of the map form.

   @return nothing
*/

static void MapFormClean(){
  debug (GTALKDBG_CALLS, ("<->MapFormClean()"));
  if (OffScreenWH) {
    WinDeleteWindow (OffScreenWH, 0);
    OffScreenWH=NULL;
  }
  GarminStopPVT ();
  GPSEndOfOperation ();
  MustCallGpslibTicks = 0;
  MemSet(&MapFlags,sizeof(MapFlags),0);
}  /* MapFormClean */
