/**   memo writting  module for PalmOS
  @file
  @author C. Claveleira
  @date 1999-2004

*/

/* 	$Id: memo.c,v 1.7 2004/11/17 21:02:54 clavelei Exp $	 */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <stdarg.h>
#include "memo.h"

#define SIZE_INC 1000

#define memoPadAppType		'memo'
#define memoPadDBType		'DATA'
static DmOpenRef MemoPadDB;
static VoidHand MemoRech;
static ULong WriteIndex;
static UInt MemoRecIndex;

static void memo_internal_close (void) MEMO_SECTION;


/**
     Opens a new memo record in memo pad data base

 @param Header pointer on a string to put at the beginning of the memo (or NULL)
 @return 0 if OK
*/

Err
memo_open (char *Header)
{
    if (MemoPadDB)
	return (0);
    MemoPadDB = DmOpenDatabaseByTypeCreator (memoPadDBType, memoPadAppType, dmModeWrite);
    if (MemoPadDB)
      {
	  MemoRecIndex = dmMaxRecordIndex;
	  MemoRech = DmNewRecord (MemoPadDB, &MemoRecIndex, SIZE_INC);
	  WriteIndex = 0;
	  if (Header)
	      memo_write (Header);
      }
    if (!MemoRech)
	return (1);
    else
	return (0);
}							   /* memo_open */


static void
memo_internal_close ()
{
    if (MemoRech)
      {
	  MemHandleResize (MemoRech, WriteIndex + 1);
	  DmReleaseRecord (MemoPadDB, MemoRecIndex, true);
      }
    if (MemoPadDB)
	DmCloseDatabase (MemoPadDB);
    MemoRech = MemoPadDB = NULL;
}							   /* memo_internal_close */


/**
   
   Closes the current memo record

   @param  Trailer pointer on a string to put at the end of the memo or NULL
   @return nothing
*/

void
memo_close (char *Trailer)
{
    if (Trailer)
	memo_write (Trailer);
    memo_internal_close ();
}


/**
   Closes and deletes the curent memo

 @return nothing
*/

void
memo_delete ()
{
    if (MemoRech)
      {
	  DmReleaseRecord (MemoPadDB, MemoRecIndex, true);
	  DmRemoveRecord (MemoPadDB, MemoRecIndex);
	  MemoRech = NULL;
      }
}							   /* memo_delete */


/**
   returns a handle on the current log

 @return handle on record
*/

VoidHand
memo_get_handle ()
{
    return (MemoRech);
}


/**
   Writes the string given in the memo

 @param p string to log
 @return nothing
*/

void
memo_write (char *p)
{
    unsigned int l;
    Err error;
    VoidPtr RecP;

	if (MemoRech)
      {
	  l = StrLen (p) + 1;				   /* we count the null */
	  if ((WriteIndex + l) > MemHandleSize (MemoRech))
	      error = MemHandleResize (MemoRech, MemHandleSize (MemoRech) + SIZE_INC);
	  else
	      error = 0;
	  if (error) {
	      memo_internal_close ();
	      error=memo_open(NULL);  /* quid header $$ */
	  }
	  if (!error)
	    {
		RecP = MemHandleLock (MemoRech);
		DmWrite (RecP, WriteIndex, p, l);
		WriteIndex += --l;
		MemHandleUnlock (MemoRech);
	    }
      }
}							   /* memo_write */


/**
   Writes "printf" formatted messages to memo

 @param  formatStr Pointer to the format specification string
 @param ... parameters
 @return nothing
*/

void
memo_printf (CharPtr formatStr, ...)
{
    va_list args;
    char buf[256];
    SWord l = 0;

    va_start (args, formatStr);
    l = StrVPrintF (buf + l, formatStr, args);
    va_end (args);
    ErrFatalDisplayIf (l >= sizeof (buf), "memo_printf overflow");
    memo_write (buf);
}							   /* memo_printf */
