/** Routes management part of GPilotS
  @file
  @author C. Claveleira
  @date 1999-2004
  
*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
 */

/* 	$Id: route.c,v 1.16 2004/11/17 21:23:28 clavelei Exp $	 */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "gtalkRsc.h"
#include "gpslib.h"
#include "gtalk.h"
#include "route.h"
#include "gpsutil.h"
#include "log.h"
#include "dbio.h"
#include "map.h"
#include "ExgMgr.h"					   // object exchange functions

/* Prototypes : */
static void RouteListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)    ROUTE_SECTION;
double LengthOfRoute (int index, Boolean exceptlast) ROUTE_SECTION;

#define LastWptName "TOTAL:"

static Boolean uneditable;


/**
   Insertion, deletion of objects in a route record

   @param  recindex index of route record in routes database
   @param  objindex index of object in route to delete or where to insert a new object
   @param  objp pointer on object to insert. If null, delete objindex, else insert *objp at objindex
   @return 0 if ok
*/

Err
EditRouteRecord (UInt recindex, UInt objindex, Object * objp)
{
    VoidHand RouteRecH;
    VoidPtr RecP = NULL,
      p;
    ULong l,
      cut,
      s;
    UInt n;
    Err error = 0;


    debug_in (GTALKDBG_CALLS, ("->EditRouteRecord(recindex=%d,objindex=%d)", recindex, objindex));

    if (recindex >= DmNumRecords (RteDB))
	return (dmErrIndexOutOfRange);
    if (!(RouteRecH = DmGetRecord (RteDB, recindex)))
	return (DmGetLastErr ());
    s = MemHandleSize (RouteRecH);			   /* taille initiale du record */
    n = ObjectsInRec (RouteRecH);			   /* nombre d'objets */
    l = OffsetOfObjectInRec (RouteRecH, objindex);
    if ((objp == NULL) && (objindex < n))
      {							   /* delete */
	  RecP = MemHandleLock (RouteRecH);
	  cut = SizeOfObject ((Object *) (RecP + l));	   /* size of object to delete */
	  error = DmWrite (RecP, l, RecP + l + cut, s - l - cut);
	  MemHandleUnlock (RouteRecH);
	  RouteRecH = DmResizeRecord (RteDB, recindex, s - cut);
      }

    else if ((objp != NULL) && (objindex <= n))
      {							   /* insertion */
	  RouteRecH = DmResizeRecord (RteDB, recindex, s + SizeOfObject (objp));
	  RecP = MemHandleLock (RouteRecH);
	  p = MemPtrNew (s - l);			   /* zone de maneuvre */
	  if (p)
	    {
		(void) MemMove (p, RecP + l, s - l);
		error = DmWrite (RecP, l + SizeOfObject (objp), p, s - l);
		MemPtrFree (p);
	    }
	  error = DmWrite (RecP, l, objp, SizeOfObject (objp));
	  MemHandleUnlock (RouteRecH);
      }
    DmReleaseRecord (RteDB, recindex, true);
    debug_out (GTALKDBG_CALLS, ("<-EditRouteRecord returns %d", error));
    return (error);
}							   /* EditRouteRecord */


#ifndef GRAPHICALROUTEEDIT
static void WptDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText) ROUTE_SECTION;

/**
   Display waypoints name. List callback proc.

 @param  itemNum list item number
 @param  bounds bounds to draw
 @param  itemsText unused
 @return nothing
*/

static void
WptDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)
{
    UInt i;
    Custom_Wpt_Type wpt;

	i = 0;
    DmSeekRecordInCategory (WptDB, &i, itemNum, dmSeekForward, Prefs.wptcategory);
    IndexToCustomWpt (i, WptDB, &wpt);
    DrawCharsInWidth (wpt.ident, bounds->topLeft.x + 0, bounds->topLeft.y, bounds->extent.x - 5, 0);
}                      /* WptDraw */


/**
   Creation and edition of routes


 @param  e event
 @return true if nothing more to do
*/

Boolean
RteEditFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    static ListPtr WptLstPtr,
      RteLstPtr;
    UInt RecordsNumber,
      i,
      modified;
    static UInt RteRecIndex;
    static UInt OldToDisplay,
      OldCurrentIndex;
    static int PrevLstSel,
      PrevWptSel;
    VoidHand RecH;
    static VoidHand RouteRecH;
    static Boolean changed;
    VoidPtr RecP;
    FieldPtr FdPtr;
    char buf[16];
    //    D201_Rte_Hdr_Type RteHdr;
    static Custom_Rte_Hdr_Type RteHdr;
    Object obj;
    Raw_D103_Wpt_Type RteWpt;
    // Custom_Wpt_Type RteWpt;
    RectangleType r;
    Err error;
    char CategoryName[dmCategoryLength];

	debug_in (GTALKDBG_CALLS, ("->RteEditFormHandleEvent(event %d)", (int) (e->eType)));
    switch (e->eType)
      {
      case frmOpenEvent:
	frm = FrmGetActiveForm ();
	handled = true;
	OldToDisplay = Prefs.todisplay;
	RecordsNumber = DmNumRecordsInCategory (WptDB, Prefs.wptcategory);
	WptLstPtr = (ListPtr) (GetObjPtr (RteEditWptLstID));
	LstSetListChoices (WptLstPtr, NULL, RecordsNumber);
	LstSetDrawFunction (WptLstPtr, &WptDraw);
	LstSetSelection (WptLstPtr, -1);
	CategoryGetName (WptDB, Prefs.wptcategory, CategoryName);
	//      FrmCopyLabel (frm, RteEditLabel2ID, CategoryName);
	CategorySetTriggerLabel (GetObjPtr (RtEditWptPopupID), CategoryName);
	RteLstPtr = (ListPtr) (GetObjPtr (RteEditRteLstID));

	OldCurrentIndex = CurrentRecordIndex;
	error=CopyOrCreateRoute(&RteHdr,CurrentRecordIndex,(EditMode == NewWpt));
	if (error) {
	  FrmReturnToForm (0);
	  break;
	}
	StrNToField (RteHdr.cmnt, StrLen (RteHdr.cmnt), RtEditCmntID);
	StrIToA (buf, RteHdr.nmbr);
	StrNToField (buf, StrLen (buf), RtEditNmbrID);
	(void) MemMove (RteWpt.ident, LastWptName, sizeof (RteWpt.ident));
	MemSet (RteWpt.cmnt, sizeof (RteWpt.cmnt), ' ');
	error = put_rte ((GarminObject *) & RteWpt, 0, 1, G103Type, sizeof (RteWpt));	/* ajout wpt "butee" */
	if (error) {
	  FrmReturnToForm (0);
	  break;
	}
	RteRecIndex = DmNumRecords (RteDB) - 1;	   /* la route de travail est la derniere */
	CurrentRecordIndex = RteRecIndex;
	RouteRecH = DmQueryRecord (RteDB, RteRecIndex);
	if (!RouteRecH){
	  FrmReturnToForm (0);
	  break;
	}
	LstSetListChoices (RteLstPtr, NULL, WptsInRteDB (RouteRecH));
	LstSetDrawFunction (RteLstPtr, &RouteListDraw);
	LstSetSelection (RteLstPtr, -1);
	changed = 0;
	FrmDrawForm (frm);
	break;

      case frmCloseEvent:
	if (RteLstPtr != NULL)
	  DmRemoveRecord (RteDB, RteRecIndex);

	Prefs.todisplay = OldToDisplay;
	break;

      case penDownEvent:
	frm = FrmGetActiveForm ();
	GetObjectBounds (TrashBMID, &r);
	//      FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, TrashBMID), &r);
	if (RctPtInRectangle (e->screenX, e->screenY, &r)
	    && ((int) LstGetSelection (RteLstPtr) >= 0))
	  {						   /* poubelle ? */
	    WinInvertRectangle (&r, 0);
	    SndPlaySystemSound (sndInfo);
	    SysTaskDelay (sysTicksPerSecond / 5);
	    WinInvertRectangle (&r, 0);
	    if (LstGetSelection (RteLstPtr) < (WptsInRteDB (RouteRecH) - 1))
	      {
		EditRouteRecord (RteRecIndex, LstGetSelection (RteLstPtr) + 1, NULL);
		changed = 1;
	      }
	    else
	      SndPlaySystemSound (sndError);
	    LstSetListChoices (RteLstPtr, NULL, WptsInRteDB (RouteRecH));
	    LstSetSelection (RteLstPtr, -1);
	    LstDrawList (RteLstPtr);
	    handled = true;
	  }
	break;

      case lstSelectEvent:				   // un element est designe
	if (e->data.lstSelect.listID == RteEditRteLstID)
	  {						   /* selection d'un elt de route */
	    if ((int) LstGetSelection (WptLstPtr) >= 0)
	      {					   /* insertion wpt */
		i = 0;
		DmSeekRecordInCategory (WptDB, &i, LstGetSelection (WptLstPtr), dmSeekForward,
					Prefs.wptcategory);
		RecH = DmQueryRecord (WptDB, i);
		if (RecH)
		  {
		    Object *objp = MemHandleLock (RecH);
		    EditRouteRecord (RteRecIndex, LstGetSelection (RteLstPtr) + 1, objp);
		    changed = 1;
		    MemHandleUnlock (RecH);
		    LstSetListChoices (RteLstPtr, NULL, WptsInRteDB (RouteRecH));
		    LstMakeItemVisible (RteLstPtr, LstGetSelection (RteLstPtr) + 1);
		    LstDrawList (RteLstPtr);
		  }
		LstSetSelection (WptLstPtr, -1);
		LstSetSelection (RteLstPtr, -1);
	      }
	  }
	else
	  {						   /* selection d'un wpt */
	    if (LstGetSelection (WptLstPtr) == PrevWptSel)
	      LstSetSelection (WptLstPtr, -1);
	    LstSetSelection (RteLstPtr, -1);
	  }
	PrevWptSel = LstGetSelection (WptLstPtr);
	PrevLstSel = LstGetSelection (RteLstPtr);
	handled = true;
	break;

      case menuEvent:
	MenuEraseStatus (NULL);

	switch (e->data.menu.itemID)
	  {
	  }

	handled = true;
	break;

      case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID)
	  {
	  case RtEditWptPopupID:
	    frm = FrmGetActiveForm ();
	    CategorySelect (WptDB, frm, RtEditWptPopupID, RtEditWptListID, true,
			    (WordPtr) & Prefs.wptcategory, CategoryName, 1, 0);
	    RecordsNumber = DmNumRecordsInCategory (WptDB, Prefs.wptcategory);
	    LstSetListChoices (WptLstPtr, NULL, RecordsNumber);
	    LstDrawList (WptLstPtr);
	    break;

	  case RteEditBtOKID:
	    debug (GTALKDBG_INFOS,
		   ("RteEditFormHandleEvent, OK hit, objects in record=%d",
		    ObjectsInRec (RouteRecH)));
	    EditRouteRecord (RteRecIndex, WptsInRteDB (RouteRecH), NULL);	/* supp. dernier wpt factice */
	    MainListDirty = 1;
	    modified = 0;
	    FdPtr = (FieldPtr) (GetObjPtr (RtEditCmntID));
	    if (FldDirty (FdPtr))
	      {
		i = StrLen (FldGetTextPtr (FdPtr));
		StrNCopySafe (RteHdr.cmnt, FldGetTextPtr (FdPtr), sizeof (RteHdr.cmnt));
		Canonicalize (RteHdr.cmnt, StrLen (RteHdr.cmnt));
		modified = 1;
	      }
	    FdPtr = GetObjPtr (RtEditNmbrID);
	    if (FldDirty (FdPtr))
	      {
		RteHdr.nmbr = StrAToI (FldGetTextPtr (FdPtr));
		modified = 1;
	      }
	    if (modified)
	      {					   /* modifs ? */
		EditRouteRecord (RteRecIndex, 0, NULL);	/* supp. header */
		CustomToRteHdrObj (&RteHdr, &obj, G201Type);	/* conversion */
		debug (GTALKDBG_INFOS,
		       ("RteEditFormHandleEvent, put header : nmbr=%d, cmnt=%s", RteHdr.nmbr,
			RteHdr.cmnt));

		error = EditRouteRecord (RteRecIndex, 0, &obj);	/* insertion nouveau header */
	      }
	    debug (GTALKDBG_INFOS, ("RteEditFormHandleEvent, before remove, objects in record=%d",
				    ObjectsInRec (RouteRecH)));

	    if (EditMode == EditWpt)
	      {
		DmRemoveRecord (RteDB, OldCurrentIndex);
	      }
	    RteLstPtr = NULL;
	    FrmReturnToForm (0);
	    break;

	  case RteEditBtCancelID:
	    DmRemoveRecord (RteDB, RteRecIndex);
	    RteLstPtr = NULL;
	    FrmReturnToForm (0);
	    break;
	  }
	handled = true;
	break;

      default:
	break;
      }
    debug_out (GTALKDBG_CALLS, ("<-RteEditFormHandleEvent"));

	return handled;
}							   /* RteEditFormHandleEvent */
#endif


#define XName 0
#define XDist 82
#define XBear 110


/**
   Display route items in list

@param  itemNum item number in the list
@param  bounds bounds of drawing rectangle
@param  itemsText not used
@return nothing
*/

static void
RouteListDraw (Int16 itemNum, RectanglePtr bounds, CharPtr * itemsText)
{
    VoidHand RecH = NULL;
    void *RecP;
    static Semicircle_Type last_pt;
    double d;
    char buf[32];
    unsigned int l,
      bearing;
    Custom_Wpt_Type wpt,
      prevp;
    LocalFormatID fmt;

	debug_in (GTALKDBG_CALLS, ("->RouteListDraw(itemnum=%d)", (int) itemNum));
    RecH = DmQueryRecord (RteDB, CurrentRecordIndex);
    if (RecH) {
	  RecP = MemHandleLock (RecH);
	  fmt = PtrToCustomWpt ((Object *) (RecP + OffsetOfObjectInRec (RecH, itemNum + 1)), &wpt);
	  if (fmt) {						   /* wpt ? */
	    RectangleType lbounds,origclip;
	//$$	GetObjectBounds (RteEditRteLstID, &lbounds);
	    RctCopyRectangle (bounds,&lbounds );
#if 0
	    if (lbounds.topLeft.y>=10){  // pour criture dcale
	      lbounds.topLeft.y-=10;
	      lbounds.extent.y+=10;
	      lbounds.extent.x-=8; // pour scrollbar
	    }
#endif
	    WinGetClip(&origclip);
	    WinSetClip (&lbounds);
		WinDrawChars (wpt.ident, StrLen (wpt.ident), bounds->topLeft.x + XName,
			      bounds->topLeft.y);
		if (StrCompare (wpt.ident, LastWptName))   /* point intermdiaire ? */
		    if (itemNum != 0) {
		      fmt = PtrToCustomWpt ((Object *) (RecP +
							OffsetOfObjectInRec (RecH, itemNum)), &prevp);
			  if (fmt)
			    {				   /* wpt ? */
				last_pt.lat = prevp.posn.lat;
				last_pt.lon = prevp.posn.lon;
				d = Distance (last_pt.lat, last_pt.lon, wpt.posn.lat, wpt.posn.lon,
					      Prefs.distunit);
				bearing =
				    Bearing (last_pt.lat, last_pt.lon, wpt.posn.lat, wpt.posn.lon);
				l = StrPrintF (buf, "%s %s", DToA (d, 0, 2),
					       DistanceUnitNames[Prefs.distunit]);
				WinDrawChars (buf, StrLen (buf),
					      bounds->topLeft.x + XDist - FntCharsWidth (buf, l),
#if 0
					      bounds->topLeft.y - bounds->extent.y / 2);
#else
					      bounds->topLeft.y);
#endif
				if (bounds->extent.x > XBear)
				  {
				      l = StrPrintF (buf, "%d", bearing);
				      WinDrawChars (buf, StrLen (buf),
						    bounds->topLeft.x + XBear - FntCharsWidth (buf,
											       l),
#if 0
						    bounds->topLeft.y - bounds->extent.y / 2);
#else
				                    bounds->topLeft.y);
#endif
				  }
			    }
		      }
		    else;
		else
		  {					   /* total */
		      l = StrPrintF (buf, "%s %s",
				     DToA (LengthOfRoute (CurrentRecordIndex, 1), 0, 2),
				     DistanceUnitNames[Prefs.distunit]);
		      WinDrawChars (buf, StrLen (buf),
				    bounds->topLeft.x + XDist - FntCharsWidth (buf, l),
				    bounds->topLeft.y);
		  }
		WinSetClip(&origclip);
	    }						   /* if (fmt) */
	  else
	    {
		WinDrawChars ("*", 4, bounds->topLeft.x + XName, bounds->topLeft.y);	/* si pas wpt, ce doit etre un link */
		uneditable = 1;
	    }
	  MemHandleUnlock (RecH);
      }
    debug_out (GTALKDBG_CALLS, ("<-RouteListDraw"));
}							   // RouteListDraw


/**
   Route form event handler

 @param  e pointer on event to handle
 @return true if evet handled
*/

Boolean
RouteFormHandleEvent (EventPtr e)
{
    Boolean Edited,
      handled = false;
    static FormPtr frm;
    static Custom_Rte_Hdr_Type RteHdr;
    ListPtr LstPtr;
    static VoidHand CurrentRouteRecH;
    char buf[48];
    UInt category,
      l,
      attributes;
    static UInt MyCategory;
    Err error;
    char CategoryName[dmCategoryLength];
#ifdef GRAPHICALROUTEEDIT
    FieldPtr FdPtr;
    Object obj;
    Boolean       modified;
#endif

	switch (e->eType)
      {
      case frmOpenEvent:
	  frm = FrmGetActiveForm ();
	  uneditable = 0;
	  CurrentRouteRecH = DmQueryRecord (RteDB, CurrentRecordIndex);
	  if (!CurrentRouteRecH)
	    {
		FrmReturnToForm (0);
		handled = true;
		break;
	    }
	  HandleToCustomRteHdr (CurrentRouteRecH, &RteHdr);
/*  	StrPrintF (buf, "Route %d", (int) (RteHdr.nmbr)); */
/*  	FrmCopyTitle (frm, buf); */
	  LstPtr = (ListPtr) (GetObjPtr (RouteListID));
	  LstSetListChoices (LstPtr, NULL, WptsInRteDB (CurrentRouteRecH));
	  LstSetSelection (LstPtr, -1);
	  LstSetDrawFunction (LstPtr, &RouteListDraw);
	  // catgorie
	  error = DmRecordInfo (RteDB, CurrentRecordIndex, &attributes, NULL, NULL);
	  MyCategory = attributes & dmRecAttrCategoryMask;
	  CategoryGetName (RteDB, MyCategory, CategoryName);
	  CategorySetTriggerLabel (GetObjPtr (RteCatPopupID), CategoryName);
#ifdef GRAPHICALROUTEEDIT
	  StrIToA (buf, RteHdr.nmbr);
	  StrNToField (buf, StrLen (buf), RteNumID);
	  StrNToField (RteHdr.cmnt, StrLen (RteHdr.cmnt), RteCmntID);
#else
	  HideObject(RteNumID);
	  HideObject(RteCmntID);
#endif
	  l = StrPrintF (buf, "%d waypoints, total : %s %s",
			 (int) (WptsInRteDB (CurrentRouteRecH)),
			 DToA (LengthOfRoute (CurrentRecordIndex, 0), 0, 2),
			 DistanceUnitNames[Prefs.distunit]);
	  StrNToField (buf, l, RteStatFDID);
	  FrmDrawForm (frm);
	  handled = true;
	  break;

      case menuEvent:
	  MenuEraseStatus (NULL);
	  switch (e->data.menu.itemID)
	    {
	    }
	  handled = true;
	  break;

      case lstSelectEvent:
	  LstSetSelection (e->data.lstSelect.pList, -1);   /* pas de selection */
#if 0
	  GetObjectBounds (RouteListID, &DrawRect);	   /* $$ recup bound box du gadget */
//      FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, RouteListID), &DrawRect);      /* $$ recup bound box du gadget */
	  LstEraseList (e->data.lstSelect.pList);	   // $$
	  DrawRoute (CurrentRecordIndex, 1, NULL);
#else
	  MapFlags.DrawRtes = 1;
	  MapFlags.AllowedOps = MapOpActiveRte;
	  FrmPopupForm (MapForm);
#endif
	  handled = true;
	  break;

      case keyDownEvent:
	  //        frm = FrmGetActiveForm ();
	  LstPtr = (ListPtr) (GetObjPtr (RouteListID));
	  if (e->data.keyDown.chr == pageUpChr)
	      LstScrollList (LstPtr, winUp, LstGetVisibleItems (LstPtr) - 1);
	  else if (e->data.keyDown.chr == pageDownChr)
	      LstScrollList (LstPtr, winDown, LstGetVisibleItems (LstPtr) - 1);
	  break;

      case ctlSelectEvent:
	  switch (e->data.ctlSelect.controlID)
	    {
	    case RteBtOKID:
#ifdef GRAPHICALROUTEEDIT
	      MainListDirty = 1;
	      modified = 0;
	      FdPtr = (FieldPtr) (GetObjPtr (RteCmntID));
	      if (FldDirty (FdPtr))
		{
		  //                i = StrLen (FldGetTextPtr (FdPtr));
		  StrNCopySafe (RteHdr.cmnt, FldGetTextPtr (FdPtr), sizeof (RteHdr.cmnt));
		  Canonicalize (RteHdr.cmnt, StrLen (RteHdr.cmnt));
		  modified = 1;
		}
	      FdPtr = GetObjPtr (RteNumID);
	      if (FldDirty (FdPtr))
		{
		  RteHdr.nmbr = StrAToI (FldGetTextPtr (FdPtr));
		  modified = 1;
		}
	      if (modified)
		{					   /* modifs ? */
		  EditRouteRecord (CurrentRecordIndex, 0, NULL);	/* supp. header */
		  CustomToRteHdrObj (&RteHdr, &obj, G201Type);	/* conversion */
		  error = EditRouteRecord (CurrentRecordIndex, 0, &obj);	/* insertion nouveau header */
		}
#endif

	      FrmReturnToForm (0);
	      handled = true;
	      break;

	    case RteButEditID:
	      EditMode = EditWpt;
	      if (uneditable)
		FrmCustomAlert (ErrorID, "Can't edit that kind of route", " ", " ");
	      else
		{
#ifdef GRAPHICALROUTEEDIT
		  MapFlags.DrawRtes = 1;
		  MapFlags.AllowedOps = MapOpRteEdit | MapOpWptDisp;
		  FrmPopupForm (MapForm);
#else
		  FrmGotoForm (EditRouteForm);
#endif
		}
	      handled = true;
	      break;

	    case RteBtDeleteID:
	      if (FrmCustomAlert (ConfID, "Remove route", " ?", " "))
		{
		  DmRemoveRecord (RteDB, CurrentRecordIndex);
		  //              SelectionAvailable = 0;
		  MainListDirty = 1;
		  FrmReturnToForm (0);
		}
	      handled = true;
	      break;
#ifdef BEAM
	    case RteBtBeamID:
	      {
		ExgSocketType sock;
		RecordSeparator RecSep;
		void *RecP;
		ULong len;

		if (!CurrentRouteRecH) break;
		RecP = MemHandleLock (CurrentRouteRecH);;
		MemSet (&sock, sizeof (sock), 0);
		sock.target = GPilotSID;
		sock.count = 1;
		sock.length = MemHandleSize (CurrentRouteRecH);	// useless
		sock.appData = DispRte;		   // useless
		sock.time = DispRte;		   // useless
		sock.description = "Route";
		sock.name = BeamTypes[DispRte];
		sock.localMode=1; // debug $$
		RecSep.length = MemHandleSize (CurrentRouteRecH);
		RecSep.reserved = 0;

		error = ExgPut (&sock);		   // put data to destination
		if (!error)
		  {
		    len=ExgSend (&sock, &RecSep, sizeof (RecSep), &error);
		    len=0;
		    while (!error && len < MemHandleSize (CurrentRouteRecH)){
		      len+=ExgSend (&sock, RecP, MemHandleSize (CurrentRouteRecH), &error);
		    }
		    error=ExgDisconnect (&sock, error);
		  }
		else SndPlaySystemSound(sndError);
		MemHandleUnlock (CurrentRouteRecH);
	      }
	      handled = true;
	      break;
#endif
	    case RteCatPopupID:			   /* categorie */
	      //        frm = FrmGetActiveForm ();                     /* factorisable */
	      category = MyCategory;
	      Edited =
		CategorySelect (RteDB, frm, RteCatPopupID, RteCatListID, false, &MyCategory,
				CtlGetLabel (e->data.ctlSelect.pControl), 1, 0);
	      if ((MyCategory != category) && (CurrentRecordIndex != dmMaxRecordIndex))	/* a deplacer vers bouton ok */
		{
		  error = SetCategory (RteDB, CurrentRecordIndex, MyCategory);
		  MainListDirty = 1;
		}
	      handled = true;
	      break;
	    }
	  break;

      default:
	  break;
      }

	return handled;
}							   // RouteFormHandleEvent


/**
   Computes the length of a route

 @param  index index of the route to compute length of
 @param  exceptlast if true the last waypoint is ignored
 @return length in units length set in preferences
*/

double
LengthOfRoute (int index, Boolean exceptlast)
{
    VoidHand RecH = NULL;
    void *RecP;
    int n,
      i;
    unsigned long lat,
      lon;
    double dist = 0.;
    Custom_Wpt_Type wpt;

    RecH = DmQueryRecord (RteDB, index);
    if (RecH)
      {
	  n = (WptsInRteDB (RecH)) - exceptlast;
	  if (n >= 2)
	    {
		RecP = MemHandleLock (RecH);
		PtrToCustomWpt ((Object *) (RecP + SizeOfObject ((Object *) RecP)), &wpt);
		lat = wpt.posn.lat;
		lon = wpt.posn.lon;
		for (i = 1; i < n; i++)
		  {
		      PtrToCustomWpt ((Object *) (RecP + OffsetOfObjectInRec (RecH, i + 1)), &wpt);
		      dist += Distance (lat, lon, wpt.posn.lat, wpt.posn.lon, Prefs.distunit);
		      lat = wpt.posn.lat;
		      lon = wpt.posn.lon;
		  }
		MemHandleUnlock (RecH);
	    }
      }
    return (dist);
}							   /* LengthOfRoute */


/**
   Search a free route number in current category

 @return route number
*/

UInt
NewRouteNumber ()
{
    Boolean found;
    UInt i,
      j;
    VoidHand RecH;
    Custom_Rte_Hdr_Type RteHdr;

    for (i = 1; i <= 99; i++)
      {							   /* recherche numero de route libre */
	  found = 1;
	  j = 0;
	  while ((RecH = DmQueryNextInCategory (RteDB, &j, Prefs.rtecategory)))
	    {
		HandleToCustomRteHdr (RecH, &RteHdr);
		if (RteHdr.nmbr == i)
		    found = 0;
		j++;
	    }
	  if (found)
	      return (i);
      }
    return (i);
}							   /* NewRouteNumber */


/**
   Duplicate or create a route

@param  RteHdrp pointer on the new route header
@param  index index of route to duplicate (if new not set)
@param  new if true a route is to created, else given route is to be duplicated
@return error code
*/

Err CopyOrCreateRoute (Custom_Rte_Hdr_Type * RteHdrp, UInt16 index, Boolean new)
{
  VoidHand RouteRecH;
  Err error=0;

  if (new) {						   /* nouvelle route ? */
    RteHdrp->nmbr = NewRouteNumber ();
    RteHdrp->cmnt[0] = 0;
    error = put_rte ((GarminObject *) RteHdrp, 0, 1, CustomRteHdr, sizeof (Custom_Rte_Hdr_Type));	/* creation du record vide */
  }

  else {						   /* edition route existante */
    RouteRecH = DmQueryRecord (RteDB, index);
    if (RouteRecH) {					   /* duplicata route a editer */
      Object *p;
      VoidPtr RecP;
      RecP = MemHandleLock (RouteRecH);
      for (p = RecP; p < (Object *) (RecP + MemHandleSize (RouteRecH));
	   p = (void *) p + SizeOfObject (p))  {
	if (p == RecP)   {				   /* header */
	  if (PtrToCustomRteHdr (RecP, RteHdrp) != Unknown)
	    error =  put_rte ((GarminObject *) RteHdrp, 0, 1, CustomRteHdr, sizeof (Custom_Rte_Hdr_Type));
	  else error=-1;
	}
	else error = put_rte (&(p->d.g), 0, 1, p->header.type, p->header.size);
	if (error) break;
      }
      MemHandleUnlock (RouteRecH);
    }
    else error=-2;
  }
  return(error);
}                    /* CopyOrCreateRoute */

void SearchWptInRoutes(void){
}
