/** Tracks management part of GPilotS
  @file 
  @author C. Claveleira
  @date 1999-2004
  
*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* 	$Id: track.c,v 1.7 2004/11/17 21:26:01 clavelei Exp $	 */

#ifndef lint
//static char vcid[] = "$Id: track.c,v 1.7 2004/11/17 21:26:01 clavelei Exp $";
#endif /* lint */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "gtalkRsc.h"
#include "gpslib.h"
#include "gtalk.h"
#include "track.h"
#include "gpsutil.h"
#include "log.h"
#include "dbio.h"
#include "map.h"
#include "ExgMgr.h"					   // object exchange functions

static void DrawSpeed (VoidHand RecH) TRACK_SECTION;
static double TrackDistance(VoidHand RecH, Boolean WithHoles) TRACK_SECTION;

/**
   Event handler of track form

  @param  e pointer on event received
  @return true if handled
 */

Boolean
TrackFormHandleEvent (EventPtr e)
{
  Boolean Edited,
    handled = false;
  static FormPtr frm;
  UInt n;
  Object *RecP;
  static VoidHand RecH;
  UInt category,
    attributes;
  static UInt MyCategory;
  Err error;
  FieldPtr fdPtr;
  char CategoryName[dmCategoryLength];
  RectangleType r;
  double dist;

  debug_in (GTALKDBG_CALLS, ("->TrackFormHandleEvent(%s)", e->eType > LASTVALIDEVTNAME ?StrIToA(gbuf1,e->eType):EvtNames[e->eType]));
  switch (e->eType)
    {
    case frmOpenEvent:
      frm = FrmGetActiveForm ();
      // catgorie
      error = DmRecordInfo (TrkDB, CurrentRecordIndex, &attributes, NULL, NULL);
      MyCategory = attributes & dmRecAttrCategoryMask;
      CategoryGetName (TrkDB, MyCategory, CategoryName);
      CategorySetTriggerLabel (GetObjPtr (TrkCatPopupID), CategoryName);
      if (Prefs.PermTrackIndex == (CurrentRecordIndex + 1))
	SetControlValue (TrkSelID, 1);
      //          FrmSetControlValue (frm, FrmGetObjectIndex (frm, TrkSelID), 1);
      FrmDrawForm (frm);				   // position a voir
      GetObjectBounds (TrkDrawingID, &DrawRect);	   /* recup bound box du gadget */
      //      FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, TrkDrawingID), &DrawRect);     /* recup bound box du gadget */
      RecH = DmQueryRecord (TrkDB, CurrentRecordIndex);
      if (RecH)
	{
	  RecP = MemHandleLock (RecH);
	  StrNToField (RecP->d.CustTrkHdr.name, StrLen (RecP->d.CustTrkHdr.name), TrkFdNameID);
	  MemHandleUnlock (RecH);
	}
      MapFlags.DrawTrk = 1;
      DrawTrack (CurrentRecordIndex, 1, 1);
      handled = true;
      break;

    case penDownEvent:
      FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, TrkDrawingID) , &r); 
      if (RctPtInRectangle (e->screenX, e->screenY, &r)) { 
	  Prefs.PermTrackIndex = CurrentRecordIndex + 1;
	  MapFlags.AllowedOps = MapOpMapTrack;
	  FrmGotoForm (MapForm);
	handled=true; 
      } 
      break;

    case menuEvent:
      MenuEraseStatus (NULL);
      switch (e->data.menu.itemID)
	{
	case MenuEraseID:
	  SysCopyStringResource (gbuf1, StrRemoveTrkID);
	  if (FrmCustomAlert (ConfID, gbuf1, " ?", " "))
	    {
	      n = CurrentRecordIndex;
	      error =
		DmSeekRecordInCategory (TrkDB, &CurrentRecordIndex, 1, dmSeekBackward,
					*CurrentCategory);
	      DmRemoveRecord (TrkDB, n);
	      MainListDirty = 1;
	      FrmReturnToForm (0);
	    }
	  break;

	case MenuSpProfID:
	  DrawSpeed (RecH);
	  break;

	case MenuTrkDistID: {
	  FormPtr fp = NULL;
	  fp = FrmInitForm (WaitForm);
	  FrmDrawForm (fp);
	  dist=TrackDistance(RecH,0);
	  n=StrPrintF (gbuf1, "%s", DToA(dist,0,1));
	  dist=TrackDistance(RecH,1);
	  n+=StrPrintF (gbuf1+n, "/%s %s", DToA(dist,0,1), DistanceUnitNames[Prefs.distunit]);
	  FrmEraseForm (fp);
	  FrmDeleteForm (fp);
	  WinDrawChars (gbuf1, n,1, DrawRect.extent.y);
	}
	  break;

	case MenuMapID:
	  Prefs.PermTrackIndex = CurrentRecordIndex + 1;
	  //	    MapFlags.AllowedOps = 0;
	  MapFlags.AllowedOps = MapOpMapTrack;
	  FrmGotoForm (MapForm);
	  break;
#ifdef BEAM
	case MenuBeamID:
	  {  // factoriser avec route.c $$
	    ExgSocketType sock;
	    RecordSeparator RecSep;
	    void *RecP;
	    ULong len;

	    if (!RecH) break;
	    RecP = MemHandleLock (RecH);
	    MemSet (&sock, sizeof (sock), 0);
	    sock.target = GPilotSID;
	    sock.count = 1;
	    sock.length = MemHandleSize (RecH);	// useless
	    sock.appData = DispTrk;		   // useless
	    sock.time = DispTrk;		   // useless
	    sock.description = "Track";
	    sock.name = BeamTypes[DispTrk];
	    //  sock.localMode=1; // debug 
	    RecSep.length = MemHandleSize (RecH);
	    RecSep.reserved = 0;

	    error = ExgPut (&sock);		   // put data to destination
	    if (!error)
	      {
		len=ExgSend (&sock, &RecSep, sizeof (RecSep), &error);
		len=0;
		while (!error && len < MemHandleSize (RecH)){
		  len+=ExgSend (&sock, RecP, MemHandleSize (RecH), &error);
		}
		error=ExgDisconnect (&sock, error);
	      }
	    MemHandleUnlock (RecH);
	  }
	  break;
#endif

	case MenuPrefID:
	  SetPreferences ();
	  MainListDirty = 1;
	  break;

	case MenuAboutID:
	  FrmAlert (idAbout);
	  break;
	}
      handled = true;
      break;

    case ctlSelectEvent:
      switch (e->data.ctlSelect.controlID)
	{
	case TrkBtOKID:
	  fdPtr = (FieldPtr) (GetObjPtr (TrkFdNameID));
	  if (FldDirty (fdPtr) && (RecH = DmGetRecord (TrkDB, CurrentRecordIndex)))
	    {
	      RecP = MemHandleLock (RecH);
	      n = FldGetTextLength (fdPtr);
	      if (n >= sizeof (RecP->d.CustTrkHdr.name))
		n = sizeof (RecP->d.CustTrkHdr.name) - 1;
	      DmWrite (RecP, OffsetOf (Object, d.CustTrkHdr.name), FldGetTextPtr (fdPtr), n + 1);
	      MemHandleUnlock (RecH);
	      DmReleaseRecord (TrkDB, CurrentRecordIndex, 1);
	      MainListDirty = 1;
	    }
	  if (CurrentRecordIndex != dmMaxRecordIndex)
	    error = SetCategory (TrkDB, CurrentRecordIndex, MyCategory);

	case BtCancelID:
	  FrmReturnToForm (0);
	  handled = true;
	  break;

#if 0
	case TrkBtDeleteID:
	  SysCopyStringResource (gbuf1, StrRemoveTrkID);
	  if (FrmCustomAlert (ConfID, gbuf1, " ?", " "))
	    {
	      n = CurrentRecordIndex;
	      error =
		DmSeekRecordInCategory (TrkDB, &CurrentRecordIndex, 1, dmSeekBackward,
					*CurrentCategory);
	      DmRemoveRecord (TrkDB, n);
	      MainListDirty = 1;
	      //      CurrentRecordIndex = CurrentRecordIndex == 0 ? 0 : CurrentRecordIndex - 1;
	      FrmReturnToForm (0);
	    }
	  handled = true;
	  break;

	case TrkMapID:
	  Prefs.PermTrackIndex = CurrentRecordIndex + 1;
	  MapFlags.AllowedOps = 0;
	  FrmGotoForm (MapForm);
	  handled = true;
	  break;
#endif
	case TrkSelID:
	  Prefs.PermTrackIndex = e->data.ctlSelect.on ? CurrentRecordIndex + 1 : 0;
	  /*  	  MapFlags.AllowedOps= MapOpMapTrack; */
	  /*  	    FrmGotoForm (MapForm); */
	  MainListDirty = 1;
	  handled = true;
	  break;
#if 0
	case TrkSpdID:
	  DrawSpeed (RecH);
	  handled = true;
	  break;
#endif
	case TrkCatPopupID:				   /* categorie */
	  category = MyCategory;
	  Edited =
	    CategorySelect (TrkDB, frm, TrkCatPopupID, TrkCatListID, false, &MyCategory,
			    CtlGetLabel (e->data.ctlSelect.pControl), 1, 0);
	  /* 	    if ((MyCategory != category) && (CurrentRecordIndex != dmMaxRecordIndex))	// a deplacer vers bouton ok */

	  /* 		error = SetCategory (TrkDB, CurrentRecordIndex, MyCategory); */
	  handled = true;
	  break;
	}
      break;

    default:
      break;
    }

  debug_out (GTALKDBG_CALLS, ("<-TrackFormHandleEvent()=%d", handled));
  return handled;
}							   /* TrackFormHandleEvent */


/**
   Drawing of speed profile of a track

   @param RecH handle of the track
   @return nothing
 */

static void DrawSpeed (VoidHand RecH)
{
#define DRAWW 128
#define DRAWH 90
#define DRAWX 20
#define DRAWY 15
#define DISPELT 10  // minimum space between lines
#define TIMEOFFSET 15

  Object *RecP;
  Custom_Trk_Point_Type *ht=NULL;
  unsigned long prevlat = 0,
    prevlon = 0,
    prevtime = 0,
    elapsedtime = 0,
    t0 = 0;
  double dist = 0;
  float f,
    pu=0; /* value range by display unit */
  FormPtr frm;
  DateTimeType dtp;
  static char date[dateStringLength];
  static char time[timeStringLength];

  UInt n,m,
    k,
    i,l,
    prevxtime,
    U = DRAWH/DISPELT, /* number of display units */
    speed,
    speeds[DRAWW],
    maxspeed = 0;

  if (!RecH)
    return;
  frm = FrmInitForm (ChartForm);

  FrmDrawForm(frm);

  n = TrkptsInTrkDB (RecH);
  k = (n-1) / DRAWW + 1;
  RecP = MemHandleLock (RecH);
  WinDrawChars (RecP->d.CustTrkHdr.name,StrLen (RecP->d.CustTrkHdr.name),45, 1);
    SysCopyStringResource (gbuf1, StrNoTimeInTrack);
  if (RecP->d.CustTrkHdr.type == CompactTrkPt) {
    //    l=StrPrintF (gbuf1, "No time information in compact track");
    WinDrawChars (gbuf1, StrLen (gbuf1),/* DRAWX */ 2, 50);
    n=0;
  }
  else {
    ht = (Custom_Trk_Point_Type *) ((void *) RecP + SizeOfObject (RecP));
    if (ht->time==0) {
      //      l=StrPrintF (gbuf1, "No time information in this track");
      WinDrawChars (gbuf1, StrLen (gbuf1),DRAWX, 50);
      n=0;
    }
  }
  if(n>=2) {                     // start and stop times
    t0=ht->time;
    elapsedtime=(ht+n-1)->time-t0;
       
    TimSecondsToDateTime (GarminTime2PalmSeconds (t0), &dtp);
    DateToAscii (dtp.month, dtp.day, dtp.year, dfDMYWithDashes, date);
    TimeToAscii (dtp.hour, dtp.minute, tfColon24h, time);
    WinDrawChars (date,StrLen (date),5, DRAWY+DRAWH+TIMEOFFSET+3);
    WinDrawChars (time,StrLen (time),5, DRAWY+DRAWH+TIMEOFFSET+3+FntCharHeight());
    WinDrawChars ("UTC",3, 5, DRAWY+DRAWH+TIMEOFFSET+3+2*FntCharHeight());
    TimSecondsToDateTime (GarminTime2PalmSeconds ((ht+n-1)->time), &dtp);
    DateToAscii (dtp.month, dtp.day, dtp.year, dfDMYWithDashes, date);
    TimeToAscii (dtp.hour, dtp.minute, tfColon24h, time);
    WinDrawChars (date,StrLen (date),DRAWX+DRAWW-20, DRAWY+DRAWH+TIMEOFFSET+3);
    WinDrawChars (time,StrLen (time),DRAWX+DRAWW-20, DRAWY+DRAWH+TIMEOFFSET+3+FntCharHeight());
  }
  else n=0;
  prevlat = ht->posn.lat;
  prevlon = ht->posn.lon;
  prevtime = ht->time;
  prevxtime=DRAWX;

  for (i = 1; i < n; i++, ht++)
    {
      if (ht->new_trk) 
	{
	  WinDrawGrayLine (DRAWX + (i-1) / k, DRAWY+DRAWH, 
			   DRAWX + ((float)DRAWW*(float)(prevtime-t0))/(float)elapsedtime, DRAWY+DRAWH+TIMEOFFSET);
	  WinDrawGrayLine (DRAWX + i / k, DRAWY+DRAWH, 
			   DRAWX + ((float)DRAWW*(float)(ht->time-t0))/(float)elapsedtime, DRAWY+DRAWH+TIMEOFFSET);

	  /* 		WinDrawLine (DRAWX + ((float)DRAWW*(float)(prevtime-t0))/(float)elapsedtime, DRAWH+TIMEOFFSET,  */
	  /* DRAWX + ((float)DRAWW*(float)(ht->time-t0))/(float)elapsedtime, DRAWH+TIMEOFFSET); */
	  prevtime = ht->time;
	  dist = 0;
	  prevxtime=DRAWX + ((float)DRAWW*(float)(ht->time-t0))/(float)elapsedtime;
	}
      else
	{
	  dist += Distance (prevlat, prevlon, ht->posn.lat, ht->posn.lon, Prefs.distunit);
	}

      if (!(i % k))
	{
	  speed =	ht->time ==  prevtime ? 0 : 
	    (UInt) (dist / (double) (ht->time - prevtime) * 36000. + 0.5);
	  speeds[i / k] = speed;

	  if (speed > maxspeed)
	    maxspeed = speed;
	  prevtime = ht->time;
	  dist = 0;
	  WinDrawLine (DRAWX + i / k, DRAWY+DRAWH, DRAWX + i / k, DRAWY+DRAWH);
	  WinDrawLine(prevxtime, DRAWY+DRAWH+TIMEOFFSET,
		      DRAWX + ((float)DRAWW*(float)(ht->time-t0))/(float)elapsedtime, DRAWY+DRAWH+TIMEOFFSET);
	  prevxtime=DRAWX + ((float)DRAWW*(float)(ht->time-t0))/(float)elapsedtime;
	}
      prevlat = ht->posn.lat;
      prevlon = ht->posn.lon;
    }
  MemHandleUnlock (RecH);
  if (n){
    /*   WinDrawLine(prevxtime, DRAWY+DRAWH+TIMEOFFSET, */
    /* 	      DRAWX + DRAWW, DRAWY+DRAWH+TIMEOFFSET); */
    WinDrawGrayLine (DRAWX + (i-1) / k, DRAWY+DRAWH, 
		     DRAWX + DRAWW , DRAWY+DRAWH+TIMEOFFSET);
    for (f=1e-6;f<1e6;f*=10.){
      pu=f;m=1;
      if ((pu*U)>=maxspeed)break;
      pu=f*2.;m=2;
      if ((pu*U)>=maxspeed)break;
      pu=f*5.;m=5;
      if ((pu*U)>=maxspeed)break;
    }
    WinDrawLine (DRAWX, DRAWY+DRAWH, DRAWX, DRAWY);
    for(i=0;i<=(U);i++) {       /* y graduations */
      WinDrawLine (DRAWX-2, DRAWY+DRAWH-(i*DISPELT), 
		   DRAWX,DRAWY+DRAWH-(i*DISPELT) );
      WinDrawGrayLine (DRAWX, DRAWY+DRAWH-(i*DISPELT), 
		       DRAWX+DRAWW, DRAWY+DRAWH-(i*DISPELT) );
      l=StrPrintF (gbuf1, "%s", DToA(pu*i/10.,0,1));
      WinDrawChars (gbuf1, StrLen (gbuf1),DRAWX-FntCharsWidth(gbuf1,l)-2, DRAWY+DRAWH-FntCharHeight()/2-(i*DISPELT));
    }

    f=(float)DRAWH/(pu*U);
    for (i = 1; (i * k) < n; i++)
      WinDrawLine (DRAWX + i, DRAWY+DRAWH, DRAWX + i, DRAWY+DRAWH - (long) (speeds[i] * f) );
    StrPrintF (gbuf1, "%s/h", DistanceUnitNames[Prefs.distunit]);
    WinDrawChars (gbuf1, StrLen (gbuf1), 1, 1);
  }
  FrmDoDialog (frm) ;
  FrmDeleteForm (frm);
}							   /* DrawSpeed */


/**
   Compute cumulative length of a track

 @param  RecH handle on the track
 @param  WithHoles if true holes are counted
 @return computed distance
*/

static double TrackDistance(VoidHand RecH, Boolean WithHoles){
  UInt n,i;
  Custom_Trk_Point_Type *ht;
  double dist = 0;
  Object *RecP;
  unsigned long prevlat = 0,
    prevlon = 0;
  static  VoidHand oldRecH;
  static double distwithholes, distwithoutholes;
  static UInt oldcount;

  if (!RecH)
    return(0.0);
  n = TrkptsInTrkDB (RecH);
  if ( (RecH==oldRecH) && (n==oldcount))   /* in cache ? */
    return (WithHoles?distwithholes:distwithoutholes);
  oldcount=n;
  oldRecH=RecH;
  distwithholes= distwithoutholes=0.0;
  RecP = MemHandleLock (RecH);
  ht = (Custom_Trk_Point_Type *) ((void *) RecP + SizeOfObject (RecP));
  prevlat = ht->posn.lat;
  prevlon = ht->posn.lon;
  if(n>=2)for (i = 1; i < n; i++, ht++)
    {
      dist = Distance (prevlat, prevlon, ht->posn.lat, ht->posn.lon, Prefs.distunit);
      distwithholes+=dist;
      if (!(ht->new_trk) ) distwithoutholes +=dist;
      prevlat = ht->posn.lat;
      prevlon = ht->posn.lon;
    }
  MemHandleUnlock (RecH);

  return(WithHoles?distwithholes:distwithoutholes);
}                                      /* TrackDistance */


#ifdef EXPORT

ExportTrack(){
  double latitude,
    longitude;
  char *p;

  memo_open ("Track\n");
  m_speed=Distance (prevlat, prevlon, ht->posn.lat, ht->posn.lon, Prefs.distunit)/(double) (ht->time - m_prevtime) * 3600.;
  SemiToDegre (&ht->posn, &latitude, &longitude);
  p = DegreToChars (latitude, 0, Prefs.displaymode);
  l = StrPrintF (gbuf1, "T%s\t",ht->new_trk?">":" ",  p);
  p = DegreToChars (longitude, 1, Prefs.displaymode);
  l += StrPrintF (gbuf1 + l, "%ld, s=%s\n",ht->time-m_prevtime,DToA(m_speed,0,2) );
  memo_write (gbuf1);
  m_prevtime=ht->time;
  memo_close (NULL);
}
#endif
