/** GPilotS common datas types 
  @file
  @author C. Claveleira
  @date 1999-2004
 Data types and enumerations used by different modules of GPilotS are
 defined in this file. It can be used to generate code not only for PalmOS,
 but also especialy for GLIB or standard Unix code.
*/

// $Id: types.h,v 1.7 2004/11/17 21:28:05 clavelei Exp $

#ifndef __TYPES_H__
#define __TYPES_H__

#if defined __palmos__            // types used are PalmOS' ones
# define PACKED
#elif defined GLIB_MAJOR_VERSION  //  type castings for glib
  typedef guint32 UInt32 ;
  typedef gint32 Int32 ;
  typedef guint32 ULong;
  typedef guint16 UInt;
  typedef guint16 UInt16;
  typedef gint16 Int16;
  typedef guint16 Err;
  typedef guint16 Word;
  typedef gint16 SWord;
  typedef char Char;
  typedef unsigned char Boolean;
  typedef void *FieldPtr;
  typedef void *MemHandle;
  typedef void *DmOpenRef;
  typedef void *RectanglePtr;
  typedef void *MemPtr;
# define MemSet(dst,num,value) memset(dst,value,num)
# define MemMove memmove
# define StrLen strlen
# define MemHandleLock(h) (h)
# define MemHandleUnlock(h)
# define min(a, b) (((a) < (b)) ? (a) : (b))
# define PACKED __attribute__ ((packed))
#else                            // default
  typedef unsigned int UInt16;
  typedef int Int16;
  typedef unsigned long UInt32;
  typedef long Int32;
# define PACKED __attribute__ ((packed))
#endif

/** Distance unit
 */
enum DistanceUnit
{
    metric, nautical, statute
};
typedef enum DistanceUnit DistanceUnit;	/* unites de distance */

/** Display format of coordinates
 */
typedef enum			       /* display mode (warning : other structures use it) */
{
    DMM, DMS, DDD
}
DisplayFormat;

/** Internal waypoint format of GPilotS
 */
typedef struct			       /* structure de waypoint "interne" */
{
    char ident[51];		       /* identifier (50 + '0') */
    Semicircle_Type posn;	       /* position (common to all Garmin types) */
    char cmnt[51];		       /* comment (50 + '0') */
    float dst;			       /* proximity distance */
    float alt;			       /* altitude */
    Int16 smbl;			       /* symbol id */
    byte dspl;			       /* display option */
    byte color;			       /* color */
}
Custom_Wpt_Type;                       /* size : 124 bytes */

/** Full size overlay type for D108 format
 */
typedef struct			       /* super D108 with place for full strings */
{
    Raw_D108_Wpt_Type w;
    char ident[51];		       /* variable length string   1-51 */
    char comment[51];		       /* waypoint user comment    1-51 */
    char facility[31];		       /* facility name            1-31 */
    char city[25];		       /* city name                1-25 */
    char addr[51];		       /* address number           1-51 */
    char cross_road[51];	       /* intersecting road label  1-51 */
}
B108_Wpt_Type;                         /* size : 308 bytes */

/** Full size overlay type for D109 format
 */
typedef struct			       /* super D109 with place for full strings (size = 312) */
{
    Raw_D109_Wpt_Type w;	       /*                           52 */
    char ident[51];		       /* variable length string  1-51 */
    char comment[51];		       /* waypoint user comment   1-51 */
    char facility[31];		       /* facility name           1-31 */
    char city[25];		       /* city name               1-25 */
    char addr[51];		       /* address number          1-51 */
    char cross_road[51];	       /* intersecting road label 1-51 */
}
B109_Wpt_Type;                         /* size : 312 bytes */

// typedef struct
// {                                   /* super D202 with place for full string */
//     char rte_ident[255];            /* null-terminated string */
// }
// B202_Rte_Hdr_Type;

/** Internal route header format of GPilotS
 */
typedef struct			       /* internal route header */
{
    byte nmbr;			       /* route number */
    char cmnt[128];		       /* comment or ident */
}
Custom_Rte_Hdr_Type;                   /* size : 129 bytes */

/** Internal track header format
 */
typedef struct			       /* internal track header */
{
    char name[256];		       /* nom du groupe de trackpoints */
    byte dspl;			       /* display on the map ? */
    byte color;			       /* color */
    LocalFormatID type;		       /* type of following track points */
    unsigned char unused;              /* padding to align size of struct between platforms */
    UInt16 number;	       /* number of track points */
    Int32 latmin,		       /* latitude min */

      latmax,			       /* latitude max */

      lonmin,			       /* longitude min */

      lonmax;			       /* longitude max */
    unsigned char unused2;              /* padding to align size of struct between platforms */
} PACKED
Custom_Trk_Hdr_Type;                   /* size : 278 bytes */

/** Standard internal track point format
 */
typedef struct
{
    Semicircle_Type posn;	       /* position */
    UInt32 time;		       /* time */
    float alt;			       /* altitude in meters */
    boolean new_trk;		       /* new track segment? */
    unsigned char unused;              /* padding to align size of struct between platforms */
} PACKED
Custom_Trk_Point_Type;                /* size : 18 bytes */

/** Compact internal track point format
 */
typedef struct			       /* custom compact track point type */
{
    Semicircle_Type posn;	       /* position */
    boolean new_trk;		       /* new track segment? */
    unsigned char unused;              /* padding to align size of struct between platforms */
} PACKED
Compact_Trk_Point_Type;                /* size : 10 bytes */

/** Structure of objects manipulated by GPilotS
 */
typedef struct			       /* structure of objects stored in records (324 bytes) */
{
    struct
    {
	LocalFormatID type;	       /* type of following datas */
	UInt16 size;	       /* size of following datas */
	Int32 version;		       /* version and spare room */
    }
    header;
    union
    {
      GarminObject g;  /* all Garmin objects (254 bytes) */
/*      Raw_D100_Wpt_Type d100; */
/*      Raw_D102_Wpt_Type d102; */
/*      Raw_D103_Wpt_Type d103; */
/*      Raw_D104_Wpt_Type d104; */
/*      Raw_D107_Wpt_Type d107; */
/*      Raw_D108_Wpt_Type d108; */
      B108_Wpt_Type b108;
      B109_Wpt_Type b109;  /* 312 bytes */
/*      Raw_D300_Trk_Point_Type d300; */
/*      D200_Rte_Hdr_Type d200; */
/*      D201_Rte_Hdr_Type d201; */
/*      D202_Rte_Hdr_Type d202; */
/*  	B202_Rte_Hdr_Type b202; */
      Custom_Wpt_Type CustWpt;
      Custom_Rte_Hdr_Type CustRteHdr;
      Custom_Trk_Hdr_Type CustTrkHdr;
    }
    d;
    byte pad[3];
}
Object;                  /* size : 324 bytes */

union TrkPts
{
    Custom_Trk_Point_Type custom;
    Compact_Trk_Point_Type compact;
};				       /* union of internal track points formats */
typedef union TrkPts TrkPts;

typedef Object *ObjectPtr;	       // 
typedef ObjectPtr *ObjectHandle;       // Object Handle


#endif
